/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.LocalDate;
import java.util.Comparator;
import java.util.Set;

import javax.annotation.PostConstruct;
import javax.transaction.Transactional;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Profile;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.PersonControllerApi;
import nl.tudelft.tam.enums.AcademicLevel;
import nl.tudelft.tam.enums.AcademicPeriod;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.*;
import nl.tudelft.tam.repository.*;

/**
 * Class to be used to load the database with pre-determined data before running the rest of the application.
 * <p>
 * This class may be implemented for a development environment. Be careful with using this class for
 * production, however. The data initialized in this class will be added every startup. If a truly persistent
 * database is used, it might need to be checked that no such object exists yet, indeed you may not want to
 * use this class at all within production.
 */
@Service
@Transactional
@Profile("dev")
public class DevDatabaseLoader {

	@Autowired
	JobOfferRepository jobOfferRepository;

	@Autowired
	ExtraWorkRepository extraWorkRepository;

	@Autowired
	PersonControllerApi personApi;

	@Autowired
	ProfileRepository profileRepository;

	@Autowired
	ApplicationRepository applicationRepository;

	@Autowired
	DeclarationRepository declarationRepository;

	@Autowired
	CoordinatorRepository coordinatorRepository;

	@Autowired
	TrainingTypeRepository trainingTypeRepository;

	@Value("${tam.filesys.payscale-proofs-dir}")
	private String RR_PROOF_DIRECTORY;

	final long ID_OOP_NOW = 2L;
	final long ID_ADS_NOW = 4L;
	final long ID_SQT_NOW = 6L;

	TrainingType regularCSE;

	JobOffer oop_now_jo_ta;
	JobOffer oop_now_jo_mentor;
	JobOffer ads_now_jo_ta;
	JobOffer ads_now_jo_mentor;
	JobOffer sqt_now_jo_ta;
	JobOffer sqt_now_jo_mentor;

	ExtraWork oop_now_ew_grading;
	ExtraWork oop_now_ew_oral_check;
	ExtraWork ads_now_ew_grading;
	ExtraWork sqt_now_ew_grading;

	nl.tudelft.tam.model.Profile cseteacher1;
	nl.tudelft.tam.model.Profile cseteacher2;
	nl.tudelft.tam.model.Profile csestudent1;
	nl.tudelft.tam.model.Profile csestudent2;

	Application oop_now_jo_ta_app_st1;
	Application oop_now_jo_ta_app_st2;
	Application oop_now_jo_men_app_st1;
	Application ads_now_jo_ta_app_st1;
	Application ads_now_jo_men_app_st1;
	Application ads_now_jo_ta_app_st2;

	Declaration oop_now_ew_grading_dec_st1_1;
	Declaration oop_now_ew_grading_dec_st1_2;

	/**
	 * Initializes the database by delegating to methods initializing individual entity types. Each
	 * initialization method should cover initializing all objects of that type.
	 */
	@PostConstruct
	private void init() throws IOException {
		initTrainingApprovals();
		initJobOffers();
		initExtraWork();
		fetchPeople();
		initApplications();
		initDeclarations();
		initRaiseRequests();
	}

	private void initJobOffers() {
		oop_now_jo_ta = jobOfferRepository.save(JobOffer.builder()
				.name("TA")
				.editionId(ID_OOP_NOW)
				.contractName("OOP-NOW-TA")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("OOP TA BAAN CODE")
				.maxHours(100)
				.requiredTrainings(Set.of(regularCSE))
				.hiringMessage("Could you provide us with the following information: "
						+ "Are you available on Tuesdays and/or Fridays? "
						+ "What is the maximum of hours you are available?")
				.period(AcademicPeriod.Q1)
				.level(AcademicLevel.BACHELOR1)
				.build());
		oop_now_jo_mentor = jobOfferRepository.save(JobOffer.builder()
				.name("Mentor (closes in a week)")
				.editionId(ID_OOP_NOW)
				.contractName("OOP-NOW-Mentor")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("OOP Mentor BAAN CODE")
				.maxHours(100)
				.requiredTrainings(Set.of(regularCSE))
				.hiringMessage("This is a different, shorter, hiring message!")
				.deadline(LocalDate.now().plusDays(7))
				.period(AcademicPeriod.S1)
				.level(AcademicLevel.BACHELOR1)
				.build());
		ads_now_jo_ta = jobOfferRepository.save(JobOffer.builder()
				.name("TA (closes in a week)")
				.editionId(ID_ADS_NOW)
				.contractName("ADS-NOW-TA")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("ADS TA BAAN CODE")
				.maxHours(100)
				.requiredTrainings(Set.of(regularCSE))
				.deadline(LocalDate.now().plusDays(7))
				.period(AcademicPeriod.Q2)
				.level(AcademicLevel.BACHELOR1)
				.build());
		ads_now_jo_mentor = jobOfferRepository.save(JobOffer.builder()
				.name("Mentor")
				.editionId(ID_ADS_NOW)
				.contractName("ADS-NOW-Mentor")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("ADS Mentor BAAN CODE")
				.maxHours(100)
				.requiredTrainings(Set.of(regularCSE))
				.rejectMessage("Sorry :(")
				.period(AcademicPeriod.SUMMER)
				.level(AcademicLevel.BACHELOR1)
				.build());
		sqt_now_jo_ta = jobOfferRepository.save(JobOffer.builder()
				.name("TA")
				.editionId(ID_SQT_NOW)
				.contractName("SQT-NOW-TA")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("SQT TA BAAN CODE")
				.maxHours(100)
				.requiredTrainings(Set.of(regularCSE))
				.period(AcademicPeriod.Q4)
				.level(AcademicLevel.BACHELOR1)
				.build());
		sqt_now_jo_mentor = jobOfferRepository.save(JobOffer.builder()
				.name("Mentor (closed)")
				.editionId(ID_SQT_NOW)
				.contractName("SQT-NOW-Mentor")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("SQT Mentor BAAN CODE")
				.maxHours(100)
				.requiredTrainings(Set.of(regularCSE))
				.period(AcademicPeriod.Q4)
				.level(AcademicLevel.BACHELOR3)
				.deadline(LocalDate.now().minusDays(1))
				.build());
	}

	private void initExtraWork() {
		oop_now_ew_grading = extraWorkRepository.save(ExtraWork.builder()
				.name("grading (no deadline)")
				.editionId(ID_OOP_NOW)
				.contractName("OOP-NOW-Grading")
				.baanCode("OOP BAAN CODE")
				.maxHours(100)
				.deadline(null)
				.build());
		oop_now_ew_oral_check = extraWorkRepository.save(ExtraWork.builder()
				.name("oral check (closes today)")
				.editionId(ID_OOP_NOW)
				.contractName("OOP-NOW-OralCheck")
				.baanCode("OOP BAAN CODE")
				.maxHours(100)
				.deadline(LocalDate.now())
				.build());
		ads_now_ew_grading = extraWorkRepository.save(ExtraWork.builder()
				.name("grading (closes in a week)")
				.editionId(ID_ADS_NOW)
				.contractName("ADS-NOW-Grading")
				.baanCode("ADS BAAN CODE")
				.maxHours(100)
				.deadline(LocalDate.now().plusDays(7))
				.build());
		sqt_now_ew_grading = extraWorkRepository.save(ExtraWork.builder()
				.name("grading (closed)")
				.editionId(ID_SQT_NOW)
				.contractName("SQT-NOW-Grading")
				.baanCode("SQT BAAN CODE")
				.maxHours(100)
				.deadline(LocalDate.now().minusDays(1))
				.build());
	}

	private void fetchPeople() {
		cseteacher1 = profileRepository.save(nl.tudelft.tam.model.Profile.builder()
				.id(3L).emailNotifications(false).build());
		cseteacher2 = profileRepository.save(nl.tudelft.tam.model.Profile.builder()
				.id(4L).emailNotifications(false).build());
		csestudent1 = profileRepository.save(nl.tudelft.tam.model.Profile.builder()
				.id(9L).emailNotifications(false).build());
		csestudent2 = profileRepository.save(nl.tudelft.tam.model.Profile.builder()
				.id(10L).emailNotifications(false).build());
		var admin1 = profileRepository.save(nl.tudelft.tam.model.Profile.builder()
				.id(1L).emailNotifications(false).build());

		// Teacher 1 => Coordinator CSE
		coordinatorRepository.save(Coordinator.builder().personId(cseteacher1.getId()).programId(1L).build());
		// Teacher 2 => Coordinator EE
		coordinatorRepository.save(Coordinator.builder().personId(cseteacher2.getId()).programId(2L).build());
		// Admin 1 => Coordinator CSE
		coordinatorRepository.save(Coordinator.builder().personId(admin1.getId()).programId(1L).build());
	}

	private void initApplications() {
		oop_now_jo_ta_app_st1 = applicationRepository.save(Application.builder()
				.id(new Application.AppId(csestudent1.getId(), oop_now_jo_ta.getId()))
				.jobOffer(oop_now_jo_ta)
				.status(Status.ACCEPTED)
				.build());
		oop_now_jo_ta_app_st2 = applicationRepository.save(Application.builder()
				.id(new Application.AppId(csestudent2.getId(), oop_now_jo_ta.getId()))
				.jobOffer(oop_now_jo_ta)
				.status(Status.OFFERED)
				.build());
		oop_now_jo_men_app_st1 = applicationRepository.save(Application.builder()
				.id(new Application.AppId(csestudent1.getId(), oop_now_jo_mentor.getId()))
				.jobOffer(oop_now_jo_mentor)
				.status(Status.OFFERED)
				.build());

		ads_now_jo_ta_app_st1 = applicationRepository.save(Application.builder()
				.id(new Application.AppId(csestudent1.getId(), ads_now_jo_ta.getId()))
				.jobOffer(ads_now_jo_ta)
				.status(Status.ACCEPTED)
				.build());
		ads_now_jo_men_app_st1 = applicationRepository.save(Application.builder()
				.id(new Application.AppId(csestudent1.getId(), ads_now_jo_mentor.getId()))
				.jobOffer(ads_now_jo_mentor)
				.status(Status.REJECTED_BY_TEACHER)
				.build());
		ads_now_jo_ta_app_st2 = applicationRepository.save(Application.builder()
				.id(new Application.AppId(csestudent2.getId(), ads_now_jo_ta.getId()))
				.jobOffer(ads_now_jo_ta)
				.status(Status.OFFERED)
				.build());
	}

	private void initDeclarations() {
		oop_now_ew_grading_dec_st1_1 = declarationRepository.save(Declaration.builder()
				.person(csestudent1)
				.extraWork(oop_now_ew_grading)
				.amount(10)
				.reason("Graded 20 exams.")
				.status(Status.ACCEPTED)
				.build());
		oop_now_ew_grading_dec_st1_2 = declarationRepository.save(Declaration.builder()
				.person(csestudent1)
				.extraWork(oop_now_ew_grading)
				.amount(5)
				.reason("Graded 1 exam.")
				.status(Status.SUBMITTED)
				.build());
	}

	private void initRaiseRequests() throws IOException {
		Path proofs = Path.of(RR_PROOF_DIRECTORY);

		// create proofs dir if it doesn't exist
		if (!Files.exists(proofs)) {
			proofs.toFile().mkdirs();
			return;
		}

		// if everything exists, just delete all proofs
		Files.walk(proofs)
				.sorted(Comparator.reverseOrder())
				.filter(p -> !p.equals(proofs))
				.map(Path::toFile)
				.forEach(File::delete);
	}

	private void initTrainingApprovals() {
		regularCSE = TrainingType.builder().name("Regular Training").isDefault(true).programId(1L).build();
		TrainingType regularEE = TrainingType.builder().name("Regular Training").isDefault(true).programId(2L)
				.build();
		TrainingType team = TrainingType.builder().name("Team Training").programId(1L).build();

		trainingTypeRepository.save(regularCSE);
		trainingTypeRepository.save(regularEE);
		trainingTypeRepository.save(team);
	}
}
