/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.io.IOException;
import java.time.LocalDateTime;
import java.util.*;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.Resource;
import org.springframework.data.domain.Pageable;
import org.springframework.data.web.PageableDefault;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;

import nl.tudelft.labracore.api.dto.ProgramDetailsDTO;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.controller.utility.PageUtil;
import nl.tudelft.tam.dto.create.ApplicationCreateDTO;
import nl.tudelft.tam.dto.util.ApplicationFilterDTO;
import nl.tudelft.tam.dto.view.details.ApplicationDetailsJobOfferDTO;
import nl.tudelft.tam.enums.AcademicPeriod;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.Application;
import nl.tudelft.tam.service.*;

@Controller
@RequestMapping("application")
public class ApplicationController {

	// region class attributes

	@Autowired
	ApplicationService applicationService;

	@Autowired
	CSVService csvService;

	@Autowired
	EditionService editionService;

	@Autowired
	ProgramService programService;

	// endregion

	// region view pages

	/**
	 * Gets the coordinator page.
	 *
	 * @param  person The authenticated person
	 * @param  filter The filter to apply to the applications
	 * @param  model  The model to add data to
	 * @return        The coordinator page
	 */
	@GetMapping("all")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public String getAllApplications(@AuthenticatedPerson Person person,
			ApplicationFilterDTO filter,
			@PageableDefault(value = 25) Pageable pageable,
			Model model) {
		List<ProgramDetailsDTO> programs = programService.getCoordinatingPrograms(person.getId());
		if (filter.getProgram() == null) {
			filter.setProgram(programs.get(0).getId());
		}

		List<Application> applications = applicationService
				.getFilteredCoordinatingApplications(person.getId(), filter);

		model.addAttribute("applications", PageUtil.pageFromList(applications, pageable));
		model.addAttribute("programs", programs);

		return "application/all";
	}

	/**
	 * Gets the coordinator page, with FlexDelft filters.
	 *
	 * @param  person The authenticated person
	 * @param  filter The filter to apply to the applications
	 * @param  model  The model to add data to
	 * @return        The coordinator page
	 */
	@GetMapping("all/flexdelft")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public String getAllApplicationsForFlexDelft(@AuthenticatedPerson Person person,
			ApplicationFilterDTO filter,
			@PageableDefault(value = 25) Pageable pageable,
			Model model) {
		List<ProgramDetailsDTO> programs = programService.getCoordinatingPrograms(person.getId());
		if (filter.getProgram() == null) {
			filter.setProgram(programs.get(0).getId());
		}

		filter.setStatuses(List.of(Status.ACCEPTED));
		List<Application> applications = applicationService
				.getFilteredCoordinatingApplications(person.getId(), filter);

		model.addAttribute("applications", PageUtil.pageFromList(applications, pageable));
		model.addAttribute("exports",
				applicationService.getExportsForPersonAndProgram(person, filter.getProgram()));
		model.addAttribute("programs", programs);

		return "application/all_flexdelft";
	}

	/**
	 * Gets all applications for a user.
	 *
	 * @param  person The currently logged in person
	 * @param  q      The search query
	 * @param  model  The model to add the data to
	 * @return        The applications page
	 */
	@GetMapping("my")
	public String getApplications(@AuthenticatedPerson Person person,
			@RequestParam(required = false) String q,
			Model model) {
		List<ApplicationDetailsJobOfferDTO> applications = applicationService
				.getFilteredApplicationsForPerson(person.getId(), q)
				.stream().filter(a -> a.getJobOffer().getEdition().getEndDate().isAfter(LocalDateTime.now()))
				.toList();

		model.addAttribute("applications", applications);
		return "application/view_many";
	}

	// endregion

	// region manipulate data

	/**
	 * Promotes/Changes role to head TA (promote) for an already accepted student TA for that job offer
	 *
	 * @param  personId The student TA who will be promoted
	 * @param  offerId  The job offer for which the student will be promoted
	 * @return          The job offer page (i.e. from where the promotion was made)
	 */
	@PostMapping("promote/{personId}/{offerId}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#offerId)")
	public String promote(@PathVariable Long personId, @PathVariable Long offerId) {
		applicationService.promote(personId, offerId);
		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Changes role of person to TA (demote) for the job offer
	 *
	 * @param  personId The student TA whose role will change
	 * @param  offerId  The job offer for which the student's role will change
	 * @return          The job offer page (i.e. from where the action was taken to change)
	 */
	@PostMapping("demote/{personId}/{offerId}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#offerId)")
	public String demote(@PathVariable Long personId, @PathVariable Long offerId) {
		applicationService.demote(personId, offerId);
		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Submits an application for a student to a specific job offer.
	 *
	 * @param  person The person making the application (ie. student)
	 * @param  create The application create DTO
	 * @param  page   The page the user was previously on (to redirect to)
	 * @return        The page to redirect the user to (either my applications or previous is provided)
	 */
	@PostMapping("submit")
	@PreAuthorize("@authorisationService.canSubmitApplication(#create.jobOffer.id)")
	public String submitApplication(@AuthenticatedPerson Person person,
			ApplicationCreateDTO create,
			@RequestParam(required = false) String page,
			@RequestParam(required = false) Long program,
			@RequestParam(required = false, name = "period") List<AcademicPeriod> filterPeriods) {
		applicationService.submit(person.getId(), create);

		String url = "applications".equals(page) ? "application/my" : "job-offer/open";

		Map<String, List<String>> params = new HashMap<>();
		if (program != null)
			params.put("program", List.of(program));
		if (filterPeriods != null)
			params.put("period",
					filterPeriods.stream().map(AcademicPeriod::name).collect(Collectors.toList()));

		return "redirect:/" + url + getURLParamsString(params);
	}

	/**
	 * Retracts the open application for a job offer of a user.
	 *
	 * @param  person the authenticated person
	 * @param  id     of the job offer (so not application)
	 * @return        The page the user came from or application/my page.
	 */
	@PostMapping("retract/{id}")
	@PreAuthorize("@authorisationService.applicationExistsFor(#id)")
	public String retractApplication(@AuthenticatedPerson Person person,
			@PathVariable Long id,
			@RequestParam(required = false) String page,
			@RequestParam(required = false) String q,
			@RequestParam(required = false) String program,
			@RequestParam(required = false, name = "period") List<AcademicPeriod> filterPeriods) {
		applicationService.retract(person.getId(), id);

		String url = "applications".equals(page) ? "application/my" : "job-offer/open";

		Map<String, List<String>> params = new HashMap<>();
		if (q != null)
			params.put("q", List.of(q));
		if (program != null)
			params.put("program", List.of(program));
		if (filterPeriods != null)
			params.put("period",
					filterPeriods.stream().map(AcademicPeriod::name).collect(Collectors.toList()));

		return "redirect:/" + url + getURLParamsString(params);
	}

	/**
	 * Rejects the application of the student for that job offer
	 *
	 * @param  personId The student whose application will be rejected
	 * @param  offerId  The job offer for which the application was made
	 * @return          The job offer page (ie. from where the rejection was made)
	 */
	@PostMapping("reject/{personId}/{offerId}")
	@PreAuthorize("@authorisationService.isManagerOfAny()")
	public String reject(@AuthenticatedPerson Person person, @PathVariable Long personId,
			@PathVariable Long offerId) {
		applicationService.reject(personId, offerId, person.getId());
		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Rejects all open (ie. SUBMITTED) applications for a job offer
	 *
	 * @param  offerId The job offer for which to reject the applications
	 * @return         The job offer page (ie. from where the rejections were made)
	 */
	@PostMapping("reject-all-open/{offerId}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#offerId)")
	public String rejectAllOpenInJobOffer(@AuthenticatedPerson Person person,
			@PathVariable Long offerId) {
		applicationService.rejectAllOpenInJobOffer(offerId, person.getId());
		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Offers a position to a person for a job offer
	 *
	 * @param  personId The person to whom to offer the position
	 * @param  offerId  The job offer for which to offer the position
	 * @return          The job offer page (ie. from where the offer was made)
	 */
	@PostMapping("offer/{personId}/{offerId}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#offerId)")
	public String offerApplication(@AuthenticatedPerson Person person,
			@PathVariable Long personId,
			@PathVariable Long offerId) {
		applicationService.offer(personId, offerId, person.getId());
		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Offers a position to a group of people.
	 *
	 * @param  offerId        The ID of the offered position
	 * @param  identifiersAll Identifiers for the people offered to
	 * @return                The job offer page
	 */
	@PostMapping("offer/{offerId}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#offerId)")
	public String bulkOfferApplication(@AuthenticatedPerson Person person,
			@PathVariable Long offerId,
			@RequestParam String identifiersAll) {
		applicationService.bulkOffer(offerId, identifiersAll, person.getId());
		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Rejects a group of people for a position
	 *
	 * @param  offerId        The ID of the position
	 * @param  identifiersAll Identifiers for the people rejected
	 * @return                The job offer page
	 */
	@PostMapping("reject/{offerId}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#offerId)")
	public String bulkRejectApplication(@AuthenticatedPerson Person person, @PathVariable Long offerId,
			@RequestParam String identifiersAll) {
		applicationService.bulkReject(offerId, identifiersAll, person.getId());
		return "redirect:/job-offer/" + offerId;
	}

	/**
	 * Accept a job offer
	 *
	 * @param  person The person accepting the application
	 * @param  id     The id of the job offer for which the application was made
	 * @param  q      Any query parameter that was present when on the previous page (for consistency)
	 * @return        The my applications page
	 */
	@PostMapping("accept/{id}")
	@PreAuthorize("@authorisationService.offerExistsFor(#id)")
	public String acceptApplication(@AuthenticatedPerson Person person,
			@PathVariable Long id,
			@RequestParam(required = false) String page,
			@RequestParam(required = false) String q,
			@RequestParam(required = false) String program,
			@RequestParam(required = false, name = "period") List<AcademicPeriod> filterPeriods) {
		applicationService.accept(person.getId(), id);

		String url = "applications".equals(page) ? "application/my" : "job-offer/open";

		Map<String, List<String>> params = new HashMap<>();
		if (q != null)
			params.put("q", List.of(q));
		if (program != null)
			params.put("program", List.of(program));
		if (filterPeriods != null)
			params.put("period",
					filterPeriods.stream().map(AcademicPeriod::name).collect(Collectors.toList()));

		return "redirect:/" + url + getURLParamsString(params);
	}

	private static String getURLParamsString(Map<String, List<String>> params) {
		StringBuilder result = new StringBuilder();
		char sep = '?';
		for (String param : params.keySet()) {
			for (String value : params.get(param)) {
				result.append(sep);
				sep = '&';
				result.append(param).append("=").append(value);
			}
		}
		return result.toString();
	}

	// endregion

	/**
	 * Exports bulk applications.
	 *
	 * @param  person The authenticated person
	 * @return        The export file as a resource
	 */
	@GetMapping("export")
	@PreAuthorize("@authorisationService.isCoordinatorForProgram(#filter.program)")
	public ResponseEntity<Resource> exportApplications(@AuthenticatedPerson Person person,
			ApplicationFilterDTO filter)
			throws IOException {
		return csvService.getResponse(applicationService.exportApplications("applications",
				applicationService.getFilteredCoordinatingApplications(person.getId(), filter)));
	}

	/**
	 * Exports bulk applications (only Accepted) with FlexDelft specific fields
	 *
	 * @param  person The authenticated person
	 * @return        The export file as a resource
	 */
	@GetMapping("flex-delft-export")
	@PreAuthorize("@authorisationService.isCoordinatorForProgram(#filter.program)")
	public ResponseEntity<Resource> exportApplicationsForFlexDelft(@AuthenticatedPerson Person person,
			ApplicationFilterDTO filter,
			@RequestParam(required = false) @DateTimeFormat(pattern = "yyyy-MM-dd'T'HH:mm:ss") LocalDateTime since,
			@RequestParam Long program,
			@RequestParam Integer batch) throws IOException {
		filter.setStatuses(List.of(Status.ACCEPTED));
		List<Application> appList = applicationService
				.getCoordinatingApplicationsForProgram(person.getId(), program).stream()
				.filter(x -> !editionService.getOrThrow(x.getJobOffer().getEditionId()).getIsArchived())
				.filter(x -> x.getStatus().equals(Status.ACCEPTED))
				.collect(Collectors.toList());
		return csvService
				.getResponse(
						applicationService.exportApplicationsForFlexDelft(person, filter.getProgram(), batch,
								"applications", appList, filter.getUpdatedSince()));
	}

}
