/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import javax.validation.Valid;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;

import nl.tudelft.labracore.api.dto.ProgramDetailsDTO;
import nl.tudelft.labracore.api.dto.ProgramSummaryDTO;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.patch.CoordinatorDefaultPatchDTO;
import nl.tudelft.tam.model.Coordinator;
import nl.tudelft.tam.model.CoordinatorDefault;
import nl.tudelft.tam.model.TrainingType;
import nl.tudelft.tam.service.*;

@Controller
@RequestMapping("coordinator")
public class CoordinatorController {

	@Autowired
	CoordinatorDefaultService coordinatorDefaultService;

	@Autowired
	CoordinatorService coordinatorService;

	@Autowired
	ProgramService programService;

	@Autowired
	PersonService personService;

	@Autowired
	TrainingTypeService trainingTypeService;

	/**
	 * Gets the coordinator programme management page
	 *
	 * @param  person The person who is logged in
	 * @param  q      The search parameter
	 * @param  model  The model to add attributes to
	 * @return        The coordinator programme management page
	 */
	@GetMapping("manage")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public String getCoordinatorManagePage(@AuthenticatedPerson Person person,
			@RequestParam(required = false) String q,
			Model model) {
		List<ProgramDetailsDTO> coordinatingPrograms = programService.getCoordinatingPrograms(person.getId());

		// Filter by query
		if (q != null && !q.isBlank()) {
			coordinatingPrograms = coordinatingPrograms.stream().filter(program -> program
					.getName()
					.toLowerCase()
					.contains(q.toLowerCase())).collect(Collectors.toList());
		}

		model.addAttribute("coordinatingPrograms", coordinatingPrograms);
		model.addAttribute("coordinators",
				coordinatingPrograms.stream()
						.collect(Collectors.toMap(ProgramDetailsDTO::getId,
								p -> personService
										.getPeopleById(coordinatorService.getCoordinatorsByProgram(p.getId())
												.stream().map(Coordinator::getPersonId).toList()))));

		return "coordinator/manage";
	}

	/**
	 * The page where the coordinator can set the defaults for their programme
	 *
	 * @return The defaults page
	 */
	@GetMapping("defaults/{programId}")
	@PreAuthorize("@authorisationService.isCoordinatorForProgram(#programId)")
	public String getDefaultsPage(@PathVariable Long programId,
			Model model) {
		CoordinatorDefault defaultObj = coordinatorDefaultService.getOrCreate(programId);
		CoordinatorDefaultPatchDTO patchDTO = coordinatorDefaultService
				.createPatchFromCoordinatorDefault(defaultObj);

		String programName = programService.get(programId).getName();
		List<Long> advertiseTo = programService.getProgramIdsAdvertisedToBy(programId);
		List<ProgramSummaryDTO> allPrograms = programService.getAllPrograms().stream()
				.filter(p -> !Objects.equals(p.getId(), programId))
				.sorted(Comparator.comparing(ProgramSummaryDTO::getName)).toList();

		model.addAttribute("patchDTO", patchDTO);
		model.addAttribute("programName", programName);
		model.addAttribute("programDefault", defaultObj);
		model.addAttribute("advertised",
				allPrograms.stream().filter(p -> advertiseTo.contains(p.getId())).toList());
		model.addAttribute("notAdvertised",
				allPrograms.stream().filter(p -> !advertiseTo.contains(p.getId())).toList());

		return "coordinator/defaults";
	}

	/**
	 * Returns the page where coordinators can manage the TA training for their programme
	 *
	 * @param  programId The program for which to manage the training
	 * @param  model     The model to add the attributes to
	 * @return           The page where coordinators can manage the TA training for their programme
	 */
	@GetMapping("training/{programId}")
	@PreAuthorize("@authorisationService.isCoordinatorForProgram(#programId)")
	public String getTrainingPage(@PathVariable Long programId,
			Model model) {

		String programName = programService.getProgramsById(List.of(programId)).get(0).getName();

		List<TrainingType> trainingTypes = trainingTypeService
				.getAllTrainingTypesForPrograms(List.of(programId));

		model.addAttribute("programName", programName);
		model.addAttribute("trainingTypes", trainingTypes);
		model.addAttribute("programId", programId);

		return "coordinator/training";
	}

	/**
	 * Modifies the coordinator defaults for a programme
	 *
	 * @param  programId The programme for which to modify the defaults
	 * @param  dto       The DTO containing the new defaults
	 * @return           The redirect to the defaults page that the user originated from
	 */
	@PatchMapping("defaults/modify/{programId}")
	@PreAuthorize("@authorisationService.isCoordinatorForProgram(#programId)")
	public String modifyDefaults(@PathVariable Long programId,
			@Valid @ModelAttribute("patchCoordinatorDefault") CoordinatorDefaultPatchDTO dto) {
		coordinatorDefaultService.patchCoordinatorDefault(programId, dto);

		return "redirect:/coordinator/defaults/" + programId;
	}

}
