/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.io.IOException;
import java.time.LocalDateTime;
import java.util.List;
import java.util.stream.Collectors;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.Resource;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.web.PageableDefault;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;

import nl.tudelft.labracore.api.dto.ProgramDetailsDTO;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.controller.utility.PageUtil;
import nl.tudelft.tam.dto.create.DeclarationCreateDTO;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.Declaration;
import nl.tudelft.tam.service.CSVService;
import nl.tudelft.tam.service.DeclarationService;
import nl.tudelft.tam.service.EditionService;
import nl.tudelft.tam.service.ProgramService;

@Controller
@RequestMapping("declaration")
public class DeclarationController {

	// region class attributes

	@Autowired
	DeclarationService declarationService;

	@Autowired
	CSVService csvService;

	@Autowired
	EditionService editionService;

	@Autowired
	ProgramService programService;

	// endregion

	// region view pages

	/**
	 * Gets the coordinator page.
	 *
	 * @param  person The authenticated person
	 * @param  model  The model to add data to
	 * @return        The coordinator page
	 */
	@GetMapping("all")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public String getAllDeclarations(@AuthenticatedPerson Person person,
			@RequestParam(required = false) Long program,
			@RequestParam(required = false) @DateTimeFormat(pattern = "yyyy-MM-dd'T'HH:mm:ss") LocalDateTime since,
			@PageableDefault(value = 25) Pageable pageable,
			Model model) {
		List<ProgramDetailsDTO> programs = programService.getCoordinatingPrograms(person.getId());
		if (program == null) {
			program = programs.get(0).getId();
		}

		List<Declaration> declarations = declarationService
				.getCoordinatingDeclarationsForProgram(person.getId(), program)
				.stream()
				.filter(x -> !editionService.getOrThrow(x.getExtraWork().getEditionId())
						.getIsArchived())
				.filter(x -> !x.getStatus().equals(Status.REJECTED_BY_STUDENT))
				.toList();
		if (since != null) {
			declarations = declarationService.filterDeclarationsChangedSince(declarations, since);
		}

		// Pagination
		Page<Declaration> pageOfDeclarations = PageUtil.pageFromList(declarations, pageable);

		model.addAttribute("declarations", pageOfDeclarations);
		model.addAttribute("exports", declarationService.getExportsForPersonAndProgram(person, program));
		model.addAttribute("programs", programs);
		return "declaration/all";
	}

	/**
	 * Gets the my declarations page
	 *
	 * @param  q The search query to filter by
	 * @return   A redirect to the open extra work page with the declarations tab open
	 */
	@GetMapping("my")
	@PreAuthorize("!@authorisationService.isStaff()")
	public String getDeclarations(@RequestParam(required = false) String q) {
		return "redirect:/extra-work/open?tab=declarations" + (q == null ? "" : "&q=" + q);
	}

	// endregion

	// region manipulate data

	/**
	 * Creates a declaration for the authenticated person
	 *
	 * @param  person    The person for whom to create the declaration
	 * @param  createDTO The declaration data
	 * @param  page      The previous page the user was on
	 * @return           The my declarations page or the previous page if provided
	 */
	@PostMapping("submit")
	@PreAuthorize("@authorisationService.canDeclareHours(#createDTO.workId)")
	public String submitDeclaration(@AuthenticatedPerson Person person,
			@ModelAttribute DeclarationCreateDTO createDTO,
			@RequestParam(required = false) String page) {
		createDTO.setPersonId(person.getId());
		declarationService.submit(createDTO);
		return "redirect:/" + (page != null ? page : "declaration/my");
	}

	/**
	 * Retracts a declaration
	 *
	 * @param  id   the id of the declaration to retract
	 * @param  page The previous page of the user
	 * @param  q    The search query on the previous page
	 * @return      The my declarations page or the previous page
	 */
	@PostMapping("retract/{id}")
	@PreAuthorize("@authorisationService.ownsDeclaration(#id)")
	public String retractDeclaration(@PathVariable Long id, @RequestParam(required = false) String page,
			@RequestParam(required = false) String q) {
		declarationService.retract(id);
		return "redirect:/" + (page != null ? page : "declaration/my") + (q != null ? "?q=" + q : "");
	}

	/**
	 * Rejects a declaration
	 *
	 * @param  id The declaration to reject
	 * @param  q  The search query on the previous page
	 * @return    The specific extra work page (ie. where the rejection was made)
	 */
	@PostMapping("reject/{id}")
	@PreAuthorize("@authorisationService.isManagerOfAny()")
	public String rejectDeclaration(@AuthenticatedPerson Person person, @PathVariable Long id,
			@RequestParam(required = false) String q) {
		declarationService.reject(id, person.getId());
		Long workId = declarationService.getExtraWorkIdFromDeclaration(id);
		return "redirect:/extra-work/" + workId + (q == null ? "" : "?q=" + q);
	}

	/**
	 * Rejects (ie. Status SUBMITTED) all open declarations for some extra work
	 *
	 * @param  id The extra work for which to reject all open declarations
	 * @return    The specific extra work page (ie. where the rejections were made)
	 */
	@PostMapping("reject-all-open/{id}")
	@PreAuthorize("@authorisationService.isManagerOfWork(#id)")
	public String rejectAllOpenInExtraWork(@AuthenticatedPerson Person person, @PathVariable Long id) {
		declarationService.rejectAllOpenInExtraWork(id, person.getId());
		return "redirect:/extra-work/" + id;
	}

	/**
	 * Approves a declaration
	 *
	 * @param  id The declaration to approve
	 * @param  q  The search query on the previous page
	 * @return    The specific extra work page (ie. where the approval was made)
	 */
	@PostMapping("approve/{id}")
	@PreAuthorize("@authorisationService.isManagerOfAny()")
	public String approveDeclaration(@AuthenticatedPerson Person person, @PathVariable Long id,
			@RequestParam(required = false) String q) {
		declarationService.approve(id, person.getId());
		Long workId = declarationService.getExtraWorkIdFromDeclaration(id);
		return "redirect:/extra-work/" + workId + (q == null ? "" : "?q=" + q);
	}

	// endregion

	/**
	 * Exports bulk declarations.
	 *
	 * @param  person The authenticated person
	 * @return        The export file as a resource
	 */
	@GetMapping("export")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public ResponseEntity<Resource> exportDeclarations(@AuthenticatedPerson Person person)
			throws IOException {
		return csvService.getResponse(declarationService.exportDeclarations("declarations",
				declarationService.getCoordinatingDeclarations(person.getId()).stream()
						.filter(x -> !editionService.getOrThrow(x.getExtraWork().getEditionId())
								.getIsArchived())
						.filter(x -> !x.getStatus().equals(Status.REJECTED_BY_STUDENT))
						.collect(Collectors.toList())));
	}

	/**
	 * Exports bulk declarations (only Accepted) with FlexDelft specific fields
	 *
	 * @param  person The authenticated person
	 * @return        The export file as a resource
	 */
	@GetMapping("flex-delft-export")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public ResponseEntity<Resource> exportDeclarationsForFlexDelft(@AuthenticatedPerson Person person,
			@RequestParam(required = false) @DateTimeFormat(pattern = "yyyy-MM-dd'T'HH:mm:ss") LocalDateTime since,
			@RequestParam Long program, @RequestParam Integer batch)
			throws IOException {
		List<Declaration> decList = declarationService
				.getCoordinatingDeclarationsForProgram(person.getId(), program).stream()
				.filter(x -> !editionService.getOrThrow(x.getExtraWork().getEditionId()).getIsArchived())
				.filter(x -> x.getStatus() == Status.ACCEPTED)
				.collect(Collectors.toList());

		return csvService
				.getResponse(declarationService.exportDeclarationsForFlexDelft(person, program, batch,
						"declarations", decList, since));
	}

}
