/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.io.IOException;
import java.time.LocalDate;
import java.util.*;
import java.util.stream.Collectors;

import javax.validation.Valid;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.InputStreamResource;
import org.springframework.core.io.Resource;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;

import com.opencsv.exceptions.CsvValidationException;

import nl.tudelft.labracore.api.dto.RoleEditionDetailsDTO;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.ExtraWorkCreateDTO;
import nl.tudelft.tam.dto.patch.ExtraWorkPatchDTO;
import nl.tudelft.tam.dto.view.details.DeclarationDetailsExtraWorkDTO;
import nl.tudelft.tam.dto.view.details.DeclarationDetailsPersonDTO;
import nl.tudelft.tam.dto.view.details.EditionDetailsOffersWorkDTO;
import nl.tudelft.tam.dto.view.details.ExtraWorkDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ExtraWorkSummaryDTO;
import nl.tudelft.tam.service.*;

@Controller
@RequestMapping("extra-work")
public class ExtraWorkController {

	// region class attributes

	@Autowired
	ExtraWorkService extraWorkService;

	@Autowired
	DeclarationService declarationService;

	@Autowired
	EditionService editionService;

	@Autowired
	PersonService personService;

	@Autowired
	CSVService csvService;

	// endregion

	// region view pages

	/**
	 * Get the specific extra work details
	 *
	 * @param  id    The id of the extra work
	 * @param  q     The search query
	 * @param  model The model to add the data to
	 * @return       The specific extra work page
	 */
	@GetMapping("{id}")
	@PreAuthorize("@authorisationService.isManagerOfWork(#id)")
	public String getExtraWorkDetails(@PathVariable Long id,
			@RequestParam(required = false) String q, Model model) {
		ExtraWorkDetailsDTO work = extraWorkService.getById(id);

		List<DeclarationDetailsPersonDTO> declarations = declarationService
				.getFilteredDeclarationsForExtraWork(id, q);

		int[] stats = declarationService.getStatsForDecs(declarations);

		model.addAttribute("work", work);
		model.addAttribute("deadlineClosed",
				work.getDeadline() != null && work.getDeadline().isBefore(LocalDate.now()));
		model.addAttribute("declarations", declarations);
		model.addAttribute("stats", stats);
		model.addAttribute("patchExtraWork", new ExtraWorkPatchDTO());

		return "extra_work/view_one";
	}

	/**
	 * Gets all open extra work
	 *
	 * @param  person The currently logged in person
	 * @param  model  The model to add the data to
	 * @param  tab    The current tab
	 * @param  q      The search query
	 * @return        The list of open extra work page (or all declarations, depending on tab)
	 */
	@GetMapping("open")
	public String getExtraWork(@AuthenticatedPerson Person person, Model model,
			@RequestParam(required = false) String tab,
			@RequestParam(required = false) String q) {
		List<DeclarationDetailsExtraWorkDTO> declarations = declarationService
				.getFilteredDeclarationsForPerson(person.getId(), q);

		List<Long> hiredEditionIds = personService.getRolesForPerson(person.getId())
				.stream()
				.filter(x -> Set.of(RoleEditionDetailsDTO.TypeEnum.TA, RoleEditionDetailsDTO.TypeEnum.HEAD_TA)
						.contains(x.getType()))
				.map(x -> Objects.requireNonNull(x.getEdition()).getId())
				.collect(Collectors.toList());

		List<ExtraWorkSummaryDTO> work = extraWorkService.getFilteredOpen(q).stream()
				.map(x -> (ExtraWorkSummaryDTO) x).collect(Collectors.toList());

		List<ExtraWorkSummaryDTO> hiredWork = work.stream()
				.filter(x -> hiredEditionIds.contains(x.getEditionId()))
				.collect(Collectors.toList());

		List<EditionDetailsOffersWorkDTO> editions = editionService
				.combineOffersAndWorkWithEditions(new ArrayList<>(), hiredWork);

		Map<Long, Double> timeRemaining = new HashMap<>();
		for (ExtraWorkSummaryDTO curr : work) {
			timeRemaining.put(curr.getId(),
					declarationService.numberOfHoursRemainingForExtraWork(person.getId(), curr.getId()));
		}

		if (tab == null || !List.of("work", "declarations").contains(tab)) {
			tab = "work";
		}

		model.addAttribute("tab", tab);
		model.addAttribute("editions", editions);
		model.addAttribute("declarations", declarations);
		model.addAttribute("timeRemaining", timeRemaining);

		return "extra_work/view_many";
	}

	// endregion

	// region manipulate data

	/**
	 * Creates new extra work
	 *
	 * @param  dto The extra work data
	 * @return     The specific extra work page
	 */
	@PostMapping
	@PreAuthorize("@authorisationService.isManagerOfAny()")
	public String createExtraWork(
			@Valid @ModelAttribute("extrawork") @DateTimeFormat(pattern = "yyyy-MM-dd") ExtraWorkCreateDTO dto) {
		Long id = extraWorkService.newExtraWork(dto);
		return "redirect:/extra-work/" + id;
	}

	/**
	 * Edits some existing extra work
	 *
	 * @param  id  The id of the extra work
	 * @param  dto The data
	 * @return     The specific extra work page
	 */
	@PatchMapping("{id}")
	@PreAuthorize("@authorisationService.isManagerOfWork(#id)")
	public String patchExtraWork(@PathVariable Long id,
			@Valid @ModelAttribute("patchExtraWork") ExtraWorkPatchDTO dto) {
		extraWorkService.patchExtraWork(id, dto);
		return "redirect:/extra-work/" + id;
	}

	/**
	 * Closes (ie. sets deadline in the past) some extra work
	 *
	 * @param  id The id of the extra work
	 * @return    The specific extra work page
	 */
	@PostMapping("close/{id}")
	@PreAuthorize("@authorisationService.isManagerOfWork(#id)")
	public String closeExtraWork(@PathVariable Long id) {
		extraWorkService.closeExtraWork(id);
		return "redirect:/extra-work/" + id;
	}

	/**
	 * Rejects all open (ie. Status SUBMITTED) declarations for some extra work
	 *
	 * @param  id The extra work for which to reject all declarations
	 * @return    The specific extra work page
	 */
	@PostMapping("reject-all-open/{id}")
	@PreAuthorize("@authorisationService.isManagerOfWork(#id)")
	public String rejectAllOpenInExtraWork(@AuthenticatedPerson Person person, @PathVariable Long id) {
		declarationService.rejectAllOpenInExtraWork(id, person.getId());
		return "redirect:/extra-work/" + id;
	}

	/**
	 * Approves all open (ie. Status SUBMITTED) declarations for some extra work
	 *
	 * @param  id The id of the extra work for which to approve all declarations
	 * @return    The specific extra work page
	 */
	@PostMapping("approve-all-open/{id}")
	@PreAuthorize("@authorisationService.isManagerOfWork(#id)")
	public String approveAllOpenInExtraWork(@AuthenticatedPerson Person person, @PathVariable Long id) {
		declarationService.approveAllOpenInExtraWork(id, person.getId());
		return "redirect:/extra-work/" + id;
	}

	// endregion

	// region import/export

	/**
	 * Exports all declarations for some extra work
	 *
	 * @param  id The id of the extra work
	 * @return    The export file as a resource
	 */
	@GetMapping("{id}/export")
	@PreAuthorize("@authorisationService.isManagerOfWork(#id)")
	public ResponseEntity<Resource> exportWork(@PathVariable Long id) throws IOException {
		return csvService.getResponse(extraWorkService.exportExtraWork(id));
	}

	/**
	 * Imports a list of extra work from a CSV file.
	 *
	 * @param  file                   The csv file containing the extra work.
	 * @throws IOException            If the file is invalid/corrupted.
	 * @throws CsvValidationException If the format of the file is invalid/corrupted.
	 */
	@PostMapping("import")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public String uploadExtraWork(@RequestParam("file") MultipartFile file)
			throws CsvValidationException, IOException {
		extraWorkService.importExtraWork(file);
		return "redirect:/job-offer/manage";
	}

	@GetMapping("/import/template")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public ResponseEntity<Resource> downloadTemplateFile() {
		Resource resource = new InputStreamResource(
				getClass().getResourceAsStream("/static/import_templates/import_extra_work_template.csv"));

		return ResponseEntity.ok()
				.contentType(MediaType.parseMediaType(MediaType.APPLICATION_OCTET_STREAM_VALUE))
				.header(HttpHeaders.CONTENT_DISPOSITION,
						"attachment; filename=import_extra_work_template.csv")
				.body(resource);
	}

	// endregion

}
