/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.io.IOException;
import java.time.LocalDate;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.validation.Valid;

import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.InputStreamResource;
import org.springframework.core.io.Resource;
import org.springframework.format.annotation.DateTimeFormat;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import org.springframework.web.servlet.mvc.support.RedirectAttributes;

import com.opencsv.exceptions.CsvValidationException;

import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.JobOfferCreateDTO;
import nl.tudelft.tam.dto.create.MissingEditionsJobOfferImportDTO;
import nl.tudelft.tam.dto.patch.JobOfferMessagePatchDTO;
import nl.tudelft.tam.dto.patch.JobOfferPatchDTO;
import nl.tudelft.tam.dto.view.details.ApplicationDetailsJobOfferDTO;
import nl.tudelft.tam.dto.view.details.ApplicationDetailsPersonDTO;
import nl.tudelft.tam.dto.view.details.EditionDetailsOffersWorkDTO;
import nl.tudelft.tam.dto.view.details.JobOfferDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ApplicationStatusSummaryDTO;
import nl.tudelft.tam.dto.view.summary.ApplicationSummaryDTO;
import nl.tudelft.tam.dto.view.summary.JobOfferSummaryDTO;
import nl.tudelft.tam.enums.AcademicPeriod;
import nl.tudelft.tam.io.ImportError;
import nl.tudelft.tam.io.JobOfferImportHandler;
import nl.tudelft.tam.security.AuthorisationService;
import nl.tudelft.tam.service.*;

@Controller
@RequestMapping("job-offer")
public class JobOfferController {

	// region class attributes

	@Autowired
	private JobOfferService jobOfferService;

	@Autowired
	private ApplicationService applicationService;

	@Autowired
	private EditionService editionService;

	@Autowired
	private CourseService courseService;

	@Autowired
	private CohortService cohortService;

	@Autowired
	private CSVService csvService;

	@Autowired
	private AuthorisationService authorisationService;

	@Autowired
	private CoordinatorService coordinatorService;

	@Autowired
	private CoordinatorDefaultService coordinatorDefaultService;

	@Autowired
	private ProgramService programService;

	@Autowired
	private TrainingTypeService trainingTypeService;

	@Autowired
	private ModelMapper mapper;

	// endregion

	// region view pages

	/**
	 * Redirects users to the correct starting page based on their roles.
	 *
	 * @return The redirected page (either open or manage)
	 */
	@GetMapping
	public String jobOffersRedirect() {
		if (authorisationService.isStaff()) {
			// Redirect to All Applications
			return "redirect:/job-offer/manage";
		} else {
			// Redirect to Job Offers
			return "redirect:/job-offer/open";
		}
	}

	/**
	 * Gets all open job offers
	 *
	 * @param  person The currently logged in person
	 * @param  q      The search query
	 * @param  model  The model to add the data to
	 * @return        The job offer page with all open job offers
	 */
	@GetMapping("open")
	public String getOpenJobOffers(@AuthenticatedPerson Person person,
			@RequestParam(required = false) Long program,
			@RequestParam(required = false, name = "period") List<AcademicPeriod> filterPeriods,
			@RequestParam(required = false) String q,
			Model model) {
		List<ApplicationDetailsJobOfferDTO> applications = applicationService
				.getFilteredApplicationsForPerson(person.getId(), q);
		List<JobOfferDetailsDTO> offerDetails = jobOfferService.getFilteredOpen(q);
		List<ProgramSummaryDTO> programsWithJobOffers = courseService.getCoursesById(
				offerDetails.stream().map(o -> o.getEdition().getCourse().getId()).distinct().toList())
				.stream().map(CourseDetailsDTO::getProgram).distinct().toList();

		List<EditionDetailsOffersWorkDTO> editions = new ArrayList<>();
		List<ProgramSummaryDTO> relevantProgrammes = programService.getAllRelevantPrograms(person.getId());
		List<ProgramDetailsDTO> programs = programService.getProgramsById(program == null ? Stream.concat(
				relevantProgrammes.stream().map(ProgramSummaryDTO::getId),
				relevantProgrammes.stream()
						.flatMap(p -> programService.getProgramIdsAdvertisingTo(p.getId()).stream()))
				.distinct().toList() : List.of(program));

		for (ProgramDetailsDTO programDTO : programs) {
			Set<Long> coursesInProgram = programDTO.getCourses().stream().map(CourseSummaryDTO::getId)
					.collect(Collectors.toSet());
			List<JobOfferSummaryDTO> offers = offerDetails.stream()
					.filter(o -> coursesInProgram.contains(o.getEdition().getCourse().getId()))
					.filter(o -> filterPeriods == null || filterPeriods.contains(o.getPeriod()))
					.map(x -> (JobOfferSummaryDTO) x).collect(Collectors.toList());

			// Sort editions based on Cohort, exclude masters and add them to the end.
			Comparator<EditionDetailsOffersWorkDTO> editionComparator = Comparator
					.<EditionDetailsOffersWorkDTO, String>comparing(o -> o.getCohort().getName())
					.reversed().thenComparing(EditionDetailsDTO::getStartDate)
					.thenComparing(EditionDetailsDTO::getName);
			editions.addAll(editionService
					.combineOffersAndWorkWithEditions(offers, new ArrayList<>())
					.stream().sorted(editionComparator).toList());
		}

		Map<Long, ApplicationStatusSummaryDTO> appForOffer = applicationService
				.getMapWithStatusAppOpenByIdForPerson(
						person.getId(),
						offerDetails.stream().map(JobOfferDetailsDTO::getId).collect(Collectors.toList()));

		model.addAttribute("applications", applications);
		model.addAttribute("programs", programsWithJobOffers.stream()
				.sorted(Comparator.comparing(ProgramSummaryDTO::getName)).toList());
		model.addAttribute("editions", editions);
		model.addAttribute("userApplications", appForOffer);

		return "job_offer/view_many";
	}

	/**
	 * Gets all editions a coordinator is a manager of.
	 *
	 * @param  person The authenticated person
	 * @param  q      The search query
	 * @param  model  The model to add details to
	 * @return        The manager page with managed editions
	 */
	@GetMapping("manage")
	@PreAuthorize("@authorisationService.isManagerOfAny()")
	public String getManagerPortal(@AuthenticatedPerson Person person,
			@RequestParam(required = false) String q,
			Model model) {

		List<Long> managedEditionIds = editionService.getAllEditionIdsThatPersonIsAMangerOf(person.getId());
		List<CourseSummaryDTO> managedCourses = courseService
				.getManagingCourses(new PersonIdDTO().id(person.getId()));
		List<Long> programIdsOfManagedCourses = managedCourses.isEmpty() ? Collections.emptyList()
				: courseService.getCoursesById(managedCourses.stream()
						.map(CourseSummaryDTO::getId)
						.collect(Collectors.toList())).stream()
						.map(x -> x.getProgram().getId())
						.collect(Collectors.toList());

		// Filter by the search query
		List<EditionDetailsOffersWorkDTO> editions = editionService
				.filter(editionService.getEditionsWithJobsAndWork(managedEditionIds), q)
				.stream().sorted(Comparator.comparing(EditionDetailsOffersWorkDTO::getStartDate).reversed())
				.toList();

		// Add the coordinator defaults to the edition objects
		coordinatorDefaultService.addDefaultsToEditionsWithJobsAndWork(editions);

		model.addAttribute("editions", editions);
		model.addAttribute("showAll", false);
		model.addAttribute("managed", managedCourses);
		model.addAttribute("cohorts", cohortService.getCohortsByProgrammeIds(programIdsOfManagedCourses));

		return "job_offer/manage";
	}

	/**
	 * Gets all editions a coordinator can manage.
	 *
	 * @param  person The authenticated person
	 * @param  q      The search query
	 * @param  model  The model to add details to
	 * @return        The manager page with all editions
	 */
	@GetMapping("manage/all")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public String getManageAllCourses(@AuthenticatedPerson Person person,
			@RequestParam(required = false) String q,
			Model model) {

		Set<Long> coordinatingProgramIds = coordinatorService.getProgramIdsByPersonId(person.getId());
		List<CourseSummaryDTO> coordinatingCourses = new ArrayList<>();
		List<Long> coordinatingEditions = new ArrayList<>();

		for (Long id : coordinatingProgramIds) {
			List<CourseSummaryDTO> coursesInProgram = courseService.getAllCoursesInProgram(id);
			coordinatingCourses.addAll(coursesInProgram);
		}

		for (CourseSummaryDTO course : coordinatingCourses) {
			coordinatingEditions.addAll(editionService.getEditionsForCourse(course.getId()));
		}

		List<CohortSummaryDTO> coordinatingCohorts = cohortService
				.getCohortsByProgrammeIds(coordinatingProgramIds.stream().toList());

		List<EditionDetailsOffersWorkDTO> editions = editionService
				.getEditionsWithJobsAndWork(coordinatingEditions)
				.stream().sorted(Comparator.comparing(EditionDetailsOffersWorkDTO::getStartDate).reversed())
				.toList();

		// Filter by the search query
		editions = editionService.filter(editions, q);

		// Add the coordinator defaults to the edition objects
		coordinatorDefaultService.addDefaultsToEditionsWithJobsAndWork(editions);

		model.addAttribute("editions", editions);
		model.addAttribute("showAll", true);
		model.addAttribute("managed", coordinatingCourses);
		model.addAttribute("cohorts", coordinatingCohorts);

		return "job_offer/manage";
	}

	/**
	 * Gets the details of a specific job offer
	 *
	 * @param  id    The id of the job offer
	 * @param  q     The search query
	 * @param  model The model to add the data to
	 * @return       The specific job offer page
	 */
	@GetMapping("{id}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#id)")
	public String getJobOfferDetails(@PathVariable Long id,
			@RequestParam(required = false) String q, Model model) {
		JobOfferDetailsDTO offer = jobOfferService.getById(id);
		Long programId = courseService.getCourseById(offer.getEdition().getCourse().getId()).getProgram()
				.getId();

		List<ApplicationDetailsPersonDTO> applications = applicationService
				.getFilteredApplicationsForJobOffer(id, q);

		int[] stats = applicationService.getStatsForApps(applications);

		model.addAttribute("offer", offer);
		model.addAttribute("deadlineClosed",
				offer.getDeadline() != null && offer.getDeadline().isBefore(LocalDate.now()));
		model.addAttribute("applications", applications.stream()
				.sorted(Comparator.comparing(ApplicationSummaryDTO::getCreatedDate)).toList());
		model.addAttribute("stats", stats);
		model.addAttribute("trainingTypes",
				trainingTypeService.getAllTrainingTypesForPrograms(List.of(programId)));

		return "job_offer/view_one";
	}

	// endregion

	// region manipulate data

	/**
	 * Creates a new job offer
	 *
	 * @param  dto The job offer data
	 * @return     The specific job offer page
	 */
	@PostMapping
	@PreAuthorize("@authorisationService.isManagerOfEdition(#dto.editionId)")
	public String createJobOffer(
			@Valid @ModelAttribute("joboffer") @DateTimeFormat(pattern = "yyyy-MM-dd") JobOfferCreateDTO dto) {
		Long id = jobOfferService.newJobOffer(dto);
		return "redirect:/job-offer/" + id;
	}

	/**
	 * Edits an existing job offer
	 *
	 * @param  id  The id of the job offer
	 * @param  dto The data of the job offer
	 * @return     The specific job offer page
	 */
	@PatchMapping("{id}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#id)")
	public String patchJobOffer(@PathVariable Long id,
			@Valid @ModelAttribute("patchJobOffer") JobOfferPatchDTO dto) {
		jobOfferService.patchJobOffer(id, dto);
		return "redirect:/job-offer/" + id;
	}

	/**
	 * Edits an existing job offer's message
	 *
	 * @param  id  The id of the job offer
	 * @param  dto The new message data of the job offer
	 * @return     The specific job offer page
	 */
	@PatchMapping("{id}/messages")
	@PreAuthorize("@authorisationService.isManagerOfJob(#id)")
	public String patchJobOffer(@PathVariable Long id,
			@Valid JobOfferMessagePatchDTO dto) {
		jobOfferService.patchJobOffer(id, dto);
		return "redirect:/job-offer/" + id;
	}

	/**
	 * Closes a job offer (ie. sets deadline to past)
	 *
	 * @param  id The id of the job offer
	 * @return    The specific job offer page
	 */
	@PostMapping("close/{id}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#id)")
	public String closeJobOffer(@PathVariable Long id) {
		jobOfferService.closeJobOffer(id);
		return "redirect:/job-offer/" + id;
	}

	/**
	 * Adds a required training to a job offer.
	 *
	 * @param id     The id of the job offer
	 * @param typeId The id of the training
	 */
	@PostMapping("{id}/required-training/{typeId}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#id)")
	public @ResponseBody void addRequiredTraining(@PathVariable Long id, @PathVariable Long typeId) {
		jobOfferService.addRequiredTraining(id, typeId);
	}

	/**
	 * Removes a required training to a job offer.
	 *
	 * @param id     The id of the job offer
	 * @param typeId The id of the training type
	 */
	@DeleteMapping("{id}/required-training/{typeId}")
	@PreAuthorize("@authorisationService.isManagerOfJob(#id)")
	public @ResponseBody void removeRequiredTraining(@PathVariable Long id, @PathVariable Long typeId) {
		jobOfferService.removeRequiredTraining(id, typeId);
	}

	// endregion

	// region import/export

	/**
	 * Exports all applications of a job offer.
	 *
	 * @param  id The id of the job offer
	 * @return    The export file as a resource
	 */
	@GetMapping("{id}/export")
	@PreAuthorize("@authorisationService.isManagerOfJob(#id)")
	public ResponseEntity<Resource> exportJob(@PathVariable Long id) throws IOException {
		return csvService.getResponse(jobOfferService.exportJobOffer(id));
	}

	/**
	 * Imports a list of job offers from a CSV file.
	 *
	 * @param  file                   The csv file containing the Job Offers
	 * @throws IOException            If the file is invalid/corrupted.
	 * @throws CsvValidationException If the file formatting is invalid/corrupted.
	 */
	@PostMapping("import")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public String uploadJobOffers(@RequestParam("file") MultipartFile file, RedirectAttributes attributes)
			throws IOException, CsvValidationException {
		var result = jobOfferService.importJobOffers(file);

		Map<EditionCreateDTO, List<JobOfferImportHandler.EditionDoesNotExistError>> editionErrors = result
				.errors().values().stream().flatMap(Collection::stream)
				.filter(e -> e instanceof JobOfferImportHandler.EditionDoesNotExistError)
				.map(e -> ((JobOfferImportHandler.EditionDoesNotExistError) e))
				.collect(Collectors.groupingBy(JobOfferImportHandler.EditionDoesNotExistError::edition));
		Map<Integer, List<ImportError>> errors = result.errors().entrySet().stream()
				.map(e -> new AbstractMap.SimpleEntry<>(e.getKey(),
						e.getValue().stream().filter(
								err -> !(err instanceof JobOfferImportHandler.EditionDoesNotExistError))
								.toList()))
				.filter(e -> !e.getValue().isEmpty())
				.collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue));

		attributes.addFlashAttribute("successfulImportSize", result.result().size());
		attributes.addFlashAttribute("importWarnings", result.warnings());
		attributes.addFlashAttribute("importErrors", errors);
		attributes.addFlashAttribute("editionErrors", editionErrors);

		MissingEditionsJobOfferImportDTO dto = new MissingEditionsJobOfferImportDTO();
		editionErrors.forEach((e, os) -> {
			dto.getEditions().add(e);
			dto.getJobOffers().addAll(
					os.stream().map(JobOfferImportHandler.EditionDoesNotExistError::jobOffer).toList());
		});
		for (int i = 0; i < dto.getEditions().size(); i++) {
			long id = i;
			editionErrors.get(dto.getEditions().get(i)).forEach(e -> e.jobOffer().setEditionId(id));
		}
		attributes.addFlashAttribute("dto", dto);

		return "redirect:/job-offer/manage";
	}

	/**
	 * Creates the job offers for editions that did not exist when importing a list of job offers.
	 *
	 * @param  dto The DTO with to be created editions and job offers
	 * @return     Redirect back to the manage page
	 */
	@PostMapping("import-missing")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public String importJobOffersWithMissingEditions(MissingEditionsJobOfferImportDTO dto) {
		jobOfferService.createJobOffersWithMissingEditions(dto);
		return "redirect:/job-offer/manage";
	}

	@GetMapping("/import/template")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public ResponseEntity<Resource> downloadTemplateFile() {
		Resource resource = new InputStreamResource(
				getClass().getResourceAsStream("/static/import_templates/import_job_offers_template.csv"));

		return ResponseEntity.ok()
				.contentType(MediaType.parseMediaType(MediaType.APPLICATION_OCTET_STREAM_VALUE))
				.header(HttpHeaders.CONTENT_DISPOSITION,
						"attachment; filename=import_job_offers_template.csv")
				.body(resource);
	}

	/**
	 * Toggles the hiding of a job offer.
	 *
	 * @param  id The id of the job offer.
	 * @return    The page to load
	 */
	@PostMapping("{id}/toggle-hide")
	@PreAuthorize("@authorisationService.isManagerOfAny()")
	public String toggleHide(@PathVariable Long id) {
		jobOfferService.toggleHide(id);
		return "redirect:/job-offer/" + id;
	}
	// endregion

}
