/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.util.Arrays;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.*;

import nl.tudelft.tam.service.CoordinatorService;
import nl.tudelft.tam.service.ProgramService;

@Controller
@RequestMapping("program")
public class ProgramController {

	private final ProgramService programService;
	private final CoordinatorService coordinatorService;

	@Autowired
	public ProgramController(ProgramService programService, CoordinatorService coordinatorService) {
		this.programService = programService;
		this.coordinatorService = coordinatorService;
	}

	/**
	 * Sets the coordinators of the programme to a new lists of coordinators.
	 *
	 * @return Redirect to the manage page
	 */
	@PatchMapping("{id}/coordinators")
	@PreAuthorize("@authorisationService.canEditCoordinatorsForProgram(#id)")
	public String updateCoordinators(@PathVariable Long id, String coordinators) {
		coordinatorService.setCoordinatorsForProgram(id,
				Arrays.stream(coordinators.split("\n")).map(String::trim).filter(s -> !s.isBlank()).toList());
		return "redirect:/coordinator/manage";
	}

	/**
	 * Adds a programme to the advertised to programmes of another programme.
	 *
	 * @param  advertiseeId   The programme advertised
	 * @param  advertisedToId The programme advertised to
	 * @return                200 OK
	 */
	@PostMapping("{advertiseeId}/advertise/{advertisedToId}")
	@PreAuthorize("@authorisationService.isCoordinatorForProgram(#advertiseeId)")
	public ResponseEntity<Void> advertiseToProgram(@PathVariable Long advertiseeId,
			@PathVariable Long advertisedToId) {
		programService.addAdvertisedToProgram(advertiseeId, advertisedToId);
		return ResponseEntity.ok().build();
	}

	/**
	 * Removed a programme from the advertised to programmes of another programme.
	 *
	 * @param  advertiseeId   The programme advertised
	 * @param  advertisedToId The programme advertised to
	 * @return                200 OK
	 */
	@DeleteMapping("{advertiseeId}/advertise/{advertisedToId}")
	public ResponseEntity<Void> removeAdvertisementFromProgram(@PathVariable Long advertiseeId,
			@PathVariable Long advertisedToId) {
		programService.removeAdvertisedFromProgram(advertiseeId, advertisedToId);
		return ResponseEntity.ok().build();
	}

}
