/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.io.IOException;
import java.util.AbstractMap;
import java.util.ArrayList;
import java.util.List;

import javax.transaction.Transactional;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.Resource;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.data.web.PageableDefault;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartHttpServletRequest;
import org.springframework.web.servlet.mvc.support.RedirectAttributes;

import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.controller.utility.PageUtil;
import nl.tudelft.tam.dto.create.RaiseRequestCreateDTO;
import nl.tudelft.tam.dto.id.ProfileIdDTO;
import nl.tudelft.tam.dto.view.details.RaiseRequestDetailsDTO;
import nl.tudelft.tam.exception.FileHandlingException;
import nl.tudelft.tam.exception.InvalidDatabaseEntryException;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.service.CoordinatorService;
import nl.tudelft.tam.service.RaiseRequestService;

@Controller
@RequestMapping("raise-request")
public class RaiseRequestController {

	@Autowired
	private RaiseRequestService raiseRequestService;

	@Autowired
	private CoordinatorService coordinatorService;

	/**
	 * Shows a list of open raise requests.
	 *
	 * @param  model                         the model
	 * @return                               link to the page with the list of open raise requests
	 * @throws InvalidDatabaseEntryException if an invalid entry is found
	 */
	@GetMapping("all")
	@PreAuthorize("@authorisationService.canEditPayScale()")
	public String getSubmittedRaiseRequests(@AuthenticatedPerson Person person,
			@RequestParam(required = false, defaultValue = "0") Integer page,
			@PageableDefault(value = 25) Pageable pageable,
			Model model) throws InvalidDatabaseEntryException {
		List<Long> coordinatingProgramIds = new ArrayList<>(
				coordinatorService.getProgramIdsByPersonId(person.getId()));

		List<RaiseRequestDetailsDTO> raiseRequests = raiseRequestService
				.getAllSubmitted(coordinatingProgramIds);

		// Pagination
		Page<RaiseRequestDetailsDTO> pageOfRaiseRequests = PageUtil.pageFromList(raiseRequests, pageable);

		model.addAttribute("raiseRequests", pageOfRaiseRequests);

		return "raise_request/submitted";
	}

	/**
	 * Submits a new claim to change the PayScale of a person.
	 *
	 * @param  person                the person requesting the change
	 * @param  raiseRequestCreateDTO the PayScale change object
	 * @return                       redirects to profile page
	 */
	@Transactional
	@PostMapping(value = "submit", consumes = { "multipart/form-data" })
	public String postChange(@AuthenticatedPerson Person person,
			@ModelAttribute("createRaiseRequestDTO") RaiseRequestCreateDTO raiseRequestCreateDTO,
			MultipartHttpServletRequest request) throws FileHandlingException {
		raiseRequestCreateDTO.setRequestedBy(new ProfileIdDTO(person.getId()));

		raiseRequestService.newRaiseRequest(raiseRequestCreateDTO,
				raiseRequestService.validateFileSubmission(request.getFileMap()));

		return "redirect:/profile";
	}

	/**
	 * Retracts the last pending claim to change the PayScale of a person.
	 *
	 * @param  person the person requesting the change
	 * @return        redirects to profile page
	 */
	@Transactional
	@PostMapping("retract")
	public String retractChange(@AuthenticatedPerson Person person) {

		raiseRequestService.retract(person);

		return "redirect:/profile";
	}

	/**
	 * Approves the claim to change the PayScale of a person.
	 *
	 * @param  person the person approving the claim
	 * @return        redirects to submitted raise requests page
	 */
	@Transactional
	@PostMapping("approve/{id}")
	@PreAuthorize("@authorisationService.canEditPayScale()")
	public String approveChange(@AuthenticatedPerson Person person, @PathVariable Long id) {
		raiseRequestService.approve(id, person);

		// TODO notify student

		return "redirect:/raise-request/all";
	}

	/**
	 * Rejects the claim to change the PayScale of a person.
	 *
	 * @param  person the person approving the claim
	 * @return        redirects to submitted raise requests page
	 */
	@Transactional
	@PostMapping("reject/{id}")
	@PreAuthorize("@authorisationService.canEditPayScale()")
	public String rejectChange(@AuthenticatedPerson Person person, @PathVariable Long id) {
		raiseRequestService.reject(id, person);

		// TODO notify student

		return "redirect:/raise-request/all";
	}

	/**
	 * Downloads the submission from the local file system.
	 *
	 * @param  id the identifier of the submission associated with the file
	 * @return    the ResponseEntity called
	 */
	@GetMapping("/{id}/proof")
	@PreAuthorize("@authorisationService.canDownloadProof(#id)")
	public ResponseEntity<Resource> downloadProofFile(@PathVariable Long id) throws IOException {
		Resource resource = raiseRequestService.downloadFile(id);

		return ResponseEntity.ok()
				.contentType(MediaType.parseMediaType(MediaType.APPLICATION_OCTET_STREAM_VALUE))
				.header(HttpHeaders.CONTENT_DISPOSITION,
						"attachment; filename=\"" + resource.getFilename() + "\"")
				.body(resource);
	}

	@GetMapping(value = "{id}/proof.pdf", produces = MediaType.APPLICATION_PDF_VALUE)
	@PreAuthorize("@authorisationService.canDownloadProof(#id)")
	public ResponseEntity<Resource> getProofFile(@PathVariable Long id) throws IOException {
		Resource resource = raiseRequestService.downloadFile(id);
		return ResponseEntity.ok()
				.header("X-Frame-Options", "SAMEORIGIN")
				.contentType(MediaType.APPLICATION_PDF)
				.body(resource);
	}

	/**
	 * Import PayScale for a student
	 *
	 * @param  person     The person rejecting the existing requests
	 * @param  identifier The person for whom the payscale will be changed
	 * @param  newValue   The new PayScale
	 * @param  overwrite  If existing approvals should be ignored
	 * @param  rAttrs     Redirect attributes to pass on necessary data
	 * @return            Redirects to the list of raise requests page
	 */
	@Transactional
	@PostMapping("import/{identifier}/{newValue}/{overwrite}")
	@PreAuthorize("@authorisationService.canEditPayScale()")
	public String importPayScale(@AuthenticatedPerson Person person,
			@PathVariable String identifier,
			@PathVariable String newValue,
			@PathVariable String overwrite,
			RedirectAttributes rAttrs) {
		List<AbstractMap.SimpleEntry<PersonSummaryDTO, Profile>> conflictList = raiseRequestService
				.importPayScale(identifier, newValue, person, overwrite.equals("overwrite"));

		rAttrs.addFlashAttribute("conflictsList", conflictList);
		rAttrs.addFlashAttribute("newValue", newValue);

		return "redirect:/raise-request/all";
	}

	/**
	 * Import PayScales for multiple students
	 *
	 * @param  person         The person rejecting the existing requests
	 * @param  identifiersAll The people for whom the PayScales will be changed
	 * @param  newValueAll    The new PayScale
	 * @param  rAttrs         Redirect attributes to pass on necessary data
	 * @return                Redirects to the list of raise requests page
	 */
	@Transactional
	@PostMapping("bulk-import")
	@PreAuthorize("@authorisationService.canEditPayScale()")
	public String bulkImportPayScales(@AuthenticatedPerson Person person,
			@RequestParam String identifiersAll,
			@RequestParam String newValueAll,
			RedirectAttributes rAttrs) {
		List<AbstractMap.SimpleEntry<PersonSummaryDTO, Profile>> conflictList = raiseRequestService
				.bulkImportPayScales(identifiersAll, newValueAll, person);

		rAttrs.addFlashAttribute("conflictsList", conflictList);
		rAttrs.addFlashAttribute("newValue", newValueAll);

		return "redirect:/raise-request/all";
	}

}
