/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import java.io.IOException;
import java.util.Map;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.InputStreamResource;
import org.springframework.core.io.Resource;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.stereotype.Controller;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.multipart.MultipartFile;
import org.springframework.web.servlet.mvc.support.RedirectAttributes;

import com.opencsv.exceptions.CsvValidationException;

import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.lib.security.user.AuthenticatedPerson;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.service.PersonService;
import nl.tudelft.tam.service.TrainingApprovalService;

@Controller
@RequestMapping("training-approval")
public class TrainingApprovalController {

	@Autowired
	private TrainingApprovalService trainingApprovalService;

	@Autowired
	private PersonService personService;

	/**
	 * Create a new training approval for one person
	 *
	 * @param  approver     The person creating the approval
	 * @param  searchQuery  The query used to search for the trainees to allow for smooth continued imports
	 * @param  trainee      The person for whom the approval is being created
	 * @param  trainingType The type of training to create the approval for
	 * @param  datePassed   The date the training was passed
	 * @param  rAttrs       Redirect attributes to pass on necessary data
	 * @return              The coordinator manage page
	 */
	@PostMapping("import/{programId}/{trainee}/{trainingType}/{datePassed}")
	@PreAuthorize("@authorisationService.canEditTaTrainingForProgram(#programId)")
	public String importTraining(@AuthenticatedPerson Person approver,
			@PathVariable Long programId,
			@RequestParam String searchQuery,
			@PathVariable String trainee,
			@PathVariable String trainingType,
			@PathVariable String datePassed,
			RedirectAttributes rAttrs) {

		trainingApprovalService.importTraining(approver, trainee, trainingType, datePassed);

		// Allow for smooth continuation of the importing process
		if (!searchQuery.isBlank() && (searchQuery.split("\\n").length > 1)) {
			Map<PersonSummaryDTO, Profile> foundPeople = personService
					.findPersonProfilePairs(searchQuery);

			rAttrs.addFlashAttribute("foundPeople", foundPeople);
			rAttrs.addFlashAttribute("searchQuery", searchQuery);
			rAttrs.addFlashAttribute("trainingType", trainingType);
			rAttrs.addFlashAttribute("datePassed", datePassed);
		}

		rAttrs.addFlashAttribute("trainingImportSuccess", true);

		return "redirect:/coordinator/training/" + programId;
	}

	/**
	 * Create a new training approval for multiple people
	 *
	 * @param  approver     The person creating the approval
	 * @param  allTrainees  The list of people for whom to create approvals
	 * @param  trainingType The type of training to create the approvals for
	 * @param  datePassed   The date the trainings were passed
	 * @param  rAttrs       Redirect attributes to pass on necessary data
	 * @return              The coordinator manage page
	 */
	@PostMapping("import/{programId}/bulk")
	@PreAuthorize("@authorisationService.canEditTaTrainingForProgram(#programId)")
	public String bulkImportTraining(@AuthenticatedPerson Person approver,
			@PathVariable Long programId,
			@RequestParam String allTrainees,
			@RequestParam String trainingType,
			@RequestParam String datePassed,
			RedirectAttributes rAttrs) {

		trainingApprovalService.bulkImportTraining(approver, allTrainees, trainingType, datePassed);

		rAttrs.addFlashAttribute("trainingImportSuccess", true);

		return "redirect:/coordinator/training/" + programId;
	}

	@GetMapping("/import/template")
	@PreAuthorize("@authorisationService.isCoordinatorOfAny()")
	public ResponseEntity<Resource> downloadTemplateFile() {
		Resource resource = new InputStreamResource(
				getClass().getResourceAsStream("/static/import_templates/import_training_template.csv"));

		return ResponseEntity.ok()
				.contentType(MediaType.parseMediaType(MediaType.APPLICATION_OCTET_STREAM_VALUE))
				.header(HttpHeaders.CONTENT_DISPOSITION,
						"attachment; filename=import_training_template.csv")
				.body(resource);
	}

	@PostMapping("import/{programId}/csv")
	@PreAuthorize("@authorisationService.canEditTaTrainingForProgram(#programId)")
	public String importTrainingCsv(@RequestParam("file") MultipartFile file,
			@PathVariable Long programId,
			RedirectAttributes rAttrs) throws IOException, CsvValidationException {

		var result = trainingApprovalService.importTraining(programId, file);

		rAttrs.addFlashAttribute("successfulImportSize", result.result().size());
		rAttrs.addFlashAttribute("importWarnings", result.warnings());
		rAttrs.addFlashAttribute("importErrors", result.errors());

		return "redirect:/coordinator/training/" + programId;
	}

}
