/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.enums;

import java.time.LocalDate;
import java.time.Month;
import java.time.MonthDay;

import nl.tudelft.tam.model.JobOffer;

public enum AcademicPeriod {

	// Only add new levels at the end of the list, since the order is used in the database.

	// The dates are not exact, but they shouldn't need to be since the midpoint of a course's start and end date
	// are used to determine the AcademicPeriod.

	Q1(
		MonthDay.of(Month.SEPTEMBER, 1), MonthDay.of(Month.NOVEMBER, 10)
	),
	Q2(
		MonthDay.of(Month.NOVEMBER, 11), MonthDay.of(Month.FEBRUARY, 5)
	),
	Q3(
		MonthDay.of(Month.FEBRUARY, 6), MonthDay.of(Month.APRIL, 20)
	),
	Q4(
		MonthDay.of(Month.APRIL, 21), MonthDay.of(Month.JUNE, 30)
	),
	SUMMER(
		MonthDay.of(Month.JULY, 1), MonthDay.of(Month.AUGUST, 31)
	),
	S1(
		MonthDay.of(Month.SEPTEMBER, 1), MonthDay.of(Month.FEBRUARY, 5)
	),
	S2(
		MonthDay.of(Month.FEBRUARY, 6), MonthDay.of(Month.AUGUST, 31)
	),
	YEAR(
		MonthDay.of(Month.SEPTEMBER, 1), MonthDay.of(Month.AUGUST, 31)
	);

	private final MonthDay start;
	private final MonthDay end;

	AcademicPeriod(MonthDay start, MonthDay end) {
		this.start = start;
		this.end = end;
	}

	/**
	 * Determines the academic period from a string.
	 *
	 * @param  toCheck The string to check
	 * @return         The academic period, or null if none was determined
	 */
	public static AcademicPeriod getAcademicPeriodFromString(String toCheck) {
		String cleanedString = toCheck.toUpperCase().replace(" ", "");

		if (cleanedString.contains("1") && cleanedString.contains("2")) {
			return S1;
		} else if (cleanedString.contains("3") && cleanedString.contains("4")) {
			return S2;
		} else if (cleanedString.contains("1")) {
			return Q1;
		} else if (cleanedString.contains("2")) {
			return Q2;
		} else if (cleanedString.contains("3")) {
			return Q3;
		} else if (cleanedString.contains("4")) {
			return Q4;
		} else if (cleanedString.contains("5") || cleanedString.contains("Summer")
				|| cleanedString.contains("Zomer")) {
			return SUMMER;
		}

		return null;
	}

	/**
	 * Determines the academic period from start and end dates
	 *
	 * @param  start The start date
	 * @param  end   The end date
	 * @return       The academic period
	 */
	public static AcademicPeriod getAcademicPeriodFromStartAndEndDates(LocalDate start, LocalDate end) {
		LocalDate midpoint = LocalDate.ofEpochDay((start.toEpochDay() + end.toEpochDay()) / 2L);
		MonthDay midpointMonthDay = MonthDay.of(midpoint.getMonth(), midpoint.getDayOfMonth());
		int numberOfDays = start.until(end).getDays();

		if (numberOfDays > 300) {
			return YEAR;
		} else if (numberOfDays > 150) {
			if (midpointMonthDay.isAfter(S1.start) && midpointMonthDay.isBefore(S1.end)) {
				return S1;
			}
			return S2;
		}

		AcademicPeriod[] quarters = { Q1, Q2, Q3, Q4 };

		for (AcademicPeriod quarter : quarters) {
			if (midpointMonthDay.isAfter(quarter.start) && midpointMonthDay.isBefore(quarter.end)) {
				return quarter;
			}
		}

		return SUMMER;
	}

	/**
	 * Magical machine learning method to determine the AcademicPeriod of a job offer!
	 *
	 * @param  jobOffer The jobOffer to determine the AcademicPeriod for
	 * @return          The most probable AcademicPeriod
	 */
	public static AcademicPeriod detectAcademicPeriod(JobOffer jobOffer) {
		LocalDate startDate = jobOffer.getContractStartDate();
		LocalDate endDate = jobOffer.getContractEndDate();

		return getAcademicPeriodFromStartAndEndDates(startDate, endDate);
	}

}
