/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.io;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.*;

import lombok.AllArgsConstructor;
import lombok.Getter;

public abstract class ImportHandler<E> {

	public record Result<E>(List<E> result, Map<Integer, List<ImportWarning>> warnings,
			Map<Integer, List<ImportError>> errors) {
	}

	/**
	 * Imports a list of splitted lines.
	 *
	 * @param  lines The list of lines
	 * @return       The result of the input, either Error, WithWarnings, or Success for every line
	 */
	public final Result<E> importFile(List<String[]> lines) {
		preImport();

		List<E> res = new ArrayList<>();
		Map<Integer, List<ImportWarning>> warnings = new HashMap<>();
		Map<Integer, List<ImportError>> errors = new HashMap<>();

		for (int i = 0; i < lines.size(); i++) {
			String[] line = lines.get(i);
			try {
				ImportResult<E> result = importLine(line);
				if (result instanceof ImportResult.Success<E> r) {
					res.add(r.value());
				} else if (result instanceof ImportResult.WithWarnings<E> r) {
					res.add(r.value());
					warnings.put(i + 1, r.warnings());
				} else if (result instanceof ImportResult.Error<E> r) {
					errors.put(i + 1, List.of(r.error()));
				}
			} catch (ImportLineException e) {
				errors.put(i + 1, List.of(e.getError()));
			}
		}

		return new Result<>(res, warnings, errors);
	}

	/**
	 * Called before a file is imported.
	 */
	public void preImport() {
	}

	/**
	 * Imports one line. Results in an error, a result, or a result with warnings.
	 *
	 * @param  line The line to import
	 * @return      The result
	 */
	protected abstract ImportResult<E> importLine(String[] line);

	/**
	 * Stops importing the current line and adds an error
	 *
	 * @param error The error
	 */
	protected void error(ImportError error) {
		throw new ImportLineException(error);
	}

	/**
	 * Stops importing the current line and adds an error
	 *
	 * @param message The error message
	 */
	protected void error(String message) {
		throw new ImportLineException(message);
	}

	protected sealed interface ImportResult<E> {

		record Success<E>(E value) implements ImportResult<E> {
		}

		record WithWarnings<E>(E value, List<ImportWarning> warnings) implements ImportResult<E> {
			public WithWarnings(E value, ImportWarning warning) {
				this(value, List.of(warning));
			}
		}

		record Error<E>(ImportError error) implements ImportResult<E> {
		}

	}

	@Getter
	@AllArgsConstructor
	protected static class ImportLineException extends RuntimeException {
		private ImportError error;

		public ImportLineException(String message) {
			this.error = new SimpleImportError(message);
		}
	}

	protected OptionalInt parseInt(String value) {
		try {
			return OptionalInt.of(Integer.parseInt(value));
		} catch (NumberFormatException ignored) {
			return OptionalInt.empty();
		}
	}

	protected Optional<Boolean> parseBoolean(String value) {
		if (Set.of("1", "true", "yes", "y").contains(value.toLowerCase()))
			return Optional.of(true);
		if (Set.of("0", "false", "no", "n").contains(value.toLowerCase()))
			return Optional.of(false);
		return Optional.empty();
	}

	protected Optional<LocalDate> parseDate(String date) {
		try {
			return Optional.of(LocalDate.parse(date, DateTimeFormatter.ofPattern("dd-MM-yyyy")));
		} catch (DateTimeParseException ignored) {
			return Optional.empty();
		}
	}

	protected Optional<LocalDateTime> parseDateTime(String date) {
		try {
			return Optional.of(LocalDateTime.parse(date, DateTimeFormatter.ofPattern("dd-MM-yyyy HH:mm")));
		} catch (DateTimeParseException ignored) {
			return Optional.empty();
		}
	}

}
