/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.io;

import java.util.ArrayList;
import java.util.List;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.stereotype.Component;

import nl.tudelft.labracore.lib.security.LabradorUserDetails;
import nl.tudelft.tam.dto.create.TrainingApprovalCreateDTO;
import nl.tudelft.tam.dto.id.ProfileIdDTO;
import nl.tudelft.tam.model.TrainingApproval;
import nl.tudelft.tam.model.TrainingType;
import nl.tudelft.tam.service.PersonService;
import nl.tudelft.tam.service.TrainingApprovalService;
import nl.tudelft.tam.service.TrainingTypeService;

@Component
public class TrainingImportHandler extends ImportHandler<TrainingApproval> {

	private final PersonService personService;

	private final TrainingApprovalService trainingApprovalService;
	private final TrainingTypeService trainingTypeService;
	private Long programId = null;

	@Autowired
	public TrainingImportHandler(PersonService personService, TrainingApprovalService trainingApprovalService,
			TrainingTypeService trainingTypeService) {
		this.personService = personService;
		this.trainingApprovalService = trainingApprovalService;
		this.trainingTypeService = trainingTypeService;
	}

	public TrainingImportHandler forProgram(Long programId) {
		this.programId = programId;
		return this;
	}

	@Override
	public void preImport() {
		if (programId == null)
			throw new IllegalStateException("No programme ID set");
	}

	@Override
	protected ImportResult<TrainingApproval> importLine(String[] line) {
		TrainingApprovalCreateDTO training = new TrainingApprovalCreateDTO();
		List<ImportWarning> warnings = new ArrayList<>();

		if (line.length != 3) {
			return new ImportResult.Error<>(new SimpleImportError("Wrong amount of entries"));
		}

		if (line[0] == null || line[0].isEmpty()) {
			return new ImportResult.Error<>(new SimpleImportError("User identifier is not specified"));
		}

		var personSearchResult = personService.searchForPeople(line[0]);

		if (personSearchResult.isEmpty()) {
			return new ImportResult.Error<>(
					new SimpleImportError("User with identifier " + line[0] + " not found"));
		}

		if (personSearchResult.size() > 1) {
			warnings.add(new SimpleImportWarning(
					"Multiple users with identifier " + line[0] + " found, using the first one"));
		}

		List<TrainingType> validTrainingTypes = trainingTypeService
				.getTrainingTypesByNameAndProgramId(line[1], programId);

		if (validTrainingTypes.isEmpty()) {
			return new ImportResult.Error<>(
					new SimpleImportError("Training type with name " + line[1] + " not found"));
		}

		if (validTrainingTypes.size() > 1) {
			warnings.add(new SimpleImportWarning(
					"Multiple training types with name " + line[1] + " found, using the first one"));
		}

		if (trainingApprovalService
				.findByPersonIdAndType(personSearchResult.get(0).getId(), validTrainingTypes.get(0))
				.isPresent()) {
			return new ImportResult.Error<>(
					new SimpleImportError("User " + line[0] + " already completed training " + line[1]));
		}

		training.setPerson(new ProfileIdDTO(personSearchResult.get(0).getId()));
		training.setType(validTrainingTypes.get(0));
		training.setDatePassed(parseDate(line[2])
				.orElseThrow(() -> new ImportLineException(
						"Deadline is not in the correct format, expected: dd-MM-yyyy")));
		training.setApprovedBy(new ProfileIdDTO(
				((LabradorUserDetails) SecurityContextHolder.getContext().getAuthentication().getPrincipal())
						.getUser().getId()));
		return warnings.isEmpty() ? new ImportResult.Success<>(training.apply())
				: new ImportResult.WithWarnings<>(training.apply(), warnings);
	}
}
