/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.io.IOException;
import java.nio.file.Path;
import java.time.LocalDateTime;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.transaction.Transactional;
import javax.validation.constraints.NotNull;

import org.apache.commons.lang3.ArrayUtils;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.ApplicationCreateDTO;
import nl.tudelft.tam.dto.create.EmailCreateDTO;
import nl.tudelft.tam.dto.id.JobOfferIdDTO;
import nl.tudelft.tam.dto.util.ApplicationFilterDTO;
import nl.tudelft.tam.dto.view.details.ApplicationDetailsJobOfferDTO;
import nl.tudelft.tam.dto.view.details.ApplicationDetailsPersonDTO;
import nl.tudelft.tam.dto.view.details.JobOfferDetailsDTO;
import nl.tudelft.tam.dto.view.details.TrainingTypeDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ApplicationStatusSummaryDTO;
import nl.tudelft.tam.dto.view.summary.ApplicationSummaryDTO;
import nl.tudelft.tam.dto.view.summary.JobOfferSummaryDTO;
import nl.tudelft.tam.dto.view.summary.TrainingTypeSummaryDTO;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.exception.ApplicationStatusException;
import nl.tudelft.tam.model.*;
import nl.tudelft.tam.model.records.FlexDelftExport;
import nl.tudelft.tam.repository.ApplicationChangeEventRepository;
import nl.tudelft.tam.repository.ApplicationExportRepository;
import nl.tudelft.tam.repository.ApplicationRepository;

@Service
public class ApplicationService {

	// region class attributes

	@Value("${tam.filesys.exports-dir}")
	String exportDir;

	@Autowired
	ApplicationRepository applicationRepository;

	@Autowired
	ApplicationChangeEventRepository applicationChangeEventRepository;

	@Autowired
	ApplicationExportRepository applicationExportRepository;

	@Autowired
	PersonService personService;

	@Autowired
	FileService fileService;

	@Autowired
	CSVService csvService;

	@Autowired
	JobOfferService jobOfferService;

	@Autowired
	RoleService roleService;

	@Autowired
	CoordinatorService coordinatorService;

	@Autowired
	EditionService editionService;

	@Autowired
	CourseService courseService;

	@Autowired
	ProfileService profileService;

	@Autowired
	CohortService cohortService;

	@Autowired
	CoordinatorDefaultService coordinatorDefaultService;

	@Autowired
	ModelMapper mapper;

	@Autowired
	EmailService emailService;

	@Autowired
	TrainingTypeService trainingTypeService;

	@Autowired
	TrainingApprovalService trainingApprovalService;

	// endregion

	// region manipulate data

	/**
	 * Save a new application into the database. In case one already exists with status rejected by student,
	 * changes the status to submitted.
	 *
	 * @param  personId                   id of the person
	 * @param  create                     The create DTO
	 * @return                            The created application or the already existing one
	 * @throws ApplicationStatusException if open application already exists
	 */
	@Transactional
	public Application submit(Long personId, ApplicationCreateDTO create) {
		Application.AppId id = new Application.AppId(personId, create.getJobOffer().getId());

		applicationRepository.findById(id).ifPresent(app -> {
			if (app.getStatus() == Status.REJECTED_BY_STUDENT) {
				applicationChangeEventRepository.deleteAllByApplicationId(id);
				applicationRepository.delete(app);
			} else {
				throw new ApplicationStatusException("Application already exists for personId:[" + personId
						+ "] and offerId:[" + create.getJobOffer().getId() + "].");
			}
		});

		Application app = create.apply();
		app.setId(id);
		app = applicationRepository.save(app);
		setStatus(app, Status.SUBMITTED);

		EditionDetailsDTO edition = editionService.getEditionById(app.getJobOffer().getEditionId());
		// Send notification of successful submission
		sendApplicationReceivedNotification(personId, app.getJobOffer().getName(),
				edition.getCourse().getName(),
				edition.getCourse().getCode(), edition.getName());

		return app;
	}

	/**
	 * Retract the application. Deleted the application if the application was not offered.
	 *
	 * @param personId id of the person
	 * @param offerId  id of the job offer
	 */
	@Transactional
	public void retract(Long personId, Long offerId) {
		Application.AppId id = new Application.AppId(personId, offerId);
		Application app = applicationRepository.findByIdOrThrow(id);
		if (app.getStatus() == Status.SUBMITTED) {
			applicationChangeEventRepository.deleteAllByApplicationId(id);
			applicationRepository.delete(app);
		} else {
			setStatus(app, Status.REJECTED_BY_STUDENT);
			applicationRepository.save(app);
		}
	}

	/**
	 * Reject the offer. Change status to rejected by teacher.
	 *
	 * @param personId id of the person
	 * @param offerId  id of the job offer
	 * @param handler  id of the person rejecting (ie. teacher)
	 */
	@Transactional
	public void reject(Long personId, Long offerId, Long handler) {
		Application.AppId id = new Application.AppId(personId, offerId);
		Application app = applicationRepository.findByIdOrThrow(id);
		app.setHandler(profileService.findByIdOrThrow(handler));
		setStatus(app, Status.REJECTED_BY_TEACHER);
		applicationRepository.save(app);

		roleService.removeRole(personId, offerId);

		// Send notification of rejection to student
		EditionDetailsDTO edition = editionService.getEditionById(app.getJobOffer().getEditionId());
		CourseDetailsDTO course = courseService.getCourseById(edition.getCourse().getId());
		String courseStaff = String.join(", ",
				course.getManagers().stream().map(PersonSummaryDTO::getDisplayName).toList());
		String rejectionMessage = app.getJobOffer().getRejectMessage();
		sendApplicationRejectedNotification(personId, app.getJobOffer().getName(),
				edition.getCourse().getName(), edition.getCourse().getCode(), edition.getName(),
				rejectionMessage, courseStaff);
	}

	/**
	 * All submitted applications for a job offer will be terminated.
	 *
	 * @param id      of the offer to be terminated.
	 * @param handler id of the person rejecting (ie. teacher)
	 */
	@Transactional
	public void rejectAllOpenInJobOffer(Long id, Long handler) {
		JobOffer jobOffer = jobOfferService.findByIdOrThrow(id);
		EditionDetailsDTO edition = editionService.getEditionById(jobOffer.getEditionId());
		CourseDetailsDTO course = courseService.getCourseById(edition.getCourse().getId());
		String courseStaff = String.join(", ",
				course.getManagers().stream().map(PersonSummaryDTO::getDisplayName).toList());
		String rejectionMessage = jobOffer.getRejectMessage();

		jobOffer.getApplications().forEach(a -> {
			if (a.getStatus() == Status.SUBMITTED) {
				a.setHandler(profileService.findByIdOrThrow(handler));
				setStatus(a, Status.REJECTED_BY_TEACHER);
				roleService.removeRole(a.getId().getPersonId(), a.getId().getJobOfferId());

				// Send notification to the student
				sendApplicationRejectedNotification(a.getId().getPersonId(), jobOffer.getName(),
						edition.getCourse().getName(), edition.getCourse().getCode(), edition.getName(),
						rejectionMessage, courseStaff);
			}
		});
	}

	/**
	 * Finds the IDs of people the position is rejected for; Rejects them for the job.
	 *
	 * @param offerId     ID of job offer
	 * @param identifiers Identifiers of people for which the job is rejected
	 * @param handler     id of the person rejecting (ie. teacher)
	 */
	@Transactional
	public void bulkReject(Long offerId, String identifiers, Long handler) {
		List<PersonSummaryDTO> persons = personService.searchForPeople(identifiers);
		persons.forEach(p -> reject(p.getId(), offerId, handler));
	}

	/**
	 * Propose an offer. Change status to offered. PersonID not validated.
	 *
	 * @param personId id of the person
	 * @param offerId  id of the job offer
	 * @param handler  id of the person offering (ie. teacher)
	 */
	@Transactional
	public void offer(Long personId, Long offerId, Long handler) {
		Application.AppId id = new Application.AppId(personId, offerId);
		Application app = applicationRepository.findById(id)
				.orElseGet(() -> submit(personId,
						ApplicationCreateDTO.builder().jobOffer(new JobOfferIdDTO(offerId)).build()));
		if (app.getStatus().equals(Status.OFFERED)
				|| app.getStatus().equals(Status.ACCEPTED))
			return;
		app.setHandler(profileService.findByIdOrThrow(handler));
		setStatus(app, Status.OFFERED);

		// Send notification of offer to student
		EditionDetailsDTO edition = editionService.getEditionById(app.getJobOffer().getEditionId());
		String hiringMessage = app.getJobOffer().getHiringMessage();
		sendApplicationOfferedNotification(personId, app.getJobOffer().getName(),
				edition.getCourse().getName(), edition.getCourse().getCode(), edition.getName(),
				hiringMessage);
	}

	/**
	 * Sets the status of an application and saves a change event.
	 *
	 * @param application The application
	 * @param status      The new status
	 */
	private void setStatus(Application application, Status status) {
		if (status == application.getStatus())
			return;

		application.setStatus(status);

		ApplicationChangeEvent event = ApplicationChangeEvent.builder().application(application)
				.status(status).build();
		applicationChangeEventRepository.save(event);
	}

	/**
	 * Finds the IDs of people the position is offered to; Makes offer to everyone for the job.
	 *
	 * @param offerId     ID of job offer
	 * @param identifiers Identifiers of people the job is offered to
	 * @param handler     id of the person offering (ie. teacher)
	 */
	@Transactional
	public void bulkOffer(Long offerId, String identifiers, Long handler) {
		List<PersonSummaryDTO> persons = personService.searchForPeople(identifiers);
		persons.forEach(p -> offer(p.getId(), offerId, handler));
	}

	/**
	 * Propose an offer. Change status to offered. Find by username.
	 *
	 * @param  personUsername username of the person
	 * @param  offerId        id of the job offer
	 * @param  handler        id of the person offering (ie. teacher)
	 *
	 * @return                success status
	 */
	@Transactional
	public boolean offerByUsername(String personUsername, Long offerId, Long handler) {
		try {
			Long personId = personService.getPersonByUsername(personUsername).getId();
			this.offer(personId, offerId, handler);
			return true;
		} catch (Exception e) {
			return false;
		}
	}

	/**
	 * Propose an offer. Change status to offered. Find by student number (with validation)
	 *
	 * @param  personNumber username of the person
	 * @param  offerId      id of the job offer
	 * @param  handler      id of the person offering (ie. teacher)
	 *
	 * @return              success status
	 */
	@Transactional
	public boolean offerByStudentNumber(Long personNumber, Long offerId, Long handler) {
		try {
			Long personId = personService.getPersonById(personNumber).getId();
			this.offer(personId, offerId, handler);
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
	}

	/**
	 * Accept the offer. Change status to accepted.
	 *
	 * @param  personId                   id of the person
	 * @param  offerId                    id of the job offer
	 * @throws ApplicationStatusException if status is not "offered"
	 */
	@Transactional
	public void accept(Long personId, Long offerId) {
		Application.AppId id = new Application.AppId(personId, offerId);
		Application app = applicationRepository.findByIdOrThrow(id);
		if (!app.getStatus().equals(Status.OFFERED))
			throw new ApplicationStatusException(
					"Trying to change application from status:[" + app.getStatus() + "] to accepted.");
		setStatus(app, Status.ACCEPTED);

		roleService.changeToTa(personId, offerId, false);
	}

	/**
	 * Promote student to head TA.
	 *
	 * @param  personId                   id of the person
	 * @param  offerId                    id of the job offer
	 * @throws ApplicationStatusException if status is not "accepted"
	 */
	@Transactional
	public void promote(Long personId, Long offerId) {
		Application.AppId id = new Application.AppId(personId, offerId);
		Application app = applicationRepository.findByIdOrThrow(id);
		if (!app.getStatus().equals(Status.ACCEPTED))
			throw new IllegalStateException(
					"Trying to change application from status:[" + app.getStatus() + "] to promoted.");

		roleService.changeToHeadTa(personId, offerId);
	}

	/**
	 * Demote student to TA.
	 *
	 * @param  personId                   id of the person
	 * @param  offerId                    id of the job offer
	 * @throws ApplicationStatusException if status is not "accepted"
	 */
	@Transactional
	public void demote(Long personId, Long offerId) {
		Application.AppId id = new Application.AppId(personId, offerId);
		Application app = applicationRepository.findByIdOrThrow(id);
		if (!app.getStatus().equals(Status.ACCEPTED))
			throw new IllegalStateException(
					"Trying to change application from status:[" + app.getStatus() + "] to demoted.");

		roleService.changeToTa(personId, offerId, true);
	}

	// endregion

	// region checkers

	/**
	 * Checks if application exists for a person in job offer.
	 *
	 * @param  personId id of the person
	 * @param  offerId  id of the person
	 * @return          true if application exists
	 */
	public boolean appExistsFor(Long personId, Long offerId) {
		Application.AppId id = new Application.AppId(personId, offerId);
		return applicationRepository.findById(id).isPresent();
	}

	/**
	 * Checks if an application has status offer for a person and offer id.
	 *
	 * @param  personId id of the person
	 * @param  offerId  id of the job offer
	 * @return          true if status is offered
	 */
	@NotNull
	public boolean hasOffer(Long personId, Long offerId) {
		return applicationRepository.findByIdOrThrow(new Application.AppId(personId, offerId)).getStatus()
				.equals(Status.OFFERED);
	}

	/**
	 * Check if a person has ta'ed a course which is provided by a job offer.
	 *
	 * @param  personId   The person for who to check
	 * @param  jobOfferId The job offer for which to check
	 * @return            True if the person has TAed the course related to the job offer.
	 */
	public boolean hasTAedBefore(Long personId, Long jobOfferId) {
		JobOffer jobOffer = jobOfferService.findByIdOrThrow(jobOfferId);
		EditionDetailsDTO edition = editionService.getEditionById(jobOffer.getEditionId());
		CourseDetailsDTO course = courseService.getOrThrow(edition.getCourse().getId());
		List<RoleDetailsDTO> previousRoles = roleService
				.getRolesById(course.getEditions().stream().map(EditionSummaryDTO::getId)
						.collect(Collectors.toList()), List.of(personId))
				.stream()
				.filter(r -> r.getType() == RoleDetailsDTO.TypeEnum.TA
						|| r.getType() == RoleDetailsDTO.TypeEnum.HEAD_TA)
				.toList();

		return !previousRoles.isEmpty();
	}

	/**
	 * Checks whether there are no valid student actions at the moment
	 *
	 * @param  editionOffers  The offers for an edition
	 * @param  offerStatusMap Map which maps offer ids to a status summary
	 * @return                Checks whether all offers of an edition require no more student actions
	 */
	public boolean checkEditionCollapsable(List<JobOfferSummaryDTO> editionOffers,
			Map<Long, ApplicationStatusSummaryDTO> offerStatusMap) {
		var collapsableFilter = Set.of(Status.ACCEPTED, Status.REJECTED_BY_TEACHER);
		return editionOffers.stream().allMatch(offer -> offer.getHiringMessage() == null)
				&& editionOffers.stream().map(offer -> offerStatusMap.get(offer.getId()))
						.allMatch(offer -> offer.exists() && collapsableFilter.contains(offer.getStatus()));
	}

	// endregion

	// region getters

	/**
	 * Retrieves a list of all applications with job details of one user. Ignore retracted applications.
	 *
	 * @param  id of the person
	 * @return    list of applications
	 */
	public List<ApplicationDetailsJobOfferDTO> getApplicationsForPerson(Long id) {
		List<Application> apps = applicationRepository.findAllByPersonId(id)
				.stream().filter(a -> !a.getStatus().equals(Status.REJECTED_BY_STUDENT))
				.toList();
		if (apps.size() == 0)
			return new ArrayList<>();
		Map<Long, EditionDetailsDTO> map = editionService
				.get(apps.stream().map(a -> a.getJobOffer().getEditionId()).collect(Collectors.toList()))
				.stream()
				.collect(Collectors.toMap(EditionDetailsDTO::getId, Function.identity(), (a, b) -> a));
		List<ApplicationDetailsJobOfferDTO> res = new ArrayList<>();
		for (Application a : apps) {
			ApplicationDetailsJobOfferDTO dto = mapper.map(a, ApplicationDetailsJobOfferDTO.class);
			JobOfferDetailsDTO extra = mapper.map(a.getJobOffer(), JobOfferDetailsDTO.class);
			extra.setEdition(map.get(extra.getEditionId()));
			dto.setJobOffer(extra);
			res.add(dto);
		}
		return res;
	}

	/**
	 * Gets all applications with a certain status for a person
	 *
	 * @param  id     The id of the person to get the applications for
	 * @param  status The status which the applications need to have
	 * @return        The list of applications
	 */
	public List<Application> getApplicationsForPersonByStatus(Long id, Status status) {
		return applicationRepository.findAllByPersonId(id)
				.stream().filter(a -> a.getStatus().equals(status))
				.toList();
	}

	/**
	 * List of all applications with person details of one job offer. Ignore retracted applications.
	 *
	 * @param  id of the job offer
	 * @return    the list of applications
	 */
	public List<ApplicationDetailsPersonDTO> getApplicationsForJobOffer(Long id) {
		List<ApplicationDetailsPersonDTO> apps = applicationRepository.findAllByOfferId(id).stream()
				.map(a -> mapper.map(a, ApplicationDetailsPersonDTO.class)).toList();
		if (apps.isEmpty())
			return apps;
		Map<Long, PersonSummaryDTO> ppl = personService
				.getPeopleById(
						apps.stream().map(ApplicationSummaryDTO::getPersonId).collect(Collectors.toList()))
				.stream().collect(Collectors.toMap(PersonSummaryDTO::getId, p -> p));
		apps.forEach(a -> {
			a.setPerson(ppl.get(a.getPersonId()));
			a.setTaBefore(hasTAedBefore(a.getPersonId(), a.getJobOfferId()));
		});
		return apps;
	}

	/**
	 * List of application with job offer details for one person, filtered by a query.
	 *
	 * @param  id     of the person
	 * @param  filter query
	 * @return        filtered list of applications
	 */
	public List<ApplicationDetailsJobOfferDTO> getFilteredApplicationsForPerson(Long id, String filter) {
		List<ApplicationDetailsJobOfferDTO> apps = getApplicationsForPerson(id);

		if (filter == null || filter.isBlank())
			return apps;
		String f = filter.toLowerCase();

		return apps.stream()
				.filter(a -> a.getJobOffer().getEdition().getCourse().getName().toLowerCase().contains(f)
						|| a.getJobOffer().getEdition().getCourse().getCode().toLowerCase().contains(f)
						|| a.getJobOffer().getEdition().getCohort().getName().toLowerCase().contains(f)
						|| a.getJobOffer().getEdition().getName().toLowerCase().contains(f)
						|| a.getJobOffer().getName().toLowerCase().contains(f)
						|| a.getStatus().name().toLowerCase().contains(f))
				.collect(Collectors.toList());
	}

	/**
	 * List of all applications with person details for one job offer, filtered by a query.
	 *
	 * @param  id     of the job offer
	 * @param  filter the query filter
	 * @return        the filtered list of applications
	 */
	public List<ApplicationDetailsPersonDTO> getFilteredApplicationsForJobOffer(Long id, String filter) {
		List<ApplicationDetailsPersonDTO> apps = getApplicationsForJobOffer(id);
		if (filter == null || filter.isBlank())
			return apps;
		String f = filter.toLowerCase();
		return apps.stream().filter(a -> a.getStatus().toString().toLowerCase().contains(f)
				|| a.getPerson().getDisplayName().toLowerCase().contains(f)
				|| a.getPerson().getUsername().toLowerCase().contains(f)
				|| a.getPerson().getExternalId().toLowerCase().contains(f)
				|| a.getPerson().getNumber().toString().contains(f)
				|| a.getPerson().getEmail().toLowerCase().contains(f)).collect(Collectors.toList());
	}

	/**
	 * By a list of offerIds of job offers, finds relative applications for a person and maps them.
	 *
	 * @param  personId id of the person
	 * @param  offerIds list of ids of the job offers
	 * @return          a map from job offer id to application if present
	 */
	public HashMap<Long, ApplicationSummaryDTO> getAppsMapForPersonByIds(Long personId, List<Long> offerIds) {
		HashMap<Long, ApplicationSummaryDTO> map = new HashMap<>();
		offerIds.forEach(id -> applicationRepository.findById(new Application.AppId(personId, id))
				.filter(a -> a.getStatus() != Status.REJECTED_BY_STUDENT)
				.ifPresent(a -> map.put(id, mapper.map(a, ApplicationSummaryDTO.class))));
		return map;
	}

	/**
	 * Counts the total number of applications, offers, accepted offers, and rejected offers.
	 *
	 * @param  applications the applications to be counted
	 * @return              an array of three integers, representing the total number of applications, offers,
	 *                      accepted offers, and rejected offers
	 */
	public <A extends ApplicationSummaryDTO> int[] getStatsForApps(List<A> applications) {
		int[] stats = { applications.size(), 0, 0, 0 };
		applications.forEach(a -> {
			switch (a.getStatus()) {
				case OFFERED -> stats[1]++;
				case ACCEPTED -> stats[2]++;
				case REJECTED_BY_STUDENT -> stats[0]--;
				case REJECTED_BY_TEACHER -> stats[3]++;
			}
		});
		return stats;
	}

	/**
	 * Computes the number of applications that still need to be handled.
	 *
	 * @param  jobId the id of the job with applications
	 * @return       the number of total applications that have not been accepted or rejected
	 */
	public int getNumberOfUnhandledApplications(Long jobId) {
		List<ApplicationDetailsPersonDTO> apps = getApplicationsForJobOffer(jobId);
		int[] stats = getStatsForApps(apps);
		return (stats[0] - stats[1] - stats[2] - stats[3]);
	}

	/**
	 * Retrieves applications for person and offers. Returns a map from offer id to whether application exists
	 * and is active.
	 *
	 * @param  personId the id of the person
	 * @param  offerIds the list of ids of job offers
	 * @return          a map from offer id to boolean, true if application exists
	 */
	public Map<Long, Boolean> getMapAppOpenByIdForPerson(Long personId, List<Long> offerIds) {
		Map<Long, Boolean> map = new HashMap<>();
		for (Long offerId : offerIds) {
			Optional<Application> app = applicationRepository
					.findById(new Application.AppId(personId, offerId));
			map.put(offerId, app.isPresent() && app.get().getStatus() != Status.REJECTED_BY_STUDENT);
		}
		return map;
	}

	/**
	 * Retrieves applications for person and offers. Returns a map from offer id to a tuple containing a
	 * boolean indicating whether the application exists and is active, and the application status.
	 *
	 * @param  personId The id of the person
	 * @param  offerIds The list of ids of job offers
	 * @return          A map from offer id to tuple of boolean and application status
	 */
	public Map<Long, ApplicationStatusSummaryDTO> getMapWithStatusAppOpenByIdForPerson(Long personId,
			List<Long> offerIds) {
		Map<Long, ApplicationStatusSummaryDTO> map = new HashMap<>();
		for (Long offerId : offerIds) {
			Optional<Application> app = applicationRepository
					.findById(new Application.AppId(personId, offerId));
			if (app.isPresent()) {
				map.put(offerId,
						new ApplicationStatusSummaryDTO(app.get().getStatus() != Status.REJECTED_BY_STUDENT,
								app.get().getStatus()));
			} else {
				map.put(offerId, new ApplicationStatusSummaryDTO(false, null));
			}
		}
		return map;
	}

	/**
	 * Gets all the applications a person coordinates.
	 *
	 * @param  personId The id of the person
	 * @return          A list of all applications the person coordinates
	 */
	public List<Application> getCoordinatingApplications(Long personId) {
		List<Long> programIds = coordinatorService.findAllByPersonId(personId).stream()
				.map(Coordinator::getProgramId).collect(Collectors.toList());
		return applicationRepository.findAll().stream()
				.filter(e -> programIds.contains(courseService
						.getOrThrow(
								editionService.getOrThrow(e.getJobOffer().getEditionId()).getCourse().getId())
						.getProgram().getId()))
				.collect(Collectors.toList());
	}

	/**
	 * Gets all the applications a person coordinates.
	 *
	 * @param  personId  The id of the person
	 * @param  programId The id of the program
	 * @return           A list of all applications the person coordinates in the program
	 */
	public List<Application> getCoordinatingApplicationsForProgram(Long personId, Long programId) {
		List<Application> applications = getCoordinatingApplications(personId);
		Map<Long, EditionDetailsDTO> editions = editionService
				.getEditionsById(
						applications.stream().map(a -> a.getJobOffer().getEditionId()).distinct().toList())
				.stream().collect(Collectors.toMap(EditionDetailsDTO::getId, Function.identity()));
		if (!applications.isEmpty()) {
			Map<Long, CourseDetailsDTO> courses = courseService
					.getCoursesById(
							editions.values().stream().map(e -> e.getCourse().getId()).distinct().toList())
					.stream().collect(Collectors.toMap(CourseDetailsDTO::getId, Function.identity()));
			return applications.stream().filter(a -> programId.equals(courses
					.get(editions.get(a.getJobOffer().getEditionId()).getCourse().getId()).getProgram()
					.getId()))
					.toList();
		} else {
			return List.of();
		}
	}

	/**
	 * Finds all applications for a specific job offer
	 *
	 * @param  id The job offer to search for
	 * @return    The list of applications found
	 */
	public List<Application> findAllByOfferId(Long id) {
		return applicationRepository.findAllByOfferId(id);
	}

	/**
	 * Finds all the exports a person did for a program.
	 *
	 * @param  person    The person
	 * @param  programId The id of the program
	 * @return           All exports done by the person
	 */
	public List<ApplicationExport> getExportsForPersonAndProgram(Person person, Long programId) {
		return applicationExportRepository.findAllByExportedByAndProgramIdOrderByExportAtDesc(person.getId(),
				programId);
	}

	/**
	 * Gets the list of coordinating applications, filtered by the provided filter.
	 *
	 * @param  personId The id of the coordinator
	 * @param  filter   The filter to apply
	 * @return          The filtered list of coordinating applications
	 */
	public List<Application> getFilteredCoordinatingApplications(Long personId,
			ApplicationFilterDTO filter) {
		List<Application> applications = getCoordinatingApplicationsForProgram(personId, filter.getProgram())
				.stream()
				.filter(a -> !editionService.getOrThrow(a.getJobOffer().getEditionId()).getIsArchived())
				.toList();
		return filterApplications(applications, filter);
	}

	/**
	 * Filter a list of applications.
	 *
	 * @param  applications The list of applications
	 * @param  filter       The filter to apply
	 * @return              The filtered list of applications
	 */
	public List<Application> filterApplications(List<Application> applications,
			ApplicationFilterDTO filter) {
		return applications.stream()
				.filter(app -> filter.getStatuses() == null || filter.getStatuses().contains(app.getStatus()))
				.filter(app -> filter.getContractStartAfter() == null
						|| !app.getJobOffer().getContractStartDate().isBefore(filter.getContractStartAfter()))
				.filter(app -> filter.getContractStartBefore() == null
						|| !app.getJobOffer().getContractStartDate().isAfter(filter.getContractStartBefore()))
				.filter(app -> filter.getUpdatedSince() == null || !applicationChangeEventRepository
						.findByApplicationIdAndChangedAtAfter(app.getId(), filter.getUpdatedSince())
						.isEmpty())
				.toList();
	}

	// endregion

	// region export

	/**
	 * Exports a list of applications.
	 *
	 * @param  name         The name of the file
	 * @param  applications The list of applications
	 * @return              The Path to the export file
	 */
	public Path exportApplications(String name, List<Application> applications) throws IOException {
		return exportApplications(name, applications, false);
	}

	/**
	 * Exports a list of applications.
	 *
	 * @param  name             The name of the file
	 * @param  applications     The list of applications
	 * @param  specificJobOffer Determines if this is a mixture of job offers or not
	 * @return                  The Path to the export file
	 */
	public Path exportApplications(String name, List<Application> applications, boolean specificJobOffer)
			throws IOException {
		String[] headers;

		if (specificJobOffer) {
			headers = new String[] { "Student Number", "Email", "Name", "TAed Before", "Status",
					"Student Response" };
		} else {
			headers = new String[] { "Student Number", "Email", "Name", "Edition", "Job Offer",
					"Status", "Student Response" };
		}

		List<TrainingTypeDetailsDTO> trainingTypes = trainingTypeService
				.getDetails(
						applications.stream().flatMap(a -> a.getJobOffer().getRequiredTrainings().stream())
								.map(TrainingType::getId).distinct().toList())
				.stream()
				.sorted(Comparator.<TrainingTypeDetailsDTO, String>comparing(t -> t.getProgram().getName())
						.thenComparing(TrainingTypeSummaryDTO::getName))
				.toList();

		for (TrainingTypeDetailsDTO trainingType : trainingTypes) {
			headers = ArrayUtils.add(headers,
					trainingType.getProgram().getName() + " - " + trainingType.getName());
		}

		Map<Long, PersonSummaryDTO> people = personService
				.get(applications.stream().map(a -> a.getId().getPersonId()).distinct())
				.stream().collect(Collectors.toMap(PersonSummaryDTO::getId, Function.identity()));
		Map<Long, EditionDetailsDTO> editions = editionService
				.get(applications.stream().map(a -> a.getJobOffer().getEditionId()).distinct().toList())
				.stream().collect(Collectors.toMap(EditionDetailsDTO::getId, Function.identity()));

		Stream<String[]> data = applications.stream().map(app -> {
			PersonSummaryDTO applicant = people.get(app.getId().getPersonId());
			EditionDetailsDTO edition = editions.get(app.getJobOffer().getEditionId());

			String[] row = new String[(specificJobOffer ? 6 : 7) + trainingTypes.size()];
			int i = 0;
			row[i++] = applicant.getNumber().toString();
			row[i++] = applicant.getEmail();
			row[i++] = applicant.getDisplayName();
			if (specificJobOffer) {
				String taedBefore = hasTAedBefore(applicant.getId(), app.getJobOffer().getId()) ? "YES"
						: "NO";
				row[i++] = taedBefore;
			} else {
				row[i++] = edition.getCourse().getName() + " - " + edition.getName();
				row[i++] = app.getJobOffer().getName();
			}
			row[i++] = app.getStatus().toString();
			row[i++] = app.getContent();

			for (TrainingTypeDetailsDTO trainingType : trainingTypes) {
				if (app.getJobOffer().getRequiredTrainings().stream()
						.anyMatch(t -> Objects.equals(t.getId(), trainingType.getId()))) {
					Optional<TrainingApproval> approval = trainingApprovalService
							.findByPersonIdAndTypeId(applicant.getId(), trainingType.getId());
					if (approval.isPresent()) {
						row[i++] = "Completed";
					} else {
						row[i++] = "Not Completed";
					}
				} else {
					row[i++] = "N/A";
				}
			}

			return row;
		});

		String fileName = fileService.toFileName(name) + ".csv";

		csvService.writeToCSV(Stream.concat(Stream.<String[]>of(headers), data).collect(Collectors.toList()),
				Path.of(exportDir), fileName);

		return Path.of(exportDir, fileName);
	}

	/**
	 * Exports a list of applications with FlexDelft specific fields
	 *
	 * @param  person       The person exporting
	 * @param  programId    The program being exported for
	 * @param  batch        The batch number
	 * @param  name         The name of the file
	 * @param  applications The list of applications
	 * @param  since        Export changes since a timestamp, null for all applications
	 * @return              The Path to the export file
	 */
	@Transactional
	public Path exportApplicationsForFlexDelft(Person person, Long programId, Integer batch, String name,
			List<Application> applications, LocalDateTime since)
			throws IOException {
		if (since != null) {
			applications = filterApplications(applications,
					ApplicationFilterDTO.builder().updatedSince(since).build());
		}

		applicationExportRepository.save(ApplicationExport.builder().exportedBy(person.getId())
				.programId(programId).batchNumber(batch).build());

		// Fetch all people and editions to cache and avoid many requests to core
		personService.get(applications.stream().map(a -> a.getId().getPersonId()));
		editionService.get(
				applications.stream().map(a -> a.getJobOffer().getEditionId()).collect(Collectors.toList()));

		String[] headers;

		headers = new String[] { "Name", "NetID", "Email", "Contract Name", "Baan Code", "Approver Name",
				"Approver Email",
				"Coordinator",
				"Max Hours", "PayScale", "Contract Start Date", "Contract End Date", "Batch", "Remarks" };

		Stream<String[]> data = applications.stream().map(app -> {
			PersonSummaryDTO applicant = personService.getOrThrow(app.getId().getPersonId());
			Profile applicantProfile = profileService.findByIdOrThrow(app.getId().getPersonId());
			JobOffer appJobOffer = app.getJobOffer();
			EditionDetailsDTO edition = editionService.getOrThrow(appJobOffer.getEditionId());
			CoordinatorDefault coordDefault = coordinatorDefaultService.getOrCreate(
					cohortService.getOrThrowById(edition.getCohort().getId()).getProgram().getId());
			PersonSummaryDTO approver = jobOfferService.getApprover(edition.getId());

			FlexDelftExport export = new FlexDelftExport(applicant, applicantProfile, appJobOffer,
					edition, coordDefault, approver, batch.toString(), "");
			return export.toExport();
		});

		String fileName = fileService.toFileName(name) + ".csv";

		csvService.writeToCSV(Stream.concat(Stream.<String[]>of(headers), data).collect(Collectors.toList()),
				Path.of(exportDir), fileName);

		return Path.of(exportDir, fileName);
	}

	// end region

	// region notifications

	/**
	 * Send an application successfully received notification
	 *
	 * @param personId   The ID of the person to send the notification to
	 * @param position   The position that the application was for
	 * @param course     The course that the application was for
	 * @param courseCode The course code of the course that the application was for
	 * @param year       The name of the edition (eg. Q1 22/23)
	 */
	public void sendApplicationReceivedNotification(Long personId, String position, String course,
			String courseCode, String year) {
		Profile profile = profileService.findByIdOrCreate(personId);
		PersonDetailsDTO person = personService.getPersonById(personId);

		EmailCreateDTO email = EmailCreateDTO.builder()
				.subject("Application (" + courseCode + ") Successfully Submitted!")
				.template("html/application_submitted")
				.contextVariables(
						Map.of("name", person.getDisplayName(), "position", position, "course", course,
								"courseCode", courseCode, "year", year))
				.build();

		if (profile.getEmailNotifications()) {
			emailService.sendEmail(List.of(person.getEmail()), email);
		}
	}

	/**
	 * Send an application offered notification
	 *
	 * @param personId      The ID of the person to send the notification to
	 * @param position      The position that the application was for
	 * @param course        The course that the application was for
	 * @param courseCode    The course code of the course that the application was for
	 * @param year          The name of the edition (eg. Q1 22/23)
	 * @param hiringMessage The hiring message from the course staff (can be null)
	 */
	public void sendApplicationOfferedNotification(Long personId, String position, String course,
			String courseCode, String year, String hiringMessage) {
		Profile profile = profileService.findByIdOrCreate(personId);
		PersonDetailsDTO person = personService.getPersonById(personId);

		Map<String, Object> contextVars = new HashMap<>(
				Map.of("name", person.getDisplayName(), "position", position, "course", course,
						"courseCode", courseCode, "year", year));

		EmailCreateDTO email = EmailCreateDTO.builder()
				.subject("Action Required: Position (" + courseCode + ") Offered!")
				.template("html/application_offered")
				.contextVariables(contextVars)
				.build();

		if (hiringMessage != null && !hiringMessage.isBlank()) {
			contextVars.put("hiringMessage", hiringMessage);
		}

		if (profile.getEmailNotifications()) {
			emailService.sendEmail(List.of(person.getEmail()), email);
		}
	}

	/**
	 * Send an application rejected notification
	 *
	 * @param personId         The ID of the person to send the notification to
	 * @param position         The position that the application was for
	 * @param course           The course that the application was for
	 * @param courseCode       The course code of the course that the application was for
	 * @param year             The name of the edition (eg. Q1 22/23)
	 * @param courseStaff      The names of the course staff (managers)
	 * @param rejectionMessage The rejection message from the course staff (can be null)
	 */
	public void sendApplicationRejectedNotification(Long personId, String position, String course,
			String courseCode, String year,
			String rejectionMessage, String courseStaff) {
		Profile profile = profileService.findByIdOrThrow(personId);
		PersonDetailsDTO person = personService.getPersonById(personId);

		Map<String, Object> contextVars = new HashMap<>(
				Map.of("name", person.getDisplayName(), "position", position, "course", course,
						"courseCode", courseCode, "year", year, "courseStaff", courseStaff));

		EmailCreateDTO email = EmailCreateDTO.builder()
				.subject("Application (" + courseCode + ") Was Not Selected")
				.template("html/application_rejected")
				.contextVariables(contextVars)
				.build();

		if (rejectionMessage != null && !rejectionMessage.isBlank()) {
			contextVars.put("rejectionMessage", rejectionMessage);
		}

		if (profile.getEmailNotifications()) {
			emailService.sendEmail(List.of(person.getEmail()), email);
		}
	}

	// end region
}
