/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.io.*;
import java.net.MalformedURLException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.util.*;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.core.io.Resource;
import org.springframework.core.io.UrlResource;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;

import com.opencsv.*;
import com.opencsv.exceptions.CsvValidationException;

import liquibase.util.BomAwareInputStream;
import nl.tudelft.tam.properties.ExportProperties;

@Service
public class CSVService {

	@Autowired
	private ExportProperties properties;

	/**
	 * Gets the response for sending a csv file.
	 *
	 * @param path the path to the file to be sent
	 */
	public ResponseEntity<Resource> getResponse(Path path) throws MalformedURLException {
		Resource resource = new UrlResource("file:" + path);
		return ResponseEntity.ok()
				.contentType(MediaType.parseMediaType("text/csv"))
				.header(HttpHeaders.CONTENT_DISPOSITION,
						"attachment; filename=\"" + path.getFileName() + "\"")
				.body(resource);
	}

	/**
	 * Reads data from a CSV file.
	 *
	 * @param  path                   the path to the file
	 * @return                        the data from the CSV file
	 * @throws IOException            if the file path is invalid or bad things happen during the read
	 * @throws CsvValidationException if a user defined validator fails
	 */
	public List<String[]> readCSV(Path path) throws IOException, CsvValidationException {
		return readCSV(path, false, 0);
	}

	/**
	 * Reads the data from a CSV file.
	 *
	 * @param  path                   the path to the file which should be read
	 * @param  ignoreQuotations       whether to ignore quotation marks in the parsing
	 * @param  skipLines              the number of lines to skip
	 * @return                        the data from the CSV file
	 * @throws IOException            if the file path is invalid or bad things happen during the read
	 * @throws CsvValidationException if a user defined validator fails
	 */
	public List<String[]> readCSV(Path path, boolean ignoreQuotations, int skipLines)
			throws IOException, CsvValidationException {
		return readCSV(
				new InputStreamReader(new BomAwareInputStream(new FileInputStream(String.valueOf(path)))),
				ignoreQuotations, skipLines);
	}

	/**
	 * Reads data from a CSV input.
	 *
	 * @param  reader                 the reader for the input
	 * @return                        the data from the CSV file
	 * @throws IOException            if the file path is invalid or bad things happen during the read
	 * @throws CsvValidationException if a user defined validator fails
	 */
	public List<String[]> readCSV(Reader reader) throws IOException, CsvValidationException {
		return readCSV(reader, false, 0);
	}

	/**
	 * Reads the data from a CSV input.
	 *
	 * @param  reader                 the reader for the input
	 * @param  ignoreQuotations       whether to ignore quotation marks in the parsing
	 * @param  skipLines              the number of lines to skip
	 * @return                        the data from the CSV file
	 * @throws IOException            if the file path is invalid or bad things happen during the read
	 * @throws CsvValidationException if a user defined validator fails
	 */
	public List<String[]> readCSV(Reader reader, boolean ignoreQuotations, int skipLines)
			throws IOException, CsvValidationException {
		List<String> content = new BufferedReader(reader).lines().toList();
		char separator = detectSeparator(content);
		reader = new StringReader(String.join("\n", content));

		// initialize the parser of the csv
		CSVParser parser = new CSVParserBuilder()
				.withSeparator(separator)
				.withIgnoreQuotations(ignoreQuotations)
				.build();

		// initialize the reader of the csv
		CSVReader csvReader = new CSVReaderBuilder(reader)
				.withCSVParser(parser)
				.withSkipLines(skipLines)
				.build();

		// read the data from the csv
		List<String[]> data = new ArrayList<>();
		String[] line;
		while ((line = csvReader.readNext()) != null) {
			data.add(line);
		}

		// close the readers
		reader.close();
		csvReader.close();

		return data;
	}

	/**
	 * Gets the separator with the least variation among the lines.
	 *
	 * @param  lines The lines in the file
	 * @return       The preferred separator
	 */
	private char detectSeparator(List<String> lines) {
		Map<Character, Double> averageOccurrences = new HashMap<>();
		Map<Character, List<Double>> amountOccurences = new HashMap<>();
		properties.getSeparators().forEach(s -> {
			averageOccurrences.put(s, 0.0);
			amountOccurences.put(s, new ArrayList<>());
		});
		for (String line : lines) {
			for (char sep : properties.getSeparators()) {
				double count = (double) line.chars().filter(c -> c == sep).count();
				averageOccurrences.put(sep, averageOccurrences.get(sep) + count / lines.size());
				amountOccurences.get(sep).add(count);
			}
		}
		return properties.getSeparators().stream().filter(s -> averageOccurrences.get(s) > 0.0)
				.min(Comparator.comparingDouble(s -> {
					// Sum of squared distances
					double mu = averageOccurrences.get(s);
					return amountOccurences.get(s).stream().mapToDouble(n -> Math.pow(mu - n, 2)).sum();
				})).orElse(properties.getSeparators().get(0));
	}

	/**
	 * Writes data to a csv file.
	 *
	 * @param  data        the data to be written
	 * @param  directory   the path to the file directory
	 * @param  fileName    the name of the file
	 * @throws IOException if the filename is wrong or an error occurs during closing of the writer
	 */
	public void writeToCSV(List<String[]> data, Path directory, String fileName) throws IOException {
		if (!directory.toFile().exists()) {
			directory.toFile().mkdirs();
		}

		File file = Path.of(directory.toString(), fileName).toFile();
		FileOutputStream fos = new FileOutputStream(file);
		// Add BOM Characters, Makes this UTF-8-SIG, but that allows cyrillic + other special (Romanian) characters
		fos.write(0xef);
		fos.write(0xbb);
		fos.write(0xbf);

		CSVWriter writer = new CSVWriter(
				new OutputStreamWriter(fos, StandardCharsets.UTF_8),
				properties.getSeparators().get(0),
				CSVWriter.DEFAULT_QUOTE_CHARACTER,
				CSVWriter.DEFAULT_ESCAPE_CHARACTER,
				CSVWriter.DEFAULT_LINE_END);
		writer.writeAll(data);
		writer.close();
	}

	/**
	 * Converts a encoded CSV file as part of a multipart file to a list of string arrays, whilst storing the
	 * file in the import directory.
	 *
	 * @param  file                   The multipartfile that should be a CSV File.
	 * @param  importDirectory        The directory where the file should be stored.
	 * @return                        A list of string arrays, where each array represents a line in the CSV
	 *                                file.
	 * @throws IOException            If the file cannot be stored.
	 * @throws CsvValidationException If the CSV format is invalid.
	 */
	public List<String[]> storeAndReadCSV(MultipartFile file, String importDirectory)
			throws IOException, CsvValidationException {
		Path directory = Path.of(importDirectory);

		if (!directory.toFile().exists()) {
			directory.toFile().mkdirs();
		}

		file.transferTo(Path.of(importDirectory + "/" + file.getOriginalFilename()));

		return readCSV(Path.of(importDirectory + "/" + file.getOriginalFilename()), false, 1);
	}

}
