/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static java.util.stream.Collectors.groupingBy;

import java.io.IOException;
import java.nio.file.Path;
import java.time.LocalDateTime;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.transaction.Transactional;
import javax.validation.constraints.NotNull;

import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.dto.CourseDetailsDTO;
import nl.tudelft.labracore.api.dto.EditionDetailsDTO;
import nl.tudelft.labracore.api.dto.PersonDetailsDTO;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.DeclarationCreateDTO;
import nl.tudelft.tam.dto.create.EmailCreateDTO;
import nl.tudelft.tam.dto.view.details.DeclarationDetailsExtraWorkDTO;
import nl.tudelft.tam.dto.view.details.DeclarationDetailsPersonDTO;
import nl.tudelft.tam.dto.view.details.ExtraWorkDetailsDTO;
import nl.tudelft.tam.dto.view.summary.DeclarationSummaryDTO;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.exception.ApplicationStatusException;
import nl.tudelft.tam.model.*;
import nl.tudelft.tam.model.records.FlexDelftExport;
import nl.tudelft.tam.repository.DeclarationChangeEventRepository;
import nl.tudelft.tam.repository.DeclarationExportRepository;
import nl.tudelft.tam.repository.DeclarationRepository;

@Service
public class DeclarationService {

	// region class attributes

	@Value("${tam.filesys.exports-dir}")
	String exportDir;

	@Autowired
	DeclarationRepository declarationRepository;

	@Autowired
	DeclarationChangeEventRepository declarationChangeEventRepository;

	@Autowired
	EditionService editionService;

	@Autowired
	PersonService personService;

	@Autowired
	CourseService courseService;

	@Autowired
	FileService fileService;

	@Autowired
	CSVService csvService;

	@Autowired
	ExtraWorkService extraWorkService;

	@Autowired
	CoordinatorService coordinatorService;

	@Autowired
	CohortService cohortService;

	@Autowired
	ProfileService profileService;

	@Autowired
	CoordinatorDefaultService coordinatorDefaultService;

	@Autowired
	ModelMapper mapper;

	@Autowired
	EmailService emailService;
	@Autowired
	private DeclarationExportRepository declarationExportRepository;

	// endregion

	// region manipulate data

	/**
	 * Save a new declaration into the database. In case one already exists with status rejected by student,
	 * changes status to submitted.
	 *
	 * @param  createDTO                  the dto create object of the declaration
	 * @throws ApplicationStatusException if open declaration already exists
	 */
	@Transactional
	public void submit(DeclarationCreateDTO createDTO) {
		Double remainingHours = numberOfHoursRemainingForExtraWork(createDTO.getPersonId(),
				createDTO.getWorkId());
		Double amountDeclared = createDTO.getAmount() * 0.25;

		if (amountDeclared > remainingHours) {
			throw new IllegalArgumentException("Declaration > Remaining Hours");
		}

		createDTO.setStatus(Status.SUBMITTED);
		Declaration dec = createDTO.apply();
		declarationRepository.save(dec);
		declarationChangeEventRepository
				.save(DeclarationChangeEvent.builder().declaration(dec).status(Status.SUBMITTED).build());
	}

	/**
	 * Retract the declaration. Delete declaration.
	 *
	 * @param id the id of the declaration
	 */
	@Transactional
	public void retract(Long id) {
		declarationRepository.deleteById(id);
	}

	/**
	 * Reject the offer. Change status to rejected by teacher.
	 *
	 * @param id      the id of the declaration
	 * @param handler the id of the rejecting person (ie. teacher)
	 */
	@Transactional
	public void reject(Long id, Long handler) {
		Declaration declaration = declarationRepository.findByIdOrThrow(id);
		declaration.setHandler(profileService.findByIdOrThrow(handler));
		setStatus(declaration, Status.REJECTED_BY_TEACHER);

		// Send notification
		EditionDetailsDTO edition = editionService.getEditionById(declaration.getExtraWork().getEditionId());
		String hours = Double.valueOf(declaration.getAmount() / 4.0).toString();
		sendDeclarationRejectedNotification(declaration.getPerson().getId(), hours,
				declaration.getExtraWork().getName(), edition.getCourse().getName(),
				edition.getCourse().getCode(), edition.getName());
	}

	/**
	 * Accept the offer. Change status to accepted.
	 *
	 * @param  id                         the id of the declaration
	 * @param  handler                    the id of the rejecting person (ie. teacher)
	 * @throws ApplicationStatusException if status is not "submitted"
	 */
	@Transactional
	public void approve(Long id, Long handler) {
		Declaration declaration = declarationRepository.findByIdOrThrow(id);
		declaration.setHandler(profileService.findByIdOrThrow(handler));
		setStatus(declaration, Status.ACCEPTED);

		// Send notification
		EditionDetailsDTO edition = editionService.getEditionById(declaration.getExtraWork().getEditionId());
		String hours = Double.valueOf(declaration.getAmount() / 4.0).toString();
		sendDeclarationApprovedNotification(declaration.getPerson().getId(), hours,
				declaration.getExtraWork().getName(), edition.getCourse().getName(),
				edition.getCourse().getCode(), edition.getName());
	}

	/**
	 * All applications in progress (status submitted or offered) for a job offer will be terminated.
	 *
	 * @param id      of the offer to be terminated.
	 * @param handler the id of the rejecting person (ie. teacher)
	 */
	@Transactional
	public void rejectAllOpenInExtraWork(Long id, Long handler) {
		ExtraWork extraWork = extraWorkService.findByIdOrThrow(id);
		EditionDetailsDTO edition = editionService.getEditionById(extraWork.getEditionId());

		extraWork.getDeclarations().forEach(d -> {
			if (!d.getStatus().handled()) {
				d.setHandler(profileService.findByIdOrThrow(handler));
				setStatus(d, Status.REJECTED_BY_TEACHER);

				// Send notification
				String hours = Double.valueOf(d.getAmount() / 4.0).toString();
				sendDeclarationRejectedNotification(d.getPerson().getId(), hours,
						extraWork.getName(), edition.getCourse().getName(),
						edition.getCourse().getCode(), edition.getName());
			}
		});
	}

	/**
	 * All applications in progress (status submitted or offered) for a job offer will be approved.
	 *
	 * @param id      of the offer to be approved.
	 * @param handler the id of the rejecting person (ie. teacher)
	 */
	@Transactional
	public void approveAllOpenInExtraWork(Long id, Long handler) {
		ExtraWork extraWork = extraWorkService.findByIdOrThrow(id);
		EditionDetailsDTO edition = editionService.getEditionById(extraWork.getEditionId());

		extraWork.getDeclarations().forEach(d -> {
			if (!d.getStatus().handled()) {
				d.setHandler(profileService.findByIdOrThrow(handler));
				setStatus(d, Status.ACCEPTED);

				// Send notification
				String hours = Double.valueOf(d.getAmount() / 4.0).toString();
				sendDeclarationApprovedNotification(d.getPerson().getId(), hours,
						extraWork.getName(), edition.getCourse().getName(),
						edition.getCourse().getCode(), edition.getName());
			}
		});
	}

	// endregion

	// region checkers

	/**
	 * Checks if a person owns a declaration, by ids.
	 *
	 * @param  personId      id of the person
	 * @param  declarationId id of the declaration
	 * @return               true if person owns the declaration
	 */
	@NotNull
	public boolean personOwnsDeclaration(Long personId, Long declarationId) {
		return personId.equals(declarationRepository.findByIdOrThrow(declarationId).getPerson().getId());
	}

	/**
	 * Returns the number of hours remaining for extra work for a person
	 *
	 * @param  personId    The person for which to check the remaining hours
	 * @param  extraWorkId The extra work for which to check the remaining hours
	 * @return             The number of remaining hours
	 */
	public Double numberOfHoursRemainingForExtraWork(Long personId, Long extraWorkId) {
		Integer maxHours = extraWorkService.getById(extraWorkId).getMaxHours();

		Integer totalAmountDeclared = declarationRepository.findAllByPersonId(personId).stream()
				.filter(d -> d.getStatus() == Status.ACCEPTED || d.getStatus() == Status.SUBMITTED)
				.filter(d -> d.getExtraWork().getId().equals(extraWorkId))
				.map(d -> d.getAmount())
				.reduce(0, (a, b) -> a + b);

		return maxHours - (totalAmountDeclared * 0.25);
	}

	// endregion

	// region getters

	/**
	 * Returns the id of the extra work associated with this declaration.
	 *
	 * @param  id id the of declaration
	 * @return    id of the extra work
	 */
	@NotNull
	public Long getExtraWorkIdFromDeclaration(Long id) {
		return declarationRepository.findByIdOrThrow(id).getExtraWork().getId();
	}

	/**
	 * Retrieves a list of all declarations with job details of one user. Ignore retracted declarations.
	 *
	 * @param  id of the person
	 * @return    list of declarations
	 */
	public List<DeclarationDetailsExtraWorkDTO> getDeclarationsForPerson(Long id) {
		List<Declaration> decs = declarationRepository.findAllByPersonId(id);
		if (decs.size() == 0)
			return new ArrayList<>();
		Map<Long, EditionDetailsDTO> map = editionService
				.getEditionsById(
						decs.stream().map(d -> d.getExtraWork().getEditionId()).collect(Collectors.toList()))
				.stream().collect(Collectors.toMap(EditionDetailsDTO::getId, e -> e));
		List<DeclarationDetailsExtraWorkDTO> res = new ArrayList<>();
		for (Declaration d : decs) {
			DeclarationDetailsExtraWorkDTO dto = mapper.map(d, DeclarationDetailsExtraWorkDTO.class);
			ExtraWorkDetailsDTO extra = mapper.map(d.getExtraWork(), ExtraWorkDetailsDTO.class);
			extra.setEdition(map.get(extra.getEditionId()));
			dto.setExtraWork(extra);
			res.add(dto);
		}
		return res;
	}

	/**
	 * List of all declarations with person details of one job offer. Ignore retracted declarations.
	 *
	 * @param  id of the job offer
	 * @return    the list of declarations
	 */
	public List<DeclarationDetailsPersonDTO> getDeclarationsForExtraWork(Long id) {
		List<DeclarationSummaryDTO> summaryDTOs = declarationRepository.findAllByWorkId(id).stream()
				.map(DeclarationSummaryDTO::map).collect(Collectors.toList());
		if (summaryDTOs.isEmpty())
			return new ArrayList<>();
		Map<Long, PersonSummaryDTO> ppl = personService
				.getPeopleById(summaryDTOs.stream().map(DeclarationSummaryDTO::getPersonId)
						.collect(Collectors.toList()))
				.stream().collect(Collectors.toMap(PersonSummaryDTO::getId, p -> p));
		List<DeclarationDetailsPersonDTO> detailsDTOs = new ArrayList<>();
		summaryDTOs.forEach(a -> {
			DeclarationDetailsPersonDTO dto = mapper.map(a, DeclarationDetailsPersonDTO.class);
			dto.setPerson(ppl.get(a.getPersonId()));
			detailsDTOs.add(dto);
		});
		return detailsDTOs;
	}

	/**
	 * Gets all declarations for a person filtered by a search query
	 *
	 * @param  id    The person for whom to find the declarations
	 * @param  query The search query to filter by
	 * @return       A list of declarations matching the search query
	 */
	public List<DeclarationDetailsExtraWorkDTO> getFilteredDeclarationsForPerson(Long id, String query) {
		List<DeclarationDetailsExtraWorkDTO> declarations = getDeclarationsForPerson(id);

		if (query == null || query.isBlank())
			return declarations;
		String f = query.toLowerCase();

		return declarations.stream()
				.filter(d -> d.getExtraWork().getEdition().getCourse().getName().toLowerCase().contains(f)
						|| d.getExtraWork().getEdition().getCourse().getCode().toLowerCase().contains(f)
						|| d.getExtraWork().getEdition().getCohort().getName().toLowerCase().contains(f)
						|| d.getExtraWork().getEdition().getName().toLowerCase().contains(f)
						|| d.getExtraWork().getName().toLowerCase().contains(f)
						|| d.getStatus().name().toLowerCase().contains(f))
				.collect(Collectors.toList());
	}

	/**
	 * Gets all declarations for some extra work filtered by a search query
	 *
	 * @param  id    The id of the extra work
	 * @param  query The search query to filter by
	 * @return       A list of declarations matching the search query
	 */
	public List<DeclarationDetailsPersonDTO> getFilteredDeclarationsForExtraWork(Long id, String query) {
		List<DeclarationDetailsPersonDTO> declarations = getDeclarationsForExtraWork(id);

		if (query == null || query.isBlank())
			return declarations;
		String f = query.toLowerCase();

		return declarations.stream()
				.filter(d -> d.getPerson().getDisplayName().toLowerCase().contains(f)
						|| d.getPerson().getUsername().toLowerCase().contains(f)
						|| d.getPerson().getExternalId().toLowerCase().contains(f)
						|| d.getPerson().getNumber().toString().contains(f)
						|| d.getPerson().getEmail().toLowerCase().contains(f)
						|| d.getStatus().name().toLowerCase().contains(f))
				.collect(Collectors.toList());
	}

	/**
	 * Returns stats about the declarations. Stat 1 is the total number of submitted declarations, stat 2 is
	 * the total number of declared hours, stat 3 is the total number of approved hours, stat 4 is the total
	 * number of reject hours.
	 *
	 * @param  declarations the list of declarations
	 * @return              array of integers with length of 4
	 */
	public <D extends DeclarationSummaryDTO> int[] getStatsForDecs(List<D> declarations) {
		int[] stats = new int[4];
		for (D d : declarations) {
			stats[0] += 1;
			stats[1] += d.getAmount();
			if (d.getStatus().equals(Status.ACCEPTED))
				stats[2] += d.getAmount();
			if (d.getStatus().equals(Status.REJECTED_BY_TEACHER))
				stats[3] += d.getAmount();
		}
		return stats;
	}

	/**
	 * Computes the number of declarations that still need to be handled.
	 *
	 * @param  workId the id of the extra work with declarations
	 * @return        the total number of hours that have not been approved
	 */
	public float getNumberOfUnhandledHours(Long workId) {
		List<DeclarationDetailsPersonDTO> decs = getDeclarationsForExtraWork(workId);
		int[] stats = getStatsForDecs(decs);
		// convert from quarters into hours
		return (stats[1] - stats[2] - stats[3]) / 4.0f;
	}

	/**
	 * Gets all the declarations a person coordinates.
	 *
	 * @param  personId The id of the person
	 * @return          A list of all declarations the person coordinates
	 */
	public List<Declaration> getCoordinatingDeclarations(Long personId) {
		List<Long> programIds = coordinatorService.findAllByPersonId(personId).stream()
				.map(Coordinator::getProgramId).collect(Collectors.toList());
		return declarationRepository.findAll().stream()
				.filter(e -> programIds.contains(courseService
						.getOrThrow(
								editionService.getOrThrow(e.getExtraWork().getEditionId()).getCourse()
										.getId())
						.getProgram().getId()))
				.collect(Collectors.toList());
	}

	/**
	 * Gets all the declarations a person coordinates.
	 *
	 * @param  personId  The id of the person
	 * @param  programId The id of the program
	 * @return           A list of all declarations the person coordinates in the program
	 */
	public List<Declaration> getCoordinatingDeclarationsForProgram(Long personId, Long programId) {
		List<Declaration> declarations = getCoordinatingDeclarations(personId);
		Map<Long, EditionDetailsDTO> editions = editionService
				.getEditionsById(
						declarations.stream().map(a -> a.getExtraWork().getEditionId()).distinct().toList())
				.stream().collect(Collectors.toMap(EditionDetailsDTO::getId, Function.identity()));
		if (!declarations.isEmpty()) {
			Map<Long, CourseDetailsDTO> courses = courseService
					.getCoursesById(
							editions.values().stream().map(e -> e.getCourse().getId()).distinct().toList())
					.stream().collect(Collectors.toMap(CourseDetailsDTO::getId, Function.identity()));
			return declarations.stream().filter(a -> programId.equals(courses
					.get(editions.get(a.getExtraWork().getEditionId()).getCourse().getId()).getProgram()
					.getId()))
					.toList();
		} else {
			return List.of();
		}
	}

	/**
	 * Find all declarations by job offer
	 *
	 * @param  id The job offer id
	 * @return    The list of declarations belonging to that job offer
	 */
	public List<Declaration> findAllByWorkId(Long id) {
		return declarationRepository.findAllByWorkId(id);
	}

	/**
	 * Finds all the exports a person did for a program.
	 *
	 * @param  person    The person
	 * @param  programId The id of the program
	 * @return           All exports done by the person
	 */
	public List<DeclarationExport> getExportsForPersonAndProgram(Person person, Long programId) {
		return declarationExportRepository.findAllByExportedByAndProgramIdOrderByExportAtDesc(person.getId(),
				programId);
	}

	/**
	 * Filter a list of declarations to only declarations changed since a timestamp.
	 *
	 * @param  declarations The list of declarations
	 * @param  since        The timestamp
	 * @return              The filtered list of declarations
	 */
	public List<Declaration> filterDeclarationsChangedSince(List<Declaration> declarations,
			LocalDateTime since) {
		return declarations.stream()
				.filter(app -> !declarationChangeEventRepository
						.findByDeclarationIdAndChangedAtAfter(app.getId(), since).isEmpty())
				.toList();
	}

	// endregion

	// region export

	/**
	 * Exports a list of declarations.
	 *
	 * @param  name         The name of the file
	 * @param  declarations The list of declarations
	 * @return              The Path to the export file
	 */
	public Path exportDeclarations(String name, List<Declaration> declarations) throws IOException {
		return exportDeclarations(name, declarations, false, true);
	}

	/**
	 * Exports a list of declarations.
	 *
	 * @param  name              The name of the file
	 * @param  declarations      The list of declarations
	 * @param  specificExtraWork Determines if this is a mixture of extra work or not
	 * @param  aggregate         Determines if aggregation of declarations should occur (by student and
	 *                           ExtraWork, and status)
	 * @return                   The Path to the export file
	 */
	public Path exportDeclarations(String name, List<Declaration> declarations, boolean specificExtraWork,
			boolean aggregate) throws IOException {
		// Fetch all people and editions to cache and avoid many requests to core
		personService.get(declarations.stream().map(d -> d.getPerson().getId()));
		editionService.get(
				declarations.stream().map(d -> d.getExtraWork().getEditionId()).collect(Collectors.toList()));

		String[] headers;
		if (specificExtraWork) {
			headers = new String[] { "Student Number", "Email", "Name", "PayScale", "Hours",
					"Status" };
		} else {
			headers = new String[] { "Student Number", "Email", "Name", "Baan Code", "Edition",
					"Contract Name", "PayScale", "Hours", "Status" };
		}

		if (aggregate) {
			record DeclarationAggregate(Profile profile, Status status, ExtraWork extraWork) {
			}

			// GroupBy Profile, Status, and Extra-Work (ExtraWork doesn't matter when specificExtraWork true)
			Map<DeclarationAggregate, List<Declaration>> groupedDeclarationMaps = declarations
					.stream()
					.collect(groupingBy(
							declaration -> new DeclarationAggregate(declaration.getPerson(),
									declaration.getStatus(), declaration.getExtraWork())));

			// Create "Grouped Declarations"
			declarations = groupedDeclarationMaps.entrySet().stream()
					.map(dec -> Declaration.builder()
							.person(dec.getKey().profile())
							.status(dec.getKey().status())
							.amount(dec.getValue().stream().map(Declaration::getAmount).reduce(0,
									Integer::sum))
							.extraWork(dec.getKey().extraWork())
							.build())
					.collect(Collectors.toList());
		}

		Stream<String[]> data = declarations.stream().map(dec -> {
			Profile declarantProfile = dec.getPerson();

			PersonSummaryDTO declarant = personService.getOrThrow(dec.getPerson().getId());
			EditionDetailsDTO edition = editionService.getOrThrow(dec.getExtraWork().getEditionId());
			String hours = String.valueOf(dec.getAmount() * 0.25);

			if (specificExtraWork) {
				return new String[] { Objects.requireNonNull(declarant.getNumber()).toString(),
						declarant.getEmail(),
						declarant.getDisplayName(), declarantProfile.getPayScale().toString(),
						hours, dec.getStatus().toString() };
			} else {
				return new String[] { Objects.requireNonNull(declarant.getNumber()).toString(),
						declarant.getEmail(),
						declarant.getDisplayName(), dec.getExtraWork().getBaanCode(),
						edition.getCourse().getName() + " - " + edition.getName(),
						dec.getExtraWork().getContractName(),
						declarantProfile.getPayScale().toString(),
						hours, dec.getStatus().toString() };
			}
		}).sorted(Comparator.comparing(o -> o[0]));

		String fileName = fileService.toFileName(name) + ".csv";
		csvService.writeToCSV(Stream.concat(Stream.<String[]>of(headers), data).collect(Collectors.toList()),
				Path.of(exportDir), fileName);

		return Path.of(exportDir, fileName);
	}

	/**
	 * Exports a list of declarations with FlexDelft specific fields
	 *
	 * @param  person       The person exporting
	 * @param  programId    The program being exported for
	 * @param  batch        The batch number
	 * @param  name         The name of the file
	 * @param  declarations The list of declarations
	 * @param  since        Export changes since a timestamp, null for all applications
	 * @return              The Path to the export file
	 */
	public Path exportDeclarationsForFlexDelft(Person person, Long programId, Integer batch, String name,
			List<Declaration> declarations, LocalDateTime since)
			throws IOException {
		if (since != null) {
			declarations = filterDeclarationsChangedSince(declarations, since);
		}

		declarationExportRepository.save(DeclarationExport.builder().exportedBy(person.getId())
				.programId(programId).batchNumber(batch).build());

		// Fetch all people and editions to cache and avoid many requests to core
		personService.get(declarations.stream().map(d -> d.getPerson().getId()));
		editionService.get(
				declarations.stream().map(d -> d.getExtraWork().getEditionId()).collect(Collectors.toList()));

		String[] headers;

		headers = new String[] { "Name", "Email", "Course Code", "Contract Name", "Baan Code",
				"Approver Name",
				"Approver Email",
				"Coordinator", "Max Hours", "Payscale", "Batch", "Remarks" };

		record PersonAndExtraWork(Profile person, ExtraWork extraWork) {
		}
		Stream<String[]> data = declarations
				.stream()
				.collect(Collectors.groupingBy(
						dec -> new PersonAndExtraWork(dec.getPerson(), dec.getExtraWork()),
						Collectors.summingDouble(dec -> dec.getAmount() * 0.25)))
				.entrySet()
				.stream()
				.map(entry -> {
					Profile declarantProfile = entry.getKey().person();
					Double hours = entry.getValue();
					PersonSummaryDTO declarant = personService.getOrThrow(declarantProfile.getId());
					ExtraWork decExtraWork = entry.getKey().extraWork();
					EditionDetailsDTO edition = editionService.getOrThrow(decExtraWork.getEditionId());
					CoordinatorDefault coordDefault = coordinatorDefaultService.getOrCreate(
							cohortService.getOrThrowById(edition.getCohort().getId()).getProgram().getId());

					FlexDelftExport export = new FlexDelftExport(declarant, declarantProfile, decExtraWork,
							edition, coordDefault, hours, batch.toString(), "");
					return export.toExport();
				});

		String fileName = fileService.toFileName(name) + ".csv";

		csvService.writeToCSV(Stream.concat(Stream.<String[]>of(headers), data).collect(Collectors.toList()),
				Path.of(exportDir), fileName);

		return Path.of(exportDir, fileName);
	}

	// endregion

	// region notifications

	/**
	 * Send a notification for a rejected declaration
	 *
	 * @param personId   The person to send the notification to
	 * @param hours      The number of hours in the rejected declaration
	 * @param extraWork  The name of the extra work
	 * @param course     The course for which the extra work was done
	 * @param courseCode The course code of the course for which the extra work was done
	 * @param year       The name of the edition (eg. Q1 22/23)
	 */
	public void sendDeclarationRejectedNotification(Long personId, String hours, String extraWork,
			String course, String courseCode, String year) {
		Profile profile = profileService.findByIdOrThrow(personId);
		PersonDetailsDTO person = personService.getPersonById(personId);

		EmailCreateDTO email = EmailCreateDTO.builder()
				.subject("Declaration (" + courseCode + ") Rejected")
				.template("html/declaration_rejected")
				.contextVariables(
						Map.of("name", person.getDisplayName(), "hours", hours, "extraWork", extraWork,
								"course", course, "courseCode", courseCode, "year", year))
				.build();

		if (profile.getEmailNotifications()) {
			emailService.sendEmail(List.of(person.getEmail()), email);
		}
	}

	/**
	 * Send a notification for a approved declaration
	 *
	 * @param personId   The person to send the notification to
	 * @param hours      The number of hours in the approved declaration
	 * @param extraWork  The name of the extra work
	 * @param course     The course for which the extra work was done
	 * @param courseCode The course code of the course for which the extra work was done
	 * @param year       The name of the edition (eg. Q1 22/23)
	 */
	public void sendDeclarationApprovedNotification(Long personId, String hours, String extraWork,
			String course, String courseCode, String year) {
		Profile profile = profileService.findByIdOrThrow(personId);
		PersonDetailsDTO person = personService.getPersonById(personId);

		EmailCreateDTO email = EmailCreateDTO.builder()
				.subject("Declaration (" + courseCode + ") Approved")
				.template("html/declaration_approved")
				.contextVariables(
						Map.of("name", person.getDisplayName(), "hours", hours, "extraWork", extraWork,
								"course", course, "courseCode", courseCode, "year", year))
				.build();

		if (profile.getEmailNotifications()) {
			emailService.sendEmail(List.of(person.getEmail()), email);
		}
	}

	// endregion

	/**
	 * Sets the status of a declaration and saves a change event.
	 *
	 * @param declaration The declaration
	 * @param status      The new status
	 */
	private void setStatus(Declaration declaration, Status status) {
		if (status == declaration.getStatus())
			return;

		declaration.setStatus(status);

		DeclarationChangeEvent event = DeclarationChangeEvent.builder().declaration(declaration)
				.status(status).build();
		declarationChangeEventRepository.save(event);
	}
}
