/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

import javax.validation.constraints.NotNull;

import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.EditionControllerApi;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.tam.cache.EditionCacheManager;
import nl.tudelft.tam.dto.view.details.EditionDetailsOffersWorkDTO;
import nl.tudelft.tam.dto.view.summary.ExtraWorkSummaryDTO;
import nl.tudelft.tam.dto.view.summary.JobOfferSummaryDTO;

@Service
public class EditionService {

	// region class attributes

	@Autowired
	ModelMapper mapper;

	@Lazy
	@Autowired
	JobOfferService jobOfferService;

	@Lazy
	@Autowired
	ExtraWorkService extraWorkService;

	@Autowired
	PersonService personService;

	@Autowired
	EditionControllerApi editionControllerApi;

	@Autowired
	EditionCacheManager editionCache;

	// endregion

	/**
	 * Adds all the students from a cohort to an edition
	 *
	 * @param editionId The edition to add the students to
	 * @param cohortId  The cohort from which to select the students to add
	 */
	public void addCohortToEdition(Long editionId, Long cohortId) {
		editionControllerApi.addCohortToEdition(editionId, cohortId);
	}

	/**
	 * Adds an edition.
	 *
	 * @param  edition The edition to create
	 * @return         The id of the created edition
	 */
	public Long addEdition(EditionCreateDTO edition) {
		return editionControllerApi.addEdition(edition).block();
	}

	// region getters

	/**
	 * Groups job offers and extra work by editions and returns the list of filled editions.
	 *
	 * @param  offers the list with job offers
	 * @param  work   the list with extra work
	 * @return        list of editions
	 */
	public List<EditionDetailsOffersWorkDTO> combineOffersAndWorkWithEditions(List<JobOfferSummaryDTO> offers,
			List<ExtraWorkSummaryDTO> work) {
		Set<Long> ids = new HashSet<>();
		offers.forEach(o -> ids.add(o.getEditionId()));
		work.forEach(w -> ids.add(w.getEditionId()));
		if (ids.isEmpty())
			return new ArrayList<>();
		Map<Long, EditionDetailsOffersWorkDTO> editions = editionControllerApi
				.getEditionsById(new ArrayList<>(ids))
				.collectList().block().stream().collect(Collectors.toMap(EditionDetailsDTO::getId,
						e -> mapper.map(e, EditionDetailsOffersWorkDTO.class)));
		editions.values().forEach(e -> {
			e.setOffers(new ArrayList<>());
			e.setWork(new ArrayList<>());
		});
		offers.forEach(o -> editions.get(o.getEditionId()).getOffers().add(o));
		work.forEach(w -> editions.get(w.getEditionId()).getWork().add(w));
		return new ArrayList<>(editions.values());
	}

	/**
	 * Filters the list of editions by a query, if there is a match in the edition, the whole edition is
	 * included. If there is not a match in the edition, the job offers and extra work is filtered too.
	 *
	 * @param  editions the list of editions
	 * @param  filter   the query
	 * @return          filtered list of editions
	 */
	public List<EditionDetailsOffersWorkDTO> filter(List<EditionDetailsOffersWorkDTO> editions,
			String filter) {
		if (filter == null || filter.isBlank())
			return editions;
		String q = filter.toLowerCase();

		List<EditionDetailsOffersWorkDTO> filtered = new ArrayList<>();

		for (EditionDetailsOffersWorkDTO e : editions) {
			if (e.getName().toLowerCase().contains(q)
					|| e.getCohort().getName().toLowerCase().contains(q)
					|| e.getCourse().getName().toLowerCase().contains(q)
					|| e.getCourse().getCode().toLowerCase().contains(q)) {
				filtered.add(e);
				continue;
			}
			e.setOffers(e.getOffers().stream().filter(o -> o.getName().toLowerCase().contains(q))
					.collect(Collectors.toList()));
			e.setWork(e.getWork().stream().filter(w -> w.getName().toLowerCase().contains(q))
					.collect(Collectors.toList()));
			if (!(e.getOffers().isEmpty() && e.getWork().isEmpty()))
				filtered.add(e);
		}

		return filtered;
	}

	/**
	 * Returns all editions by ids filled with job offers and extra work. In comparison to
	 * `combineOffersAndWorkWithEditions()` this method returns also editions with no job offers or extra
	 * work.
	 *
	 * @param  editionIds the list of editions ids
	 * @return            list of editions
	 */
	public List<EditionDetailsOffersWorkDTO> getEditionsWithJobsAndWork(List<Long> editionIds) {
		if (editionIds.isEmpty())
			return Collections.emptyList();
		List<JobOfferSummaryDTO> offers = jobOfferService.getAllInEditions(editionIds);
		List<ExtraWorkSummaryDTO> work = extraWorkService.getAllInEditions(editionIds);
		Map<Long, EditionDetailsOffersWorkDTO> editions = editionControllerApi
				.getEditionsById(editionIds).collectList().block().stream()
				.collect(Collectors.toMap(EditionDetailsDTO::getId,
						e -> mapper.map(e, EditionDetailsOffersWorkDTO.class)));
		editions.values().forEach(e -> {
			e.setOffers(new ArrayList<>());
			e.setWork(new ArrayList<>());
		});
		offers.forEach(o -> editions.get(o.getEditionId()).getOffers().add(o));
		work.forEach(w -> editions.get(w.getEditionId()).getWork().add(w));
		return new ArrayList<>(editions.values());
	}

	/**
	 * Gets all edition ids
	 *
	 * @return The list of all edition ids
	 */
	public List<Long> getAllEditionIds() {
		return editionControllerApi.getAllEditions().map(EditionDetailsDTO::getId).collectList()
				.block();
	}

	/**
	 * Removes a person from an edition
	 *
	 * @param editionId The edition to remove the person from
	 * @param personId  The person to remove
	 */
	public void removePersonFromEdition(Long editionId, Long personId) {
		editionControllerApi.removePersonFromEdition(editionId, personId).block();
	}

	/**
	 * Gets the edition object or throws an error
	 *
	 * @param  id The id of the edition to find
	 * @return    The edition details object
	 */
	@NotNull
	public EditionDetailsDTO getOrThrow(Long id) {
		return editionCache.getOrThrow(id);
	}

	/**
	 * Finds all editions with provided ids
	 *
	 * @param  ids The editions to find
	 * @return     The list of edition details objects matching the ids
	 */
	public List<EditionDetailsDTO> getEditionsById(List<Long> ids) {
		ids = ids.stream().distinct().collect(Collectors.toList());
		return editionCache.get(ids);
	}

	/**
	 * Gets the edition object by id
	 *
	 * @param  id The id of the edition
	 * @return    The found editon object
	 */
	@NotNull
	public EditionDetailsDTO getEditionById(Long id) {
		return editionCache.getOrThrow(id);
	}

	/**
	 * Gets a list of editions with provided ids
	 *
	 * @param  ids The list of ids to look for
	 * @return     The list of edition details objects found
	 */
	public List<EditionDetailsDTO> get(List<Long> ids) {
		ids = ids.stream().distinct().collect(Collectors.toList());
		return editionCache.get(ids);
	}

	/**
	 * Gets a list of editions based on their course.
	 *
	 * @param  time the time at which the edition should be active
	 * @return      The list of edition summary objects found
	 */
	public List<EditionSummaryDTO> getAllActiveEditions(LocalDateTime time) {
		return editionControllerApi.getAllEditionsActiveAtDate(time).toStream().toList();
	}

	/**
	 * Gets a list of editions active during a period.
	 *
	 * @param  period the period at which the edition should be active
	 * @return        The list of edition summary objects found
	 */
	public List<EditionSummaryDTO> getAllActiveEditions(LocalDateTime start, LocalDateTime end) {
		return editionControllerApi.getAllEditionsActiveDuringPeriod(start, end).toStream().toList();
	}

	/**
	 * Gets a list of edition ids that a person is the manager of
	 *
	 * @param  personId The person to get the ids for
	 * @return          The list of edition ids
	 */
	public List<Long> getAllEditionIdsThatPersonIsAMangerOf(Long personId) {
		return personService.getRolesForPerson(personId)
				.stream().filter(r -> r.getType().equals(RoleEditionDetailsDTO.TypeEnum.TEACHER))
				.map(r -> Objects.requireNonNull(r.getEdition()).getId()).collect(Collectors.toList());
	}

	/**
	 * Get all edition ids for a course
	 *
	 * @param  courseId The course id
	 * @return          The list of edition ids
	 */
	public List<Long> getEditionsForCourse(Long courseId) {
		return editionControllerApi.getAllEditionsByCourse(courseId).map(EditionDetailsDTO::getId)
				.collect(Collectors.toList()).block();
	}

	// endregion
}
