/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.util.Optional;

import javax.validation.constraints.NotNull;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import nl.tudelft.tam.dto.patch.NotificationPatchDTO;
import nl.tudelft.tam.dto.patch.ShirtPatchDTO;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.repository.ProfileRepository;

@Service
public class ProfileService {

	@Autowired
	ProfileRepository repository;

	/**
	 * Finds the profile object by id
	 *
	 * @param  id The id to find
	 * @return    The object found
	 */
	@NotNull
	public Profile findByIdOrThrow(Long id) {
		return repository.findByIdOrThrow(id);
	}

	/**
	 * Finds a profile or creates one if one does not exist.
	 *
	 * @param  id The id to find
	 * @return    The profile
	 */
	@NotNull
	public Profile findByIdOrCreate(Long id) {
		return repository.findById(id).orElseGet(() -> createProfile(id));
	}

	/**
	 * Finds the profile object by id
	 *
	 * @param  id The id to find
	 * @return    The object found wrapped in an optional
	 */
	public Optional<Profile> findById(Long id) {
		return repository.findById(id);
	}

	/**
	 * Updates the polo information for a person
	 *
	 * @param personId The person for which to update the information
	 * @param polo     The polo information
	 */
	public void updatePoloData(Long personId, ShirtPatchDTO polo) {
		repository.save(polo.apply(repository.findByIdOrThrow(personId)));
	}

	/**
	 * Updates the notification preferences for a person
	 *
	 * @param personId                The person for which to update the information
	 * @param notificationPreferences The notification preferences
	 */
	public void updateNotificationPreferences(Long personId, NotificationPatchDTO notificationPreferences) {
		repository.save(notificationPreferences.apply(repository.findByIdOrThrow(personId)));
	}

	/**
	 * Create a profile for a non-existent user
	 *
	 * @param  id The id of the person
	 * @return    The profile which was saved
	 */
	public Profile createProfile(Long id) {
		return repository.save(Profile.builder().id(id).build());
	}

}
