/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

import javax.transaction.Transactional;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Service;

import nl.tudelft.labracore.api.ProgramControllerApi;
import nl.tudelft.labracore.api.dto.ProgramDetailsDTO;
import nl.tudelft.labracore.api.dto.ProgramSummaryDTO;
import nl.tudelft.labracore.lib.security.user.DefaultRole;
import nl.tudelft.tam.model.TAMProgram;
import nl.tudelft.tam.repository.CoordinatorRepository;
import nl.tudelft.tam.repository.TAMProgramRepository;
import nl.tudelft.tam.security.AuthorisationService;

@Service
public class ProgramService {

	@Autowired
	private ProgramControllerApi programApi;

	@Autowired
	private TAMProgramRepository programRepository;

	@Autowired
	private CoordinatorRepository coordinatorRepository;

	@Lazy
	@Autowired
	private AuthorisationService authorisationService;

	/**
	 * Get a program by id
	 *
	 * @param  id The id of the program to get
	 * @return    The ProgramDetailsDTO object
	 */
	public ProgramDetailsDTO get(Long id) {
		return programApi.getProgramById(id).block();
	}

	/**
	 * Gets a list of programs corresponding to a list of ids.
	 *
	 * @param  ids The program ids
	 * @return     The list of program details
	 */
	public List<ProgramDetailsDTO> getProgramsById(List<Long> ids) {
		if (ids.isEmpty())
			return Collections.emptyList();
		return programApi.getAllProgramsById(ids).collectList().block();
	}

	public List<ProgramSummaryDTO> getAllPrograms() {
		return programApi.getAllPrograms().collectList().block();
	}

	/**
	 * Gets all programs coordinated by a coordinator.
	 *
	 * @param  coordinatorId The id of the coordinator
	 *
	 * @return               All program details coordinated
	 */
	public List<ProgramDetailsDTO> getCoordinatingPrograms(Long coordinatorId) {
		return authorisationService.getAuthPerson().getDefaultRole() == DefaultRole.ADMIN
				? programApi.getAllProgramsById(
						programApi.getAllPrograms().map(ProgramSummaryDTO::getId).collectList().block())
						.collectList().block()
				: coordinatorRepository.findAllByPersonId(coordinatorId).stream()
						.map(c -> programApi.getProgramById(c.getProgramId()).block()).toList();
	}

	/**
	 * Returns a list of all the programs which are relevant to the person.
	 *
	 * @param  personId The person to get the programs for
	 * @return          The list of programs
	 */
	public List<ProgramSummaryDTO> getAllRelevantPrograms(Long personId) {
		return programApi.getAllRelevantPrograms(personId).collectList().block();
	}

	/**
	 * Gets a TAM programme, or creates one if one does not exist.
	 *
	 * @param  programId The ID of the program
	 * @return           The TAM programme
	 */
	@Transactional
	public TAMProgram getTAMProgram(Long programId) {
		return programRepository.findById(programId)
				.orElseGet(() -> programRepository.save(TAMProgram.builder().id(programId).build()));
	}

	/**
	 * Gets all the IDs of programmes advertised by the specified program.
	 *
	 * @param  programId The ID of the programme advertised by
	 * @return           The list of programme IDs advertised to by that program
	 */
	public List<Long> getProgramIdsAdvertisedToBy(Long programId) {
		return getTAMProgram(programId).getAdvertisesTo().stream().map(TAMProgram::getId).toList();
	}

	/**
	 * Gets all the IDs of programmes advertising to the specified program.
	 *
	 * @param  programId The ID of the programme advertised to
	 * @return           The list of programme IDs advertising to that program
	 */
	public List<Long> getProgramIdsAdvertisingTo(Long programId) {
		return getTAMProgram(programId).getAdvertisedToBy().stream().map(TAMProgram::getId).toList();
	}

	/**
	 * Advertises a programme to another programme.
	 *
	 * @param programId           The programme to advertise to
	 * @param advertisedProgramId The programme to advertise
	 */
	@Transactional
	public void addAdvertisedToProgram(Long programId, Long advertisedProgramId) {
		TAMProgram program = getTAMProgram(programId);
		TAMProgram advertisedProgram = getTAMProgram(advertisedProgramId);
		program.getAdvertisesTo().add(advertisedProgram);
		programRepository.save(program);
	}

	/**
	 * Removes an advertised programme from another programme.
	 *
	 * @param programId           The programme that was advertised to
	 * @param advertisedProgramId The programme that was advertised
	 */
	@Transactional
	public void removeAdvertisedFromProgram(Long programId, Long advertisedProgramId) {
		TAMProgram program = getTAMProgram(programId);
		program.setAdvertisesTo(program.getAdvertisesTo().stream()
				.filter(p -> !Objects.equals(p.getId(), advertisedProgramId)).collect(Collectors.toSet()));
		programRepository.save(program);
	}

}
