/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

import javax.validation.constraints.NotNull;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Lazy;
import org.springframework.core.io.Resource;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;

import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.api.dto.ProgramSummaryDTO;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.librador.dto.view.View;
import nl.tudelft.tam.dto.create.RaiseRequestCreateDTO;
import nl.tudelft.tam.dto.view.details.RaiseRequestDetailsDTO;
import nl.tudelft.tam.dto.view.summary.RaiseRequestSummaryDTO;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.exception.FileHandlingException;
import nl.tudelft.tam.exception.InvalidDatabaseEntryException;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.model.RaiseRequest;
import nl.tudelft.tam.repository.RaiseRequestRepository;

@Service
public class RaiseRequestService {

	@Autowired
	private RaiseRequestRepository raiseRequestRepository;

	@Autowired
	private FileService fileService;

	@Autowired
	private ProfileService profileService;

	@Autowired
	private PersonService personService;

	@Lazy
	@Autowired
	private ProgramService programService;

	@Value("${tam.filesys.payscale-proofs-dir}")
	private String PROOF_DIRECTORY;

	/**
	 * Validates that file uploaded as a proof of pay scale is a single pdf file.
	 *
	 * @param  fileMap               the file
	 * @throws FileHandlingException if file is not validated
	 */
	public MultipartFile validateFileSubmission(Map<String, MultipartFile> fileMap)
			throws FileHandlingException {
		if (fileMap.size() != 1) {
			throw new FileHandlingException("Attempted to submit more than one file in raise request proof.");
		}

		MultipartFile file = fileMap.values().iterator().next();
		String[] allowedTypes = { "image/", "application/pdf" };

		if (Arrays.stream(allowedTypes)
				.noneMatch(t -> Objects.requireNonNull(file.getContentType()).toLowerCase().contains(t))) {
			throw new FileHandlingException("Uploaded file is not allowed.");
		}

		return file;
	}

	/**
	 * Makes a new RaiseRequest entry in database.
	 *
	 * @param  raiseRequestCreateDTO the request to be created in database
	 * @param  file                  the file proof
	 * @return                       whether creating the request was successful
	 */
	public boolean newRaiseRequest(RaiseRequestCreateDTO raiseRequestCreateDTO, MultipartFile file) {
		// validate no other ongoing raise request are pending for this profile
		if (raiseRequestRepository.findByRequesterAndSubmitted(raiseRequestCreateDTO.getRequestedBy().getId())
				.isPresent()) {
			return false;
		}

		RaiseRequest request = raiseRequestRepository.save(raiseRequestCreateDTO.apply());
		try {
			fileService.uploadFile(file, request.getId(), PROOF_DIRECTORY);
		} catch (IOException e) {
			raiseRequestRepository.delete(request);
			e.printStackTrace();
			return false;
		}
		return true;
	}

	/**
	 * A getter for RaiseRequest object by its id.
	 *
	 * @param  id the id of the object to be returned
	 * @return    the RaiseRequest object with matching id
	 */
	@NotNull
	public RaiseRequest getBy(Long id) {
		return raiseRequestRepository.findByIdOrThrow(id);
	}

	/**
	 * Returns the file proof of a given raise request.
	 *
	 * @param  id          of the raise request
	 * @return             the file resource
	 * @throws IOException if an error occurs while getting the file to download
	 */
	public Resource downloadFile(Long id) throws IOException {
		return fileService.downloadFile(id, PROOF_DIRECTORY);
	}

	/**
	 * Retracts the last open claim for the change in PayScale for a person.
	 *
	 * @param person the person retracting their claim.
	 */
	public void retract(Person person) {
		raiseRequestRepository.findByRequesterAndSubmitted(person.getId())
				.ifPresent(raiseRequest -> raiseRequestRepository.delete(raiseRequest));
	}

	/**
	 * Approves a claim for change in PayScale.
	 *
	 * @param id         the id of the RaiseRequest to be accepted
	 * @param approvedBy the person approving the claim
	 */
	public void approve(Long id, Person approvedBy) {
		RaiseRequest raiseRequest = raiseRequestRepository.findByIdOrThrow(id);
		raiseRequest.getRequestedBy().setPayScale(raiseRequest.getPayScale());
		raiseRequest.setApprovedBy(profileService.findByIdOrThrow(approvedBy.getId()));
		raiseRequest.setStatus(Status.ACCEPTED);
	}

	/**
	 * Rejects a claim for change in PayScale.
	 *
	 * @param id         the id of the RaiseRequest to be rejected
	 * @param approvedBy the person rejecting the claim
	 */
	public void reject(Long id, Person approvedBy) {
		RaiseRequest raiseRequest = raiseRequestRepository.findByIdOrThrow(id);
		raiseRequest.setApprovedBy(profileService.findByIdOrThrow(approvedBy.getId()));
		raiseRequest.setStatus(Status.REJECTED_BY_TEACHER);
	}

	/**
	 * Return all raise requests with status submitted.
	 *
	 * @param  programIds list of program ids for which to get all submitted raise requests
	 * @return            list of found raise requests
	 */
	public List<RaiseRequestDetailsDTO> getAllSubmitted(List<Long> programIds)
			throws InvalidDatabaseEntryException {

		List<RaiseRequest> requests = raiseRequestRepository.findAllByStatus(Status.SUBMITTED);

		List<RaiseRequestDetailsDTO> list = View.convert(requests, RaiseRequestDetailsDTO.class);

		if (list.isEmpty())
			return list; // api.getPeopleById throw error on empty list request

		List<Profile> profiles = requests.stream().map(RaiseRequest::getRequestedBy)
				.collect(Collectors.toList());

		List<PersonSummaryDTO> people = personService.getPeopleById(
				profiles.stream().map(Profile::getId).collect(Collectors.toList()));

		if (list.size() != people.size() || people.size() != profiles.size()) {
			throw new InvalidDatabaseEntryException(
					"The list of raise requests contains an invalid request, " +
							"a matching person or profile was not found for at least one request.");
		}

		List<RaiseRequestDetailsDTO> filteredList = new ArrayList<>();

		for (int i = 0; i < list.size(); i++) {
			RaiseRequestDetailsDTO request = list.get(i);

			List<Long> relevantProgramIds = programService
					.getAllRelevantPrograms(people.get(i).getId())
					.stream()
					.map(ProgramSummaryDTO::getId)
					.collect(Collectors.toList());

			for (Long programId : relevantProgramIds) {
				if (programIds.contains(programId)) {
					request.setPerson(people.get(i));
					request.setCurrentPayScale(profiles.get(i).getPayScale());
					request.setRelevantProgramIds(relevantProgramIds);

					filteredList.add(request);
					break;
				}
			}

		}

		return filteredList;
	}

	/**
	 * Returns the last raise request for a person that is open.
	 *
	 * @param  person the person that submitted the raise request
	 * @return        the raise request summary dto
	 */
	public RaiseRequestSummaryDTO getOpenForPerson(Person person) {

		Optional<RaiseRequest> res = raiseRequestRepository.findByRequesterAndSubmitted(person.getId());

		if (res.isEmpty())
			return null;

		return RaiseRequestSummaryDTO.builder()
				.id(res.get().getId())
				.payScale(res.get().getPayScale())
				.status(res.get().getStatus())
				.build();
	}

	/**
	 * Import a PayScale for a student
	 *
	 * @param  rejector          The person rejecting the existing requests
	 * @param  studentIdString   The student for which to create the request
	 * @param  newPayScaleString The new PayScale
	 * @param  overwrite         If existing approvals should be ignored
	 *
	 * @return                   A list of conflicts that were skipped
	 */
	public List<AbstractMap.SimpleEntry<PersonSummaryDTO, Profile>> importPayScale(String studentIdString,
			String newPayScaleString,
			Person rejector,
			Boolean overwrite) {
		Long studentId = Long.parseLong(studentIdString);
		PayScale newPayScale = PayScale.valueOf(newPayScaleString);

		Profile profile = profileService.findByIdOrThrow(studentId);

		Optional<RaiseRequest> existingRequest = raiseRequestRepository
				.findByRequesterAndSubmitted(studentId);

		if (!overwrite && profile.getPayScale().isHigherThan(newPayScale)) {
			// Existing PayScale Higher => Add to conflicts
			return List.of(new AbstractMap.SimpleEntry<>(personService.getOrThrow(studentId), profile));
		}

		if (existingRequest.isPresent() && !existingRequest.get().getPayScale().equals(newPayScale)) {
			// Existing Raise Request => Don't add to lost of conflicts or overwrite, but skip!
			return List.of();
		}

		existingRequest.ifPresent(raiseRequest -> reject(raiseRequest.getId(), rejector));
		profile.setPayScale(newPayScale);

		return List.of();
	}

	/**
	 * Import PayScales for multiple students
	 *
	 * @param  rejector    The person rejecting the existing requests
	 * @param  identifiers The list of students
	 * @param  newPayScale The new payscale
	 *
	 * @return             A list of conflicts that were skipped
	 */
	public List<AbstractMap.SimpleEntry<PersonSummaryDTO, Profile>> bulkImportPayScales(String identifiers,
			String newPayScale,
			Person rejector) {
		List<PersonSummaryDTO> people = personService.searchForPeople(identifiers);
		List<AbstractMap.SimpleEntry<PersonSummaryDTO, Profile>> conflicts = new ArrayList<>();

		for (PersonSummaryDTO person : people) {
			conflicts.addAll(importPayScale(person.getId().toString(), newPayScale, rejector, false));
		}

		return conflicts;
	}
}
