/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import java.io.IOException;
import java.time.LocalDate;
import java.util.*;

import javax.transaction.Transactional;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;

import com.opencsv.exceptions.CsvValidationException;

import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.TrainingApprovalCreateDTO;
import nl.tudelft.tam.dto.id.ProfileIdDTO;
import nl.tudelft.tam.io.ImportHandler;
import nl.tudelft.tam.io.TrainingImportHandler;
import nl.tudelft.tam.model.TrainingApproval;
import nl.tudelft.tam.model.TrainingType;
import nl.tudelft.tam.repository.TrainingApprovalRepository;

@Service
@Transactional
public class TrainingApprovalService {

	@Autowired
	private TrainingApprovalRepository trainingApprovalRepository;

	@Autowired
	private TrainingTypeService trainingTypeService;

	@Autowired
	private ProfileService profileService;

	@Autowired
	private PersonService personService;

	@Autowired
	private CSVService csvService;

	@Autowired
	private TrainingImportHandler importHandler;

	@Value("${tam.filesys.imports-dir}")
	private String importDirectory;

	/**
	 * Create or overwrite a training approval for a student
	 *
	 * @param approver           The person creating the approval
	 * @param traineeIdString    The student id as a string
	 * @param trainingTypeString The type of training
	 * @param datePassedString   The date on which the training was completed
	 */
	@Transactional
	public void importTraining(Person approver, String traineeIdString, String trainingTypeString,
			String datePassedString) {
		ProfileIdDTO approvedId = new ProfileIdDTO(approver.getId());
		ProfileIdDTO traineeId = new ProfileIdDTO(Long.parseLong(traineeIdString));
		LocalDate datePassed = LocalDate.parse(datePassedString);
		TrainingType trainingType = trainingTypeService
				.getTrainingTypeById(Long.parseLong(trainingTypeString));

		Optional<TrainingApproval> existingApproval = trainingApprovalRepository
				.findByPersonIdAndType(traineeId.getId(), trainingType);

		if (existingApproval.isEmpty()) {
			TrainingApprovalCreateDTO createDTO = TrainingApprovalCreateDTO.builder()
					.approvedBy(approvedId)
					.person(traineeId)
					.type(trainingType)
					.datePassed(datePassed)
					.build();

			trainingApprovalRepository.save(createDTO.apply());
		} else {
			TrainingApproval existing = existingApproval.get();
			existing.setDatePassed(datePassed);
			existing.setApprovedBy(profileService.findByIdOrThrow(approver.getId()));
			trainingApprovalRepository.save(existing);
		}
	}

	/**
	 * Bulk import training approvals
	 *
	 * @param approver           The person importing
	 * @param allTraineeString   The list of students (newline separated)
	 * @param trainingTypeString The type of training
	 * @param datePassedString   The date on which they completed the training
	 */
	@Transactional
	public void bulkImportTraining(Person approver, String allTraineeString, String trainingTypeString,
			String datePassedString) {
		List<PersonSummaryDTO> trainees = personService.searchForPeople(allTraineeString);

		for (PersonSummaryDTO trainee : trainees) {
			importTraining(approver, trainee.getId().toString(), trainingTypeString, datePassedString);
		}
	}

	/**
	 * Delete all the training approvals for a training type
	 *
	 * @param trainingTypeId The ID of the training type for which to delete the approvals
	 */
	@Transactional
	public void deleteAllTrainingApprovalsForTrainingType(Long trainingTypeId) {
		trainingApprovalRepository.deleteAllByType(trainingTypeService.getTrainingTypeById(trainingTypeId));
	}

	/**
	 * Find a training approval by person id and type
	 *
	 * @param  personId The person to find the approval for
	 * @param  type     The training type to find the approval for
	 * @return          An optional of the training approval
	 */
	public Optional<TrainingApproval> findByPersonIdAndType(Long personId, TrainingType type) {
		return trainingApprovalRepository.findByPersonIdAndType(personId, type);
	}

	/**
	 * Find a training approval by person id and type id
	 *
	 * @param  personId The person to find the approval for
	 * @param  typeId   The training type to find the approval for
	 * @return          An optional of the training approval
	 */
	public Optional<TrainingApproval> findByPersonIdAndTypeId(Long personId, Long typeId) {
		return trainingApprovalRepository.findByPersonIdAndTypeId(personId, typeId);
	}

	/**
	 * Imports a list of training approvals from a CSV file. Format:<code>
	 *     [0] = assistant id, [1] = training type, [2] = date passed
	 * </code>
	 *
	 * @param  programId              The ID of the program to import for
	 * @param  file                   The CSV file containing the training approvals.
	 * @return                        A record containing successful, warning and failed imports.
	 * @throws IOException            If the file is invalid/corrupted.
	 * @throws CsvValidationException If the file format is invalid.
	 */
	@Transactional
	public ImportHandler.Result<TrainingApproval> importTraining(Long programId, MultipartFile file)
			throws IOException, CsvValidationException {

		List<String[]> lines = csvService.storeAndReadCSV(file, importDirectory);
		var result = importHandler.forProgram(programId).importFile(lines);
		trainingApprovalRepository.saveAll(result.result());

		return result;
	}
}
