/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import static org.mockito.Mockito.*;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithUserDetails;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.security.AuthorisationService;
import nl.tudelft.tam.service.ApplicationService;
import nl.tudelft.tam.service.PersonService;

@AutoConfigureMockMvc
@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
public class PersonControllerTest {

	@Autowired
	MockMvc mvc;

	@MockBean
	PersonService personService;

	@MockBean
	ApplicationService applicationService;

	@MockBean
	AuthorisationService authService;

	PersonSummaryDTO person1, person2;
	final long JOFFER_ID_1 = 1001;
	final long PERSON_ID_1 = 1;
	final long PERSON_ID_2 = 2;

	@BeforeEach
	void setup() {
		person1 = new PersonSummaryDTO().displayName("One").id(PERSON_ID_1).username("User1");
		person2 = new PersonSummaryDTO().displayName("Two").id(PERSON_ID_2).username("User2");
	}

	@Test
	@WithUserDetails("admin")
	void findCandidatesTest() throws Exception {
		when(authService.isManagerOfAny()).thenReturn(true);
		when(personService.searchForPeople(anyString())).thenReturn(List.of(person1, person2));
		when(applicationService.appExistsFor(anyLong(), anyLong())).thenReturn(false);

		String user = "test";

		mvc.perform(post("/person/find-candidate")
				.param("offerId", String.valueOf(JOFFER_ID_1))
				.param("identifiers", user).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(flash().attribute("specificOfferCandidates", List.of(person1, person2)))
				.andExpect(flash().attribute("specificOfferQuery", user))
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).isManagerOfAny();
		verify(personService).searchForPeople(user);
	}

	@Test
	@WithUserDetails("admin")
	void findCandidatesToRejectTest() throws Exception {
		when(authService.isManagerOfAny()).thenReturn(true);
		when(personService.searchForPeople(anyString())).thenReturn(List.of(person1, person2));
		when(applicationService.appExistsFor(anyLong(), anyLong())).thenReturn(false);

		String user = "test";

		mvc.perform(post("/person/find-candidate-to-reject")
						.param("offerId", String.valueOf(JOFFER_ID_1))
						.param("identifiers", user).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(flash().attribute("specificRejectCandidates", List.of(person1, person2)))
				.andExpect(flash().attribute("specificRejectQuery", user))
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).isManagerOfAny();
		verify(personService).searchForPeople(user);
	}

	@Test
	@WithUserDetails("admin")
	void searchForPeopleTest() throws Exception {
		Map<PersonSummaryDTO, Profile> entrymap = Map.ofEntries(Map.entry(person1, new Profile()),
				Map.entry(person2, new Profile()));
		when(authService.isCoordinatorOfAny()).thenReturn(true);
		when(personService.findPersonProfilePairs(anyString())).thenReturn(entrymap);

		String user = "test";

		mvc.perform(get("/person/search")
				.param("q", user))
				.andExpect(status().isOk()) // Expect HTTP 200 status
				.andExpect(model().attribute("people", entrymap)) // Expect the "people" attribute in the model
				.andExpect(view().name("person/search"));

		verify(authService).isCoordinatorOfAny();
		verify(personService).findPersonProfilePairs(user);
	}

	@Test
	@WithUserDetails("admin")
	void searchForPeopleWithoutIdentifierTest() throws Exception {
		when(authService.isCoordinatorOfAny()).thenReturn(true);

		mvc.perform(MockMvcRequestBuilders.get("/person/search"))
				.andExpect(status().isOk())
				.andExpect(model().attribute("people", Map.of()))
				.andExpect(view().name("person/search"));

		verify(authService).isCoordinatorOfAny();
		verify(personService, times(0)).findPersonProfilePairs(anyString());
		verify(personService).findPersonProfilePairs(null);
	}

	@Test
	@WithUserDetails("student")
	void searchForPeopleAsStudentTest() throws Exception {
		when(authService.isCoordinatorOfAny()).thenReturn(false);

		mvc.perform(MockMvcRequestBuilders.get("/person/search"))
				.andExpect(status().isForbidden());

		verify(authService).isCoordinatorOfAny();
		verify(personService, times(0)).findPersonProfilePairs(anyString());
	}

	@Test
	@WithUserDetails("admin")
	void findCandidatesWithStudentNumberTest() throws Exception {
		when(authService.isManagerOfAny()).thenReturn(true);
		when(personService.searchForPeople(anyString())).thenReturn(List.of(person1, person2));
		// Make sure that there is an application for person 1 and not for person 2
		// both should still be returned in the result.
		when(applicationService.appExistsFor(eq(PERSON_ID_1), anyLong())).thenReturn(true);

		String user = "12";

		mvc.perform(post("/person/find-candidate")
				.param("offerId", String.valueOf(JOFFER_ID_1))
				.param("identifiers", user).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(flash().attribute("specificOfferCandidates", List.of(person1, person2)))
				.andExpect(flash().attribute("specificOfferQuery", user))
				.andExpect(redirectedUrl("/job-offer/" + JOFFER_ID_1));

		verify(authService).isManagerOfAny();
		verify(personService).searchForPeople(user);
	}

}
