/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.controller;

import static org.mockito.Mockito.*;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.util.List;

import javax.transaction.Transactional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.MethodSource;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.security.test.context.support.WithUserDetails;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.request.MockHttpServletRequestBuilder;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;
import org.springframework.web.multipart.MultipartFile;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.lib.security.user.DefaultRole;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.RaiseRequestCreateDTO;
import nl.tudelft.tam.dto.id.ProfileIdDTO;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.repository.ProfileRepository;
import nl.tudelft.tam.security.AuthorisationService;
import nl.tudelft.tam.service.RaiseRequestService;
import nl.tudelft.tam.test.TestUserDetailsService;

@AutoConfigureMockMvc
@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class RaiseRequestControllerTest {

	@Autowired
	private MockMvc mvc;

	@Autowired
	private ProfileRepository profileRepository;

	@MockBean
	private RaiseRequestService raiseRequestService;

	@MockBean
	private AuthorisationService authorisationService;

	private Profile profile;

	@BeforeEach
	void setup() {
		profile = profileRepository.save(Profile.builder()
				.id(TestUserDetailsService.id)
				.build());
	}

	@Test
	@WithUserDetails("teacher")
	void getSubmittedRaiseRequests() throws Exception {
		when(authorisationService.canEditPayScale()).thenReturn(true);

		mvc.perform(get("/raise-request/all"))
				.andExpect(status().isOk())
				.andExpect(view().name("raise_request/submitted"))
				.andExpect(model().attributeExists("raiseRequests"));
	}

	@Test
	@WithUserDetails("student")
	void getSubmittedRaiseRequestsAccessDenied() throws Exception {
		mvc.perform(get("/raise-request/all"))
				.andExpect(status().is4xxClientError());
	}

	@Test
	@Transactional
	@WithUserDetails("student")
	void postChange() throws Exception {
		PayScale newScale = PayScale.SA4;
		String proof = "proof-13984";
		MultipartFile proofFile = new MockMultipartFile(proof, proof.getBytes());

		RaiseRequestCreateDTO requestDTO = RaiseRequestCreateDTO.builder().payScale(newScale).build();

		when(raiseRequestService.validateFileSubmission(anyMap())).thenReturn(proofFile);
		when(raiseRequestService.newRaiseRequest(any(RaiseRequestCreateDTO.class), any(MultipartFile.class)))
				.thenReturn(true);

		mvc.perform(MockMvcRequestBuilders.multipart("/raise-request/submit")
				.file(new MockMultipartFile("file", proof.getBytes())).with(csrf())
				.contentType(MediaType.MULTIPART_FORM_DATA_VALUE)
				.flashAttr("createRaiseRequestDTO", requestDTO))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/profile"));

		requestDTO.setRequestedBy(new ProfileIdDTO(profile.getId()));

		verify(raiseRequestService, times(1)).validateFileSubmission(anyMap());
		verify(raiseRequestService, times(1)).newRaiseRequest(eq(requestDTO), eq(proofFile));
	}

	@Test
	@Transactional
	@WithUserDetails("student")
	void retractChange() throws Exception {
		doNothing().when(raiseRequestService).retract(any(Person.class));

		mvc.perform(post("/raise-request/retract").with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/profile"));

		String username = "student";

		verify(raiseRequestService).retract(
				Person.builder()
						.id(profile.getId())
						.number(username.hashCode())
						.username(username)
						.displayName(username)
						.externalId(username + "@tudelft.nl")
						.email(username + "@tudelft.nl")
						.defaultRole(DefaultRole.STUDENT)
						.build());
	}

	@Test
	@WithUserDetails("teacher")
	void approveChange() throws Exception {
		doNothing().when(raiseRequestService).approve(anyLong(), any(Person.class));
		when(authorisationService.canEditPayScale()).thenReturn(true);

		Long id = 846815L;

		mvc.perform(post("/raise-request/approve/{id}", id).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/raise-request/all"));

		String username = "teacher";
		verify(raiseRequestService).approve(id, Person.builder()
				.id(profile.getId())
				.number(username.hashCode())
				.username(username)
				.displayName(username)
				.externalId(username + "@tudelft.nl")
				.email(username + "@tudelft.nl")
				.defaultRole(DefaultRole.TEACHER)
				.build());
	}

	@Test
	@WithUserDetails("teacher")
	void rejectChange() throws Exception {
		when(authorisationService.canEditPayScale()).thenReturn(true);
		doNothing().when(raiseRequestService).reject(anyLong(), any(Person.class));

		Long id = 1548992L;

		mvc.perform(post("/raise-request/reject/{id}", id).with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("/raise-request/all"));

		String username = "teacher";
		verify(raiseRequestService).reject(id, Person.builder()
				.id(profile.getId())
				.number(username.hashCode())
				.username(username)
				.displayName(username)
				.externalId(username + "@tudelft.nl")
				.email(username + "@tudelft.nl")
				.defaultRole(DefaultRole.TEACHER)
				.build());
	}

	@Test
	@WithUserDetails("teacher")
	void downloadProofFile() throws Exception {
		when(authorisationService.canDownloadProof(anyLong())).thenReturn(true);
		String proof = "proof-13984";
		MultipartFile proofFile = new MockMultipartFile(proof, proof.getBytes());

		when(raiseRequestService.downloadFile(anyLong())).thenReturn(proofFile.getResource());

		long PROOF_ID = 4685487L;

		mvc.perform(get("/raise-request/{assignmentId}/proof", PROOF_ID))
				.andExpect(status().isOk());

		verify(raiseRequestService, times(1)).downloadFile(PROOF_ID);
	}

	@ParameterizedTest
	@MethodSource(value = "protectedEndpoints")
	void testWithoutUserDetailsIsForbidden(MockHttpServletRequestBuilder request) throws Exception {
		mvc.perform(request.with(csrf()))
				.andExpect(status().is3xxRedirection())
				.andExpect(redirectedUrl("http://localhost/login"));
	}

	private static List<MockHttpServletRequestBuilder> protectedEndpoints() {
		return List.of(
				get("/raise-request"),
				post("/raise-request/submit"),
				delete("/raise-request/retract"),
				post("/raise-request/approve"),
				post("/raise-request/reject"),
				post("/raise-request/1/proof"));
	}

}
