/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.io;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.Mockito.when;

import java.time.LocalDate;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithUserDetails;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.tam.model.TrainingApproval;
import nl.tudelft.tam.model.TrainingType;
import nl.tudelft.tam.service.PersonService;
import nl.tudelft.tam.service.TrainingApprovalService;
import nl.tudelft.tam.service.TrainingTypeService;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class TrainingImportHandlerTest {

	@MockBean
	private PersonService personService;

	@MockBean
	private TrainingApprovalService trainingApprovalService;

	@MockBean
	private TrainingTypeService trainingTypeService;

	private TrainingImportHandler importHandler;

	Long programId;

	TrainingType t1;

	TrainingType t2;

	@BeforeEach
	void setup() {
		programId = 1L;

		t1 = TrainingType.builder().id(1L).name("Regular Training").programId(programId).build();
		t2 = TrainingType.builder().id(2L).name("Team Training").programId(programId).build();

		when(personService.searchForPeople("csestudent1")).thenReturn(List.of(new PersonSummaryDTO().id(1L)));
		when(personService.searchForPeople("csestudent2")).thenReturn(List.of(new PersonSummaryDTO().id(2L)));

		when(trainingTypeService.getTrainingTypesByNameAndProgramId("Regular Training", programId))
				.thenReturn(List.of(t1));
		when(trainingTypeService.getTrainingTypesByNameAndProgramId("Team Training", programId))
				.thenReturn(List.of(t2));

		when(trainingApprovalService.findByPersonIdAndType(anyLong(), any())).thenReturn(Optional.empty());

		this.importHandler = new TrainingImportHandler(personService, trainingApprovalService,
				trainingTypeService)
				.forProgram(programId);
	}

	@Test
	@WithUserDetails("cseteacher1")
	void importNormal() {

		when(personService.searchForPeople("csestudent3")).thenReturn(List.of(new PersonSummaryDTO().id(3L)));
		String content = """
				csestudent1,Regular Training,26-02-2003
				csestudent2,Team Training,31-04-2018
				csestudent3,Regular Training,31-06-2018
				bogusEntry, Regular Training, 31-06-2018, bogus
				""".stripIndent();

		ImportHandler.Result<TrainingApproval> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.warnings()).isEmpty();
		assertThat(res.errors()).hasSize(1);
		assertThat(res.result()).hasSize(3);
		assertThat(res.errors()).containsEntry(4, List.of(new SimpleImportError("Wrong amount of entries")));
		assertThat(res.result().get(0).getType()).isEqualTo(t1);
		assertThat(res.result().get(0).getDatePassed()).isEqualTo(LocalDate.of(2003, 2, 26));
		assertThat(res.result().stream().map(x -> x.getPerson().getId()).distinct().toList())
				.containsExactlyInAnyOrder(1L, 2L, 3L);

	}

	@Test
	@WithUserDetails("cseteacher1")
	void importWhereThirdPersonIsNotFound() {

		when(personService.searchForPeople("csestudent3")).thenReturn(Collections.emptyList());

		String content = """
				csestudent1,Regular Training,26-02-2003
				csestudent2,Team Training,31-04-2018
				csestudent3,Regular Training,31-06-2018
				""".stripIndent();

		ImportHandler.Result<TrainingApproval> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.warnings()).isEmpty();
		assertThat(res.errors()).hasSize(1);
		assertThat(res.result()).hasSize(2);

		assertThat(res.errors()).containsEntry(3,
				List.of(new SimpleImportError("User with identifier csestudent3 not found")));

	}

	@Test
	@WithUserDetails("cseteacher1")
	void dateIsWrong() {
		String content = """
				csestudent1,Regular Training,26-02-2003
				csestudent2,Team Training,31-04-20189
				""".stripIndent();

		ImportHandler.Result<TrainingApproval> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.warnings()).isEmpty();
		assertThat(res.errors()).hasSize(1);
		assertThat(res.result()).hasSize(1);

		assertThat(res.errors()).containsEntry(2, List
				.of(new SimpleImportError("Deadline is not in the correct format, expected: dd-MM-yyyy")));

	}

	@Test
	@WithUserDetails("cseteacher1")
	void multipleTrainingsCoincideForOne() {

		when(trainingTypeService.getTrainingTypesByNameAndProgramId("Regular Training", programId))
				.thenReturn(List.of(t1, t2));
		String content = """
				csestudent1,Regular Training,26-02-2003
				csestudent2,Team Training,31-04-2018
				""".stripIndent();

		ImportHandler.Result<TrainingApproval> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.warnings()).hasSize(1);
		assertThat(res.errors()).isEmpty();
		assertThat(res.result()).hasSize(2);
		assertThat(res.warnings()).containsEntry(1, List.of(new SimpleImportWarning(
				"Multiple training types with name Regular Training found, using the first one")));

	}

	@Test
	@WithUserDetails("cseteacher1")
	void trainingDoesntExist() {

		when(trainingTypeService.getTrainingTypesByNameAndProgramId("Regular Training", programId))
				.thenReturn(Collections.emptyList());
		String content = """
				csestudent1,Regular Training,26-02-2003
				csestudent2,Team Training,31-04-2018
				""".stripIndent();

		ImportHandler.Result<TrainingApproval> res = importHandler
				.importFile(Arrays.stream(content.split("\n")).map(l -> l.split(",")).toList());

		assertThat(res.warnings()).isEmpty();
		assertThat(res.errors()).hasSize(1);
		assertThat(res.result()).hasSize(1);
		assertThat(res.errors()).containsKey(1);

	}

}
