/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.repository;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;

import java.time.LocalDate;
import java.util.List;

import javax.transaction.Transactional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;

import application.test.TestTAMApplication;
import nl.tudelft.tam.enums.AcademicLevel;
import nl.tudelft.tam.enums.AcademicPeriod;
import nl.tudelft.tam.model.Application;
import nl.tudelft.tam.model.JobOffer;

@SpringBootTest(classes = TestTAMApplication.class)
class ApplicationRepositoryTest {

	@Autowired
	private ApplicationRepository repository;

	@Autowired
	private JobOfferRepository jobOfferRepository;

	@BeforeEach
	void setup() {
		jobOfferRepository.deleteAll();
		repository.deleteAll();
	}

	@Test
	@Transactional
	void findByIdOrThrow() {
		Application.AppId id = new Application.AppId(42L, 69L);

		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id));

		JobOffer jobOffer = jobOfferRepository.save(JobOffer.builder()
				.id(4654L)
				.editionId(2354L)
				.contractName("Test")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.BACHELOR1)
				.period(AcademicPeriod.Q1)
				.build());

		Application application = repository.save(Application.builder()
				.id(id)
				.jobOffer(jobOffer)
				.build());

		assertThat(repository.findByIdOrThrow(application.getId()))
				.isEqualTo(application);
	}

	@Test
	@Transactional
	void findAllByPersonId() {
		Long pId1 = 4200L;
		Long pId2 = 6900L;
		Long joId1 = 98531L;
		Long joId2 = 97506L;
		Application.AppId id1 = new Application.AppId(pId1, joId1);
		Application.AppId id2 = new Application.AppId(pId1, joId2);
		Application.AppId id3 = new Application.AppId(pId2, joId2);

		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id1));
		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id2));
		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id3));

		JobOffer jobOffer1 = jobOfferRepository.save(JobOffer.builder()
				.id(joId1)
				.editionId(2354L)
				.contractName("Test")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.BACHELOR1)
				.period(AcademicPeriod.Q1)
				.build());

		JobOffer jobOffer2 = jobOfferRepository.save(JobOffer.builder()
				.id(joId2)
				.editionId(2354L)
				.contractName("Test 2")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 2 JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.MINOR)
				.period(AcademicPeriod.S1)
				.build());

		Application application1 = repository.save(Application.builder()
				.id(id1)
				.jobOffer(jobOffer1)
				.build());

		Application application2 = repository.save(Application.builder()
				.id(id2)
				.jobOffer(jobOffer2)
				.build());

		Application application3 = repository.save(Application.builder()
				.id(id3)
				.jobOffer(jobOffer2)
				.build());

		List<Application> res = repository.findAllByPersonId(pId1);
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(application1, application2);
		assertThat(res).doesNotContain(application3);
	}

	@Test
	@Transactional
	void findAllByOfferId() {
		JobOffer jobOffer1 = jobOfferRepository.save(JobOffer.builder()
				.editionId(2354L)
				.contractName("Test 2")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 2 JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.BACHELOR1)
				.period(AcademicPeriod.Q1)
				.build());

		JobOffer jobOffer2 = jobOfferRepository.save(JobOffer.builder()
				.editionId(2354L)
				.contractName("Test 2")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 2 JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.MINOR)
				.period(AcademicPeriod.S1)
				.build());

		Long pId1 = 4200L;
		Long pId2 = 6900L;
		Long joId1 = jobOffer1.getId();
		Long joId2 = jobOffer2.getId();

		Application.AppId id1 = new Application.AppId(pId1, joId1);
		Application.AppId id2 = new Application.AppId(pId1, joId2);
		Application.AppId id3 = new Application.AppId(pId2, joId2);

		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id1));
		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id2));
		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id3));

		Application application1 = repository.save(Application.builder()
				.id(id1)
				.jobOffer(jobOffer1)
				.build());

		Application application2 = repository.save(Application.builder()
				.id(id2)
				.jobOffer(jobOffer2)
				.build());

		Application application3 = repository.save(Application.builder()
				.id(id3)
				.jobOffer(jobOffer2)
				.build());

		List<Application> res = repository.findAllByOfferId(joId2);
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(application2, application3);
		assertThat(res).doesNotContain(application1);
	}
}
