/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.repository;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;

import java.util.List;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;

import application.test.TestTAMApplication;
import nl.tudelft.tam.model.Declaration;
import nl.tudelft.tam.model.ExtraWork;
import nl.tudelft.tam.model.Profile;

@SpringBootTest(classes = TestTAMApplication.class)
class DeclarationRepositoryTest {

	@Autowired
	private DeclarationRepository repository;

	@Autowired
	private ProfileRepository profileRepository;

	@Autowired
	private ExtraWorkRepository extraWorkRepository;

	@AfterEach
	void cleanUp() {
		repository.deleteAll();
		extraWorkRepository.deleteAll();
	}

	@Test
	void findByIdOrThrow() {
		Long id = 48567L;

		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id));

		Profile profile = profileRepository.save(Profile.builder().id(42L).build());
		ExtraWork extraWork = extraWorkRepository
				.save(ExtraWork.builder().id(69L).editionId(420L).name("Extra Work")
						.contractName("Extra-Work-Contract").baanCode("Extra Work Baan Code").maxHours(100)
						.build());

		Declaration declaration = repository.save(Declaration.builder()
				.id(id)
				.amount(654)
				.reason("Test Reason")
				.person(profile)
				.extraWork(extraWork)
				.build());

		assertThat(repository.findByIdOrThrow(declaration.getId()))
				.isEqualTo(declaration);
	}

	@Test
	void findAllByPersonIdTest() {
		Long person1 = 5001L;
		Long person2 = 5002L;

		Profile profile1 = profileRepository.save(Profile.builder().id(person1).build());
		Profile profile2 = profileRepository.save(Profile.builder().id(person2).build());
		ExtraWork extraWork = extraWorkRepository
				.save(ExtraWork.builder().id(69L).editionId(420L).name("Extra Work")
						.contractName("Extra-Work-Contract").baanCode("Extra Work Baan Code").maxHours(100)
						.build());

		Declaration declaration1 = repository.save(Declaration.builder()
				.amount(10)
				.reason("Test Reason 1")
				.person(profile1)
				.extraWork(extraWork)
				.build());
		Declaration declaration2 = repository.save(Declaration.builder()
				.amount(20)
				.reason("Test Reason 2")
				.person(profile1)
				.extraWork(extraWork)
				.build());
		Declaration declaration3 = repository.save(Declaration.builder()
				.amount(30)
				.reason("Test Reason 2")
				.person(profile2)
				.extraWork(extraWork)
				.build());

		List<Declaration> res = repository.findAllByPersonId(person1);
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(declaration1, declaration2);
	}

	@Test
	void findAllByWorkIdTest() {
		Profile profile = profileRepository.save(Profile.builder().id(204L).build());
		ExtraWork extraWork1 = extraWorkRepository
				.save(ExtraWork.builder().editionId(420L).name("Extra Work")
						.contractName("Extra-Work-Contract").baanCode("Extra Work Baan Code").maxHours(100)
						.build());
		ExtraWork extraWork2 = extraWorkRepository
				.save(ExtraWork.builder().editionId(420L).name("Extra Work")
						.contractName("Extra-Work-Contract").baanCode("Extra Work Baan Code").maxHours(100)
						.build());

		Declaration declaration1 = repository.save(Declaration.builder()
				.amount(10)
				.reason("Test Reason 1")
				.person(profile)
				.extraWork(extraWork1)
				.build());
		Declaration declaration2 = repository.save(Declaration.builder()
				.amount(20)
				.reason("Test Reason 2")
				.person(profile)
				.extraWork(extraWork1)
				.build());
		Declaration declaration3 = repository.save(Declaration.builder()
				.amount(30)
				.reason("Test Reason 3")
				.person(profile)
				.extraWork(extraWork2)
				.build());

		List<Declaration> res = repository.findAllByWorkId(extraWork1.getId());
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(declaration1, declaration2);
	}
}
