/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.repository;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;

import java.time.LocalDate;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;

import application.test.TestTAMApplication;
import nl.tudelft.tam.model.ExtraWork;

@SpringBootTest(classes = TestTAMApplication.class)
class ExtraWorkRepositoryTest {

	@Autowired
	private ExtraWorkRepository repository;

	@BeforeEach
	void setup() {
		repository.deleteAll();
	}

	@Test
	void findByIdOrThrow() {
		Long id = 6546L;

		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id));

		ExtraWork extraWork = repository.save(ExtraWork.builder()
				.id(id)
				.editionId(42L)
				.name("Extra Work")
				.contractName("Extra-Work-Contract").baanCode("Extra Work Baan Code").maxHours(100).build());

		assertThat(repository.findByIdOrThrow(extraWork.getId()))
				.isEqualTo(extraWork);
	}

	@Test
	void findAllOpenTest() {
		Long id1 = 789765L;
		Long id2 = 984654L;
		Long id3 = 321498L;

		assert (repository.findAllOpen()).isEmpty();

		ExtraWork extraWork1 = repository.save(ExtraWork.builder()
				.id(6845L)
				.editionId(id1)
				.deadline(LocalDate.now().minusDays(1))
				.name("grading")
				.contractName("Extra-Work-1-Contract").baanCode("Extra Work 1 Baan Code").maxHours(100)
				.build());

		ExtraWork extraWork2 = repository.save(ExtraWork.builder()
				.id(89465L)
				.editionId(id2)
				.deadline(LocalDate.now())
				.name("grading").contractName("Extra-Work-2-Contract").baanCode("Extra Work 2 Baan Code")
				.maxHours(100)
				.build());

		ExtraWork extraWork3 = repository.save(ExtraWork.builder()
				.id(948534L)
				.editionId(id3)
				.deadline(LocalDate.now().plusDays(1))
				.name("grading")
				.contractName("Extra-Work-3-Contract").baanCode("Extra Work 3 Baan Code").maxHours(100)
				.build());

		List<ExtraWork> res = repository.findAllOpen();

		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(extraWork2, extraWork3);
	}

	@Test
	void findAllClosedTest() {
		Long id1 = 789765L;
		Long id2 = 984654L;
		Long id3 = 321498L;

		assert (repository.findAllOpen()).isEmpty();

		ExtraWork extraWork1 = repository.save(ExtraWork.builder()
				.id(6845L)
				.editionId(id1)
				.deadline(LocalDate.now().minusDays(1))
				.name("grading")
				.contractName("Extra-Work-1-Contract").baanCode("Extra Work 2 Baan Code").maxHours(100)
				.build());

		ExtraWork extraWork2 = repository.save(ExtraWork.builder()
				.id(89465L)
				.editionId(id2)
				.deadline(LocalDate.now())
				.name("grading").contractName("Extra-Work-2-Contract").baanCode("Extra Work 2 Baan Code")
				.maxHours(100)
				.build());

		ExtraWork extraWork3 = repository.save(ExtraWork.builder()
				.id(948534L)
				.editionId(id3)
				.deadline(LocalDate.now().plusDays(1))
				.name("grading")
				.contractName("Extra-Work-3-Contract").contractName("Extra-Work-3-Contract")
				.baanCode("Extra Work 3 Baan Code").maxHours(100).build());

		List<ExtraWork> res = repository.findAllClosed();

		assertThat(res.size()).isEqualTo(1);
		assertThat(res).contains(extraWork1);
	}

	@Test
	void findAllByEditions() {
		Long id1 = 789765L;
		Long id2 = 984654L;
		Long id3 = 321498L;

		assert (repository.findAllOpen()).isEmpty();

		ExtraWork extraWork1 = repository.save(ExtraWork.builder()
				.id(6845L)
				.editionId(id1)
				.deadline(LocalDate.now().minusDays(1))
				.name("grading")
				.contractName("Extra-Work-1-Contract").baanCode("Extra Work 1 Baan Code").maxHours(100)
				.build());

		ExtraWork extraWork2 = repository.save(ExtraWork.builder()
				.id(89465L)
				.editionId(id2)
				.deadline(LocalDate.now())
				.name("grading").contractName("Extra-Work-2-Contract").baanCode("Extra Work 2 Baan Code")
				.maxHours(100)
				.build());

		ExtraWork extraWork3 = repository.save(ExtraWork.builder()
				.id(948534L)
				.editionId(id3)
				.deadline(LocalDate.now().plusDays(1))
				.name("grading")
				.contractName("Extra-Work-3-Contract").baanCode("Extra Work 3 Baan Code").maxHours(100)
				.build());

		List<ExtraWork> res = repository.findAllByEditions(List.of(id1, id3));

		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(extraWork1, extraWork3);
	}
}
