/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.repository;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertThrows;

import java.time.LocalDate;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.tam.enums.AcademicLevel;
import nl.tudelft.tam.enums.AcademicPeriod;
import nl.tudelft.tam.model.JobOffer;

@Transactional
@SpringBootTest(classes = TestTAMApplication.class)
class JobOfferRepositoryTest {

	@Autowired
	private JobOfferRepository repository;

	@BeforeEach
	void setup() {
		repository.deleteAll();
	}

	@Test
	void findByIdOrThrow() {
		Long id = 6546L;

		assertThrows(ResourceNotFoundException.class, () -> repository.findByIdOrThrow(id));

		JobOffer jobOffer = repository.save(JobOffer.builder()
				.id(id)
				.editionId(42L)
				.contractName("Test")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test JO BAAN CODE")
				.maxHours(100)
				.deadline(LocalDate.now())
				.level(AcademicLevel.BACHELOR1)
				.period(AcademicPeriod.Q1)
				.build());

		assertThat(repository.findByIdOrThrow(jobOffer.getId()))
				.isEqualTo(jobOffer);
	}

	@Test
	void findAllOpen() {
		Long id1 = 789765L;
		Long id2 = 984654L;
		Long id3 = 321498L;

		assert (repository.findAllOpen()).isEmpty();

		JobOffer jobOffer1 = repository.save(JobOffer.builder()
				.id(id1)
				.editionId(42L)
				.deadline(LocalDate.now().minusDays(1))
				.contractName("Test")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.BACHELOR1)
				.period(AcademicPeriod.Q1)
				.build());

		JobOffer jobOffer2 = repository.save(JobOffer.builder()
				.id(id2)
				.editionId(42L)
				.deadline(LocalDate.now())
				.contractName("Test 3")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 3 JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.MINOR)
				.period(AcademicPeriod.S1)
				.build());

		JobOffer jobOffer3 = repository.save(JobOffer.builder()
				.id(id3)
				.editionId(42L)
				.deadline(LocalDate.now().plusDays(1))
				.contractName("Test 3")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 3 JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.MASTER1)
				.period(AcademicPeriod.YEAR)
				.build());

		List<JobOffer> res = repository.findAllOpen();

		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(jobOffer2, jobOffer3);
	}

	@Test
	void findAllClosed() {
		Long id1 = 789765L;
		Long id2 = 984654L;
		Long id3 = 321498L;

		assert (repository.findAllOpen()).isEmpty();

		JobOffer jobOffer1 = repository.save(JobOffer.builder()
				.id(id1)
				.editionId(42L)
				.deadline(LocalDate.now().minusDays(1))
				.contractName("Test")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.BACHELOR1)
				.period(AcademicPeriod.Q1)
				.build());

		JobOffer jobOffer2 = repository.save(JobOffer.builder()
				.id(id2)
				.editionId(42L)
				.deadline(LocalDate.now())
				.contractName("Test 2")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 2 JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.MINOR)
				.period(AcademicPeriod.S1)
				.build());

		JobOffer jobOffer3 = repository.save(JobOffer.builder()
				.id(id3)
				.editionId(42L)
				.deadline(LocalDate.now().plusDays(1))
				.contractName("Test 3")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 3 JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.MASTER1)
				.period(AcademicPeriod.YEAR)
				.build());

		List<JobOffer> res = repository.findAllClosed();

		assertThat(res.size()).isEqualTo(1);
		assertThat(res).contains(jobOffer1);
	}

	@Test
	void findAllByEditions() {
		Long id1 = 789765L;
		Long id2 = 984654L;
		Long id3 = 321498L;

		assert (repository.findAllOpen()).isEmpty();

		JobOffer jobOffer1 = repository.save(JobOffer.builder()
				.id(4684531L)
				.editionId(id1)
				.contractName("Test")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test JO BAAN CODE")
				.maxHours(100)
				.deadline(LocalDate.now().minusDays(1))
				.level(AcademicLevel.BACHELOR1)
				.period(AcademicPeriod.Q1)
				.build());

		JobOffer jobOffer2 = repository.save(JobOffer.builder()
				.id(6846846L)
				.editionId(id2)
				.contractName("Test 2")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 2 JO BAAN CODE")
				.maxHours(100)
				.deadline(LocalDate.now())
				.level(AcademicLevel.MINOR)
				.period(AcademicPeriod.S1)
				.build());

		JobOffer jobOffer3 = repository.save(JobOffer.builder()
				.id(987654L)
				.editionId(id3)
				.deadline(LocalDate.now().plusDays(1))
				.contractName("Test 3")
				.contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(2))
				.baanCode("Test 3 JO BAAN CODE")
				.maxHours(100)
				.level(AcademicLevel.MASTER1)
				.period(AcademicPeriod.YEAR)
				.build());

		List<JobOffer> res = repository.findAllByEditions(List.of(id1, id3));

		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(jobOffer1, jobOffer3);
	}

}
