/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.security;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.ProgramControllerApi;
import nl.tudelft.labracore.api.dto.CourseSummaryDTO;
import nl.tudelft.labracore.api.dto.ProgramDetailsDTO;
import nl.tudelft.labracore.lib.security.user.DefaultRole;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.service.CoordinatorService;
import nl.tudelft.tam.service.CourseService;
import reactor.core.publisher.Mono;

@SpringBootTest(classes = TestTAMApplication.class)
public class AuthorisationServiceTest {

	@SpyBean
	AuthorisationService service;

	@MockBean
	CoordinatorService coordinatorService;

	@MockBean
	CourseService courseService;

	@Autowired
	ProgramControllerApi programApi;

	Person admin, teacher, student;

	@BeforeEach
	void setup() {
		admin = Person.builder().defaultRole(DefaultRole.ADMIN).id(1L).build();
		teacher = Person.builder().defaultRole(DefaultRole.TEACHER).id(2L).build();
		student = Person.builder().defaultRole(DefaultRole.STUDENT).id(3L).build();
	}

	@Test
	void isAdminTest() {
		doReturn(admin).when(service).getAuthPerson();
		assertThat(service.isAdmin()).isTrue();

		doReturn(student).when(service).getAuthPerson();
		assertThat(service.isAdmin()).isFalse();
	}

	@Test
	void isStaffTest() {
		doReturn(admin).when(service).getAuthPerson();
		assertThat(service.isStaff()).isTrue();

		doReturn(teacher).when(service).getAuthPerson();
		assertThat(service.isStaff()).isTrue();

		doReturn(student).when(service).getAuthPerson();
		assertThat(service.isStaff()).isFalse();
	}

	@Test
	void isCoordinatorOfAnyTest() {
		doReturn(admin).when(service).getAuthPerson();
		assertThat(service.isCoordinatorOfAny()).isTrue();

		doReturn(teacher).when(service).getAuthPerson();
		assertThat(service.isCoordinatorOfAny()).isFalse();

		doReturn(true).when(coordinatorService).existsByPersonId(anyLong());
		assertThat(service.isCoordinatorOfAny()).isTrue();
	}

	@Test
	void isCoordinatorForProgramTest() {
		when(programApi.getProgramById(anyLong()))
				.thenReturn(Mono.just(new ProgramDetailsDTO().coordinators(Collections.emptyList())));

		doReturn(admin).when(service).getAuthPerson();
		assertThat(service.isCoordinatorForProgram(1L)).isTrue();

		doReturn(teacher).when(service).getAuthPerson();
		assertThat(service.isCoordinatorForProgram(1L)).isFalse();

		doReturn(true).when(coordinatorService).existsByPersonIdAndProgramId(anyLong(), anyLong());
		assertThat(service.isCoordinatorForProgram(1L)).isTrue();
	}

	@Test
	void isManagerOfCourseTest() {
		doReturn(admin).when(service).getAuthPerson();
		assertThat(service.isManagerOfCourse(1L)).isTrue();

		doReturn(teacher).when(service).getAuthPerson();
		assertThat(service.isManagerOfCourse(1L)).isFalse();

		doReturn(List.of(new CourseSummaryDTO().id(1L))).when(courseService).getManagingCourses(any());
		assertThat(service.isManagerOfCourse(1L)).isTrue();
	}

	@Test
	void isManagerOfAnyTest() {
		doReturn(admin).when(service).getAuthPerson();
		assertThat(service.isManagerOfAny()).isTrue();

		doReturn(teacher).when(service).getAuthPerson();
		assertThat(service.isManagerOfAny()).isTrue();
	}

	@Test
	void canCreateEditionTest() {
		doReturn(true).when(service).isManagerOfCourse(anyLong());
		assertThat(service.canCreateEdition(1L)).isTrue();

		verify(service).isManagerOfCourse(eq(1L));
	}

	@Test
	void canEditTaTrainingTest() {
		doReturn(true).when(service).isAdmin();
		assertThat(service.canEditTaTraining()).isTrue();

		verify(service).isAdmin();
	}

	@Test
	void canEditPayScaleTest() {
		doReturn(true).when(service).isAdmin();
		assertThat(service.canEditPayScale()).isTrue();

		verify(service).isAdmin();
	}

}
