/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.time.LocalDate;
import java.util.*;

import javax.transaction.Transactional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.CsvSource;
import org.mockito.ArgumentCaptor;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.*;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.cache.CourseCacheManager;
import nl.tudelft.tam.cache.EditionCacheManager;
import nl.tudelft.tam.dto.view.details.ApplicationDetailsJobOfferDTO;
import nl.tudelft.tam.dto.view.details.ApplicationDetailsPersonDTO;
import nl.tudelft.tam.dto.view.details.JobOfferDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ApplicationSummaryDTO;
import nl.tudelft.tam.enums.NotificationFrequency;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.model.*;
import nl.tudelft.tam.repository.ApplicationChangeEventRepository;
import nl.tudelft.tam.repository.ApplicationRepository;
import nl.tudelft.tam.repository.JobOfferRepository;
import nl.tudelft.tam.repository.ProfileRepository;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class ApplicationServiceMockTest {

	@SpyBean
	ApplicationService service;

	@MockBean
	ApplicationRepository repository;

	@MockBean
	ApplicationChangeEventRepository applicationChangeEventRepository;

	@MockBean
	JobOfferRepository jobOfferRepository;

	@MockBean
	JobOfferService jobOfferService;

	@MockBean
	RoleService roleService;

	@MockBean
	CoordinatorService coordinatorService;

	@MockBean
	PersonService personService;

	@MockBean
	EditionService editionService;

	@MockBean
	CourseService courseService;

	@MockBean
	EditionCacheManager editionCache;

	@MockBean
	CourseCacheManager courseCacheManager;

	@MockBean
	ProfileRepository profileRepository;

	@MockBean
	CSVService csvService;

	@MockBean
	CohortService cohortService;

	@MockBean
	CoordinatorDefaultService coordinatorDefaultService;

	@Autowired
	ModelMapper mapper;

	@MockBean
	EmailService emailService;

	JobOffer offer1, offer2;
	Application app1a, app1b, app1c, app2a;
	PersonSummaryDTO personSummary1, personSummary2;
	EditionDetailsDTO editionDetails;
	CourseDetailsDTO courseDetails;
	CourseSummaryDTO courseSummary;
	CohortDetailsDTO cohortDetails;
	ProgramSummaryDTO programSummary;
	CohortSummaryDTO cohortSummary;

	final long JOFFER_ID_1 = 1001;
	final long JOFFER_ID_2 = 1002;
	final long PERSON_ID_1 = 5001;
	final long PERSON_ID_2 = 5002;
	final long EDITION_ID = 7000;
	final long ADMIN_ID = 329476;
	final long COURSE_ID = 10932;
	final long PROGRAM_ID = 1;
	final long COHORT_ID = 219;

	@BeforeEach
	void setUp() {
		offer1 = JobOffer.builder().name("offer-1").id(JOFFER_ID_1).editionId(EDITION_ID)
				.contractStartDate(LocalDate.of(2023, 2, 1))
				.contractEndDate(LocalDate.of(2023, 2, 2))
				.deadline(LocalDate.now().plusDays(1)).build();
		offer2 = JobOffer.builder().name("offer-2").id(JOFFER_ID_2).editionId(EDITION_ID)
				.contractStartDate(LocalDate.of(2023, 2, 1))
				.contractEndDate(LocalDate.of(2023, 2, 2))
				.deadline(LocalDate.now().plusDays(2)).build();

		app1a = Application.builder().id(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).jobOffer(offer1)
				.status(Status.SUBMITTED).build();
		app1b = Application.builder().id(new Application.AppId(PERSON_ID_2, JOFFER_ID_1)).jobOffer(offer1)
				.status(Status.OFFERED).build();
		app1c = Application.builder().id(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).jobOffer(offer1)
				.status(Status.ACCEPTED).build();
		app2a = Application.builder().id(new Application.AppId(PERSON_ID_1, JOFFER_ID_2)).jobOffer(offer2)
				.status(Status.REJECTED_BY_STUDENT).build();
		offer1.setApplications(List.of(app1a, app1b));

		personSummary1 = new PersonSummaryDTO().id(PERSON_ID_1).username("Person1").displayName("Person 1")
				.email("Person1@email.com").number(1);
		personSummary2 = new PersonSummaryDTO().id(PERSON_ID_2).username("Person2").displayName("Person 2")
				.email("Person2@email.com").number(2);

		courseDetails = new CourseDetailsDTO().id(COURSE_ID).name("Course").code("Test")
				.managers(List.of(personSummary1, personSummary2));
		courseSummary = new CourseSummaryDTO().id(COURSE_ID).name("Course").code("Test");
		programSummary = new ProgramSummaryDTO().id(PROGRAM_ID).name("Program");
		cohortDetails = new CohortDetailsDTO().id(COHORT_ID).name("Cohort").program(programSummary);
		cohortSummary = new CohortSummaryDTO().id(COHORT_ID).name("Cohort");
		editionDetails = new EditionDetailsDTO().id(EDITION_ID).name("Edition").course(courseSummary)
				.cohort(cohortSummary);

		offer1.setApplications(List.of(app1a, app1b));
		offer2.setApplications(List.of(app2a));
	}

	@Test
	void offerByUsernameTest() {
		String username = "USER-name-42";
		Application.AppId id = app1a.getId();

		when(personService.getPersonByUsername(anyString()))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).username(username));
		when(repository.findById(any())).thenReturn(Optional.of(app1a));
		when(profileRepository.findByIdOrThrow(eq(ADMIN_ID)))
				.thenReturn(Profile.builder().id(ADMIN_ID).build());
		when(editionService.getEditionById(anyLong())).thenReturn(editionDetails);
		when(personService.getPersonById(eq(PERSON_ID_1)))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).email("test@test.com")
						.displayName("Test Person"));
		when(profileRepository.findById(eq(PERSON_ID_1))).thenReturn(Optional.of(Profile.builder()
				.id(PERSON_ID_1)
				.emailNotifications(true)
				.emailNotificationFrequency(NotificationFrequency.Daily)
				.payScale(PayScale.SA2)
				.createdDate(LocalDate.now())
				.lastModifiedDate(LocalDate.now())
				.build()));
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		boolean result = service.offerByUsername(username, JOFFER_ID_1, ADMIN_ID);

		assertThat(result).isTrue();

		verify(personService).getPersonByUsername(username);
		verify(repository).findById(id);
		assertThat(app1a.getStatus()).isEqualTo(Status.OFFERED);
		assertThat(app1a.getHandler().getId()).isEqualTo(ADMIN_ID);
	}

	@Test
	void offerByStudentNumberTest() {
		Long failingStudentNumber = 293894L;
		Long workingStudentNumber = PERSON_ID_1;

		Application.AppId id = app1a.getId();

		when(personService.getPersonById(eq(workingStudentNumber)))
				.thenReturn(new PersonDetailsDTO().id(workingStudentNumber));
		when(personService.getPersonById(eq(failingStudentNumber)))
				.thenThrow(new NoSuchElementException());
		when(repository.findById(any())).thenReturn(Optional.of(app1a));
		when(profileRepository.findByIdOrThrow(eq(ADMIN_ID)))
				.thenReturn(Profile.builder().id(ADMIN_ID).build());
		when(editionService.getEditionById(anyLong())).thenReturn(editionDetails);
		when(personService.getPersonById(eq(PERSON_ID_1)))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).email("test@test.com")
						.displayName("Test Person"));
		when(profileRepository.findById(eq(PERSON_ID_1))).thenReturn(Optional.of(Profile.builder()
				.id(PERSON_ID_1)
				.emailNotifications(true)
				.emailNotificationFrequency(NotificationFrequency.Daily)
				.payScale(PayScale.SA2)
				.createdDate(LocalDate.now())
				.lastModifiedDate(LocalDate.now())
				.build()));
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		boolean workingResult = service.offerByStudentNumber(workingStudentNumber, JOFFER_ID_1, ADMIN_ID);
		assertThat(workingResult).isTrue();

		boolean failingResult = service.offerByStudentNumber(failingStudentNumber, JOFFER_ID_1, ADMIN_ID);
		assertThat(failingResult).isFalse();

		verify(personService, times(2)).getPersonById(workingStudentNumber);
		verify(repository).findById(id);
		assertThat(app1a.getStatus()).isEqualTo(Status.OFFERED);
		assertThat(app1a.getHandler().getId()).isEqualTo(ADMIN_ID);
	}

	@Test
	void getApplicationsForPersonTest() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID).name("test-name");
		JobOfferDetailsDTO offer = mapper.map(offer1, JobOfferDetailsDTO.class);
		offer.setEdition(edition);
		ApplicationDetailsJobOfferDTO app = mapper.map(app1a, ApplicationDetailsJobOfferDTO.class);
		app.setJobOffer(offer);

		when(repository.findAllByPersonId(anyLong())).thenReturn(List.of(app1a, app2a));
		when(editionService.get(anyList())).thenReturn(List.of(edition));

		List<ApplicationDetailsJobOfferDTO> res = service.getApplicationsForPerson(PERSON_ID_1);
		assertThat(res.size()).isEqualTo(1);
		assertThat(res.get(0)).isEqualTo(app);

		verify(repository).findAllByPersonId(PERSON_ID_1);
		verify(editionService).get(List.of(EDITION_ID));
	}

	@Test
	void getApplicationsForPersonTestEmpty() {
		when(repository.findAllByPersonId(anyLong())).thenReturn(List.of());

		List<ApplicationDetailsJobOfferDTO> res = service.getApplicationsForPerson(PERSON_ID_1);
		assertThat(res.isEmpty()).isTrue();

		verify(repository).findAllByPersonId(PERSON_ID_1);
		verify(editionService, never()).getEditionsById(anyList());
	}

	@Test
	void getApplicationsForJobOfferTest() {
		PersonSummaryDTO person1 = new PersonSummaryDTO().id(PERSON_ID_1).username("user1");
		PersonSummaryDTO person2 = new PersonSummaryDTO().id(PERSON_ID_2).username("user2");
		ApplicationDetailsPersonDTO app1 = mapper.map(app1a, ApplicationDetailsPersonDTO.class);
		ApplicationDetailsPersonDTO app2 = mapper.map(app1b, ApplicationDetailsPersonDTO.class);
		app1.setPerson(person1);
		app2.setPerson(person2);

		when(repository.findAllByOfferId(anyLong())).thenReturn(List.of(app1a, app1b));
		when(jobOfferService.findByIdOrThrow(app1a.getJobOffer().getId())).thenReturn(app1a.getJobOffer());
		when(jobOfferService.findByIdOrThrow(app2a.getJobOffer().getId())).thenReturn(app2a.getJobOffer());

		when(personService.getPeopleById(anyList())).thenReturn(List.of(person1, person2));
		var edition = new EditionDetailsDTO().id(1L).name("Comma, Edition")
				.course(new CourseSummaryDTO().name("Course"));

		when(editionService.getOrThrow(eq(1L)))
				.thenReturn(edition);

		when(editionService.getEditionById(any()))
				.thenReturn(edition);
		when(courseService.getOrThrow(any()))
				.thenReturn(new CourseDetailsDTO().id(1L).name("course")
						.editions(List.of(new EditionSummaryDTO().id(1L))));
		when(roleService.getRolesById(any(), any())).thenReturn(List.of());

		List<ApplicationDetailsPersonDTO> res = service.getApplicationsForJobOffer(JOFFER_ID_1);
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(app1, app2);

		verify(repository).findAllByOfferId(JOFFER_ID_1);
		verify(personService).getPeopleById(List.of(PERSON_ID_1, PERSON_ID_2));
	}

	@Test
	void getApplicationsForJobOfferTestEmpty() {
		when(repository.findAllByOfferId(anyLong())).thenReturn(List.of());

		assertThat(service.getApplicationsForJobOffer(JOFFER_ID_1).size()).isEqualTo(0);

		verify(repository).findAllByOfferId(JOFFER_ID_1);
		verify(personService, never()).getPeopleById(anyList());
	}

	@ParameterizedTest
	@CsvSource({ ", true",
			"course-name, true",
			"course-code, true",
			"cohort-name, true",
			"edition-name, true",
			"offer-1, true",
			"SUBMITTED, true",
			"er-1, true",
			"IttED, true",
			"RSE-CO, true",
			"ABCalsdkfj, false", })
	void getFilteredApplicationsForPersonTest(String filter, boolean present) {
		JobOfferDetailsDTO offer = mapper.map(offer1, JobOfferDetailsDTO.class);
		ApplicationDetailsJobOfferDTO app = mapper.map(app1a, ApplicationDetailsJobOfferDTO.class);
		EditionDetailsDTO edition = new EditionDetailsDTO().id(EDITION_ID)
				.course(new CourseSummaryDTO().name("course-name").code("course-code"))
				.cohort(new CohortSummaryDTO().name("cohort-name")).name("edition-name");
		offer.setEdition(edition);
		app.setJobOffer(offer);

		when(repository.findAllByPersonId(anyLong())).thenReturn(List.of(app1a));
		when(editionService.get(anyList())).thenReturn(List.of(edition));

		List<ApplicationDetailsJobOfferDTO> res = service.getFilteredApplicationsForPerson(PERSON_ID_1,
				filter);
		assertThat(res.isEmpty()).isEqualTo(!present);

		verify(repository).findAllByPersonId(PERSON_ID_1);
		verify(editionService).get(List.of(EDITION_ID));
	}

	@ParameterizedTest
	@CsvSource({ ", true",
			"user1, true",
			"display, true",
			"external, true",
			"40469555, true",
			"mail@domain.zw, true",
			"@domain.zw, true",
			"69, true",
			"UsEr, true",
			"abcdef703485, false" })
	void getFilteredApplicationsForJobOfferTest(String filter, boolean present) {
		PersonSummaryDTO person1 = new PersonSummaryDTO().id(PERSON_ID_1).username("user1")
				.displayName("display").externalId("external").number(40469555).email("mail@domain.zw");
		ApplicationDetailsPersonDTO app1 = mapper.map(app1a, ApplicationDetailsPersonDTO.class);
		app1.setPerson(person1);

		when(repository.findAllByOfferId(anyLong())).thenReturn(List.of(app1a));
		when(jobOfferService.findByIdOrThrow(any())).thenReturn(app1a.getJobOffer());
		when(personService.getPeopleById(anyList())).thenReturn(List.of(person1));
		var edition = new EditionDetailsDTO().id(1L).name("Comma, Edition")
				.course(new CourseSummaryDTO().name("Course"));

		when(editionService.getOrThrow(eq(1L)))
				.thenReturn(edition);

		when(editionService.getEditionById(any()))
				.thenReturn(edition);
		when(courseService.getOrThrow(any()))
				.thenReturn(new CourseDetailsDTO().id(1L).name("course")
						.editions(List.of(new EditionSummaryDTO().id(1L))));
		when(roleService.getRolesById(any(), any())).thenReturn(List.of());

		assertThat(service.getFilteredApplicationsForJobOffer(JOFFER_ID_1, filter).isEmpty())
				.isEqualTo(!present);

		verify(repository).findAllByOfferId(JOFFER_ID_1);
		verify(personService).getPeopleById(List.of(PERSON_ID_1));
	}

	@Test
	void getAppsMapForPersonByIdsTest() {
		app2a.setStatus(Status.ACCEPTED);

		var id1 = new Application.AppId(PERSON_ID_1, JOFFER_ID_1);
		var id2 = new Application.AppId(PERSON_ID_1, JOFFER_ID_2);

		when(repository.findById(eq(id1))).thenReturn(Optional.of(app1a));
		when(repository.findById(eq(id2))).thenReturn(Optional.of(app2a));

		HashMap<Long, ApplicationSummaryDTO> res = service.getAppsMapForPersonByIds(PERSON_ID_1,
				List.of(JOFFER_ID_1, JOFFER_ID_2));
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).containsKeys(JOFFER_ID_1, JOFFER_ID_2);
		assertThat(res.get(JOFFER_ID_1)).isEqualTo(mapper.map(app1a, ApplicationSummaryDTO.class));
		assertThat(res.get(JOFFER_ID_2)).isEqualTo(mapper.map(app2a, ApplicationSummaryDTO.class));

		verify(repository, times(1)).findById(id1);
		verify(repository, times(1)).findById(id2);
	}

	@Test
	void accept() {
		when(repository.findByIdOrThrow(app1b.getId())).thenReturn(app1b);
		assertThat(app1b.getStatus()).isEqualTo(Status.OFFERED);

		doNothing().when(roleService).changeToTa(anyLong(), anyLong(), anyBoolean());

		service.accept(PERSON_ID_2, JOFFER_ID_1);

		assertThat(app1b.getStatus()).isEqualTo(Status.ACCEPTED);

		verify(roleService).changeToTa(PERSON_ID_2, JOFFER_ID_1, false);
	}

	@Test
	void testPromoteFromAcceptedStatus() {
		app1b.setStatus(Status.ACCEPTED);
		when(repository.findByIdOrThrow(app1b.getId())).thenReturn(app1b);
		assertThat(app1b.getStatus()).isEqualTo(Status.ACCEPTED);

		doNothing().when(roleService).changeToHeadTa(anyLong(), anyLong());

		service.promote(PERSON_ID_2, JOFFER_ID_1);

		assertThat(app1b.getStatus()).isEqualTo(Status.ACCEPTED);

		verify(roleService).changeToHeadTa(PERSON_ID_2, JOFFER_ID_1);
	}

	@Test
	void testPromoteFromNotAcceptedStatus() {
		when(repository.findByIdOrThrow(app1b.getId())).thenReturn(app1b);
		assertThrows(IllegalStateException.class, () -> service.promote(PERSON_ID_2, JOFFER_ID_1));

		verify(roleService, never()).changeToHeadTa(anyLong(), anyLong());
	}

	@Test
	void testDemoteFromAcceptedStatus() {
		app1b.setStatus(Status.ACCEPTED);
		when(repository.findByIdOrThrow(app1b.getId())).thenReturn(app1b);
		assertThat(app1b.getStatus()).isEqualTo(Status.ACCEPTED);

		doNothing().when(roleService).changeToTa(anyLong(), anyLong(), anyBoolean());

		service.demote(PERSON_ID_2, JOFFER_ID_1);

		assertThat(app1b.getStatus()).isEqualTo(Status.ACCEPTED);

		verify(roleService).changeToTa(PERSON_ID_2, JOFFER_ID_1, true);
	}

	@Test
	void testDemoteFromNotAcceptedStatus() {
		when(repository.findByIdOrThrow(app1b.getId())).thenReturn(app1b);
		assertThrows(IllegalStateException.class, () -> service.demote(PERSON_ID_2, JOFFER_ID_1));

		verify(roleService, never()).changeToTa(anyLong(), anyLong(), anyBoolean());
	}

	@Test
	void reject() {
		when(repository.findByIdOrThrow(app1a.getId())).thenReturn(app1a);
		assertThat(app1a.getStatus()).isEqualTo(Status.SUBMITTED);
		when(profileRepository.findByIdOrThrow(eq(ADMIN_ID)))
				.thenReturn(Profile.builder().id(ADMIN_ID).build());

		doNothing().when(roleService).removeRole(anyLong(), anyLong());
		when(courseService.getCourseById(anyLong())).thenReturn(courseDetails);
		when(editionService.getEditionById(anyLong())).thenReturn(editionDetails);
		when(personService.getPersonById(eq(PERSON_ID_1)))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).email("test@test.com")
						.displayName("Test Person"));
		when(profileRepository.findByIdOrThrow(eq(PERSON_ID_1))).thenReturn(Profile.builder()
				.id(PERSON_ID_1)
				.emailNotifications(true)
				.emailNotificationFrequency(NotificationFrequency.Daily)
				.payScale(PayScale.SA2)
				.createdDate(LocalDate.now())
				.lastModifiedDate(LocalDate.now())
				.build());
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		service.reject(PERSON_ID_1, JOFFER_ID_1, ADMIN_ID);

		assertThat(app1a.getStatus()).isEqualTo(Status.REJECTED_BY_TEACHER);
		assertThat(app1a.getHandler().getId()).isEqualTo(ADMIN_ID);

		verify(roleService).removeRole(PERSON_ID_1, JOFFER_ID_1);
	}

	@Test
	void bulkRejectTest() {
		PersonSummaryDTO s1 = new PersonSummaryDTO().id(PERSON_ID_1);
		PersonSummaryDTO s2 = new PersonSummaryDTO().id(PERSON_ID_2);
		List<PersonSummaryDTO> persons = List.of(s1, s2);

		doNothing().when(service).reject(anyLong(), anyLong(), anyLong());
		when(personService.searchForPeople(anyString())).thenReturn(persons);

		service.bulkReject(anyLong(), anyString(), anyLong());

		verify(service, times(persons.size())).reject(anyLong(), anyLong(), anyLong());
	}

	@Test
	void rejectAllOpenInJobOffer() {
		when(jobOfferService.findByIdOrThrow(JOFFER_ID_1)).thenReturn(offer1);
		when(profileRepository.findByIdOrThrow(eq(ADMIN_ID)))
				.thenReturn(Profile.builder().id(ADMIN_ID).build());

		offer1.setApplications(List.of(Application.builder().status(Status.SUBMITTED)
				.id(new Application.AppId(PERSON_ID_1, JOFFER_ID_1)).build()));

		doNothing().when(roleService).removeRole(anyLong(), anyLong());
		when(courseService.getCourseById(anyLong())).thenReturn(courseDetails);
		when(editionService.getEditionById(anyLong())).thenReturn(editionDetails);
		when(personService.getPersonById(eq(PERSON_ID_1)))
				.thenReturn(new PersonDetailsDTO().id(PERSON_ID_1).email("test@test.com")
						.displayName("Test Person"));
		when(profileRepository.findByIdOrThrow(eq(PERSON_ID_1))).thenReturn(Profile.builder()
				.id(PERSON_ID_1)
				.emailNotifications(true)
				.emailNotificationFrequency(NotificationFrequency.Daily)
				.payScale(PayScale.SA2)
				.createdDate(LocalDate.now())
				.lastModifiedDate(LocalDate.now())
				.build());
		// Should be a Future, but result is never used
		when(emailService.sendEmail(anyList(), any())).thenReturn(null);

		service.rejectAllOpenInJobOffer(JOFFER_ID_1, ADMIN_ID);

		assertThat(offer1.getApplications().get(0).getStatus())
				.isEqualTo(Status.REJECTED_BY_TEACHER);
		assertThat(offer1.getApplications().get(0).getHandler().getId()).isEqualTo(ADMIN_ID);

		verify(roleService).removeRole(PERSON_ID_1, JOFFER_ID_1);
	}

	@Test
	void getCoordinatingApplicationsTest() {
		Long programId = 1L;
		Long courseId = 100L;

		Coordinator coordinator = new Coordinator();
		coordinator.setProgramId(programId);

		when(coordinatorService.findAllByPersonId(PERSON_ID_1)).thenReturn(List.of(coordinator));

		when(repository.findAll()).thenReturn(List.of(app1a));

		when(editionService.getOrThrow(eq(EDITION_ID)))
				.thenReturn(
						new EditionDetailsDTO().id(EDITION_ID).course(new CourseSummaryDTO().id(courseId)));
		when(courseService.getOrThrow(eq(courseId)))
				.thenReturn(
						new CourseDetailsDTO().id(courseId).program(new ProgramSummaryDTO().id(programId)));

		List<Application> result = service.getCoordinatingApplications(PERSON_ID_1);

		assertThat(result).isEqualTo(List.of(app1a));

		verify(coordinatorService).findAllByPersonId(anyLong());
		verify(repository).findAll();
	}

	@Test
	void exportApplicationsNonSpecific() throws IOException {
		List<Application> applications = new ArrayList<>();
		applications.add(app1a);
		applications.add(app1b);
		applications.add(app2a);

		String name = "test-application-non-specific";

		when(personService.get(any()))
				.thenReturn(List.of(personSummary1, personSummary2));

		doReturn(true).when(service).hasTAedBefore(anyLong(), anyLong());

		when(editionService.get(any())).thenReturn(List.of(editionDetails));

		@SuppressWarnings("unchecked") ArgumentCaptor<List<String[]>> captor = ArgumentCaptor
				.forClass(List.class);
		service.exportApplications(name, applications);
		verify(csvService).writeToCSV(captor.capture(), any(), eq(name + ".csv"));

		assertThat(captor.getValue())
				.contains(
						new String[] { "Student Number", "Email", "Name", "Edition", "Job Offer",
								"Status", "Student Response" },

						new String[] { "1",
								"Person1@email.com",
								"Person 1",
								"Course - Edition",
								"offer-1",
								"SUBMITTED",
								null },
						new String[] { "2",
								"Person2@email.com",
								"Person 2",
								"Course - Edition",
								"offer-1",
								"OFFERED",
								null },
						new String[] { "1",
								"Person1@email.com",
								"Person 1",
								"Course - Edition",
								"offer-2",
								"REJECTED_BY_STUDENT",
								null
						});
	}

	@Test
	void exportApplicationsSpecific() throws IOException {
		List<Application> applications = new ArrayList<>();
		applications.add(app1a);
		applications.add(app1b);

		String name = "test-application-non-specific";

		when(personService.get(any()))
				.thenReturn(List.of(personSummary1, personSummary2));
		when(personService.getOrThrow(PERSON_ID_1)).thenReturn(personSummary1);
		when(personService.getOrThrow(PERSON_ID_2)).thenReturn(personSummary2);

		doReturn(true).when(service).hasTAedBefore(anyLong(), anyLong());

		when(editionService.getOrThrow(EDITION_ID)).thenReturn(editionDetails);

		@SuppressWarnings("unchecked") ArgumentCaptor<List<String[]>> captor = ArgumentCaptor
				.forClass(List.class);
		service.exportApplications(name, applications, true);
		verify(csvService).writeToCSV(captor.capture(), any(), eq(name + ".csv"));

		assertThat(captor.getValue())
				.contains(
						new String[] { "Student Number", "Email", "Name", "TAed Before", "Status",
								"Student Response" },

						new String[] { "1",
								"Person1@email.com",
								"Person 1",
								"YES",
								"SUBMITTED",
								null },
						new String[] { "2",
								"Person2@email.com",
								"Person 2",
								"YES",
								"OFFERED",
								null });
	}

	@Test
	void exportApplicationsFlexDelft() throws IOException {
		List<Application> applications = new ArrayList<>();
		applications.add(app1c);
		applications.add(app1b);

		Person exporter = Person.builder()
				.id(PERSON_ID_1)
				.externalId(personSummary1.getExternalId())
				.username(personSummary1.getUsername())
				.displayName(personSummary1.getDisplayName())
				.email(personSummary1.getEmail())
				.number(personSummary1.getNumber())
				.build();

		String name = "test-application-flex-delft";

		when(personService.get(any()))
				.thenReturn(List.of(personSummary1, personSummary2));
		when(personService.getOrThrow(PERSON_ID_1)).thenReturn(personSummary1);
		when(personService.getOrThrow(PERSON_ID_2)).thenReturn(personSummary2);

		doReturn(true).when(service).hasTAedBefore(anyLong(), anyLong());

		when(editionService.getOrThrow(EDITION_ID)).thenReturn(editionDetails);
		when(cohortService.get(any())).thenReturn(List.of(cohortDetails));
		when(cohortService.getOrThrowById(anyLong())).thenReturn(cohortDetails);

		when(jobOfferService.getApprover(anyLong())).thenReturn(personSummary2);
		when(profileRepository.findByIdOrThrow(anyLong())).thenReturn(Profile.builder()
				.id(PERSON_ID_1)
				.payScale(PayScale.SA2)
				.build());
		when(coordinatorDefaultService.getOrCreate(anyLong())).thenReturn(CoordinatorDefault.builder()
				.programId(PROGRAM_ID)
				.coordinatorContractEmail("email@coordinator.com")
				.coordinatorContractName("Contract Name Default")
				.build());

		@SuppressWarnings("unchecked") ArgumentCaptor<List<String[]>> captor = ArgumentCaptor
				.forClass(List.class);
		service.exportApplicationsForFlexDelft(exporter, PROGRAM_ID, 1, name, applications, null);
		verify(csvService).writeToCSV(captor.capture(), any(), eq(name + ".csv"));

		assertThat(captor.getValue())
				.contains(
						new String[] { "Name",
								"NetID",
								"Email",
								"Contract Name",
								"Baan Code",
								"Approver Name",
								"Approver Email",
								"Coordinator",
								"Max Hours",
								"PayScale",
								"Contract Start Date",
								"Contract End Date",
								"Batch",
								"Remarks" },

						new String[] { "Person 1",
								"Person1",
								"Person1@email.com",
								null,
								null,
								"Person 2",
								"Person2@email.com",
								"Contract Name Default",
								"null",
								"SA2",
								"01-02-2023",
								"02-02-2023",
								"1",
								"" },

						new String[] { "Person 2",
								"Person2",
								"Person2@email.com",
								null,
								null,
								"Person 2",
								"Person2@email.com",
								"Contract Name Default",
								"null",
								"SA2",
								"01-02-2023",
								"02-02-2023",
								"1",
								"" });
	}

	@Test
	void getNumberOfUnhandledApplicationsTest() {
		when(service.getApplicationsForJobOffer(anyLong())).thenReturn(List.of());
		when(service.getStatsForApps(anyList())).thenReturn(new int[] { 10, 1, 2, 3 });

		assertEquals(4, service.getNumberOfUnhandledApplications(1L));
	}
}
