/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.Mockito.when;

import java.time.LocalDate;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.CohortDetailsDTO;
import nl.tudelft.labracore.api.dto.CohortSummaryDTO;
import nl.tudelft.labracore.api.dto.ProgramSummaryDTO;
import nl.tudelft.tam.dto.patch.CoordinatorDefaultPatchDTO;
import nl.tudelft.tam.dto.patch.embeddables.ExtraWorkDefaultPatchDTO;
import nl.tudelft.tam.dto.patch.embeddables.JobOfferDefaultPatchDTO;
import nl.tudelft.tam.dto.view.details.EditionDetailsOffersWorkDTO;
import nl.tudelft.tam.model.CoordinatorDefault;
import nl.tudelft.tam.model.embeddables.ExtraWorkDefault;
import nl.tudelft.tam.model.embeddables.JobOfferDefault;
import nl.tudelft.tam.repository.CoordinatorDefaultRepository;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
public class CoordinatorDefaultServiceTest {

	@Autowired
	CoordinatorDefaultService service;

	@Autowired
	CoordinatorDefaultRepository repository;

	@MockBean
	CohortService cohortService;

	@Autowired
	ModelMapper mapper;

	CoordinatorDefault cseDefault;
	Long CSE_DEFAULT_ID = 1L;

	@BeforeEach
	void setup() {
		repository.deleteAll();

		JobOfferDefault joDefault = JobOfferDefault.builder().contractName("TEST JO")
				.contractBaanCode("TEST BAAN JO").contractStartDate(LocalDate.now())
				.contractEndDate(LocalDate.now().plusWeeks(1)).applicationDeadline(LocalDate.now()).build();
		ExtraWorkDefault ewDefault = ExtraWorkDefault.builder().contractName("TEST EW")
				.contractBaanCode("TEST BAAN EW").build();

		cseDefault = CoordinatorDefault.builder().programId(CSE_DEFAULT_ID)
				.jobOfferDefault(joDefault).extraWorkDefault(ewDefault).build();

		repository.save(cseDefault);
	}

	@Test
	void findByIdOrThrowTest() {
		assertThat(service.findByIdOrThrow(CSE_DEFAULT_ID)).isEqualTo(cseDefault);
	}

	@Test
	void getOrCreate() {
		assertThat(service.getOrCreate(CSE_DEFAULT_ID)).isEqualTo(cseDefault);

		Long nonExistentId = 293289L;
		assertThrows(ResourceNotFoundException.class, () -> service.findByIdOrThrow(nonExistentId));
		assertThat(service.getOrCreate(nonExistentId))
				.isEqualTo(CoordinatorDefault.builder().programId(nonExistentId)
						.jobOfferDefault(
								JobOfferDefault.builder().contractBaanCode("").contractName("").build())
						.extraWorkDefault(
								ExtraWorkDefault.builder().contractBaanCode("").contractName("").build())
						.coordinatorContractName("")
						.coordinatorContractEmail("")
						.build());
	}

	@Test
	void addDefaultsToEditionsWithJobsAndWork() {
		EditionDetailsOffersWorkDTO edition = mapper.map(new EditionDetailsOffersWorkDTO()
				.cohort(new CohortSummaryDTO().id(1L)), EditionDetailsOffersWorkDTO.class);

		assertThat(edition.getDefaults()).isNull();

		CohortDetailsDTO cohortDetails = new CohortDetailsDTO().id(1L)
				.program(new ProgramSummaryDTO().id(CSE_DEFAULT_ID));
		when(cohortService.getOrThrowById(anyLong())).thenReturn(cohortDetails);

		List<EditionDetailsOffersWorkDTO> editions = List.of(edition);

		service.addDefaultsToEditionsWithJobsAndWork(editions);

		assertThat(edition.getDefaults()).isEqualTo(cseDefault);
	}

	@Test
	void patchCoordinatorDefault() {
		LocalDate date = LocalDate.now().plusWeeks(29292);

		JobOfferDefaultPatchDTO joPatchDto = new JobOfferDefaultPatchDTO(date, "patched name JO",
				"patched baan JO", date, date, "patched Description", "patched Hiring Message",
				"patched Reject Message");
		ExtraWorkDefaultPatchDTO ewPatchDto = new ExtraWorkDefaultPatchDTO("patched name EW",
				"patched baan EW", date);

		CoordinatorDefaultPatchDTO patchDto = new CoordinatorDefaultPatchDTO(joPatchDto, ewPatchDto,
				"patched coordinator name", "patched coordinator email");

		CoordinatorDefault prev = service.findByIdOrThrow(CSE_DEFAULT_ID);

		assertThat(prev.getJobOfferDefault().getApplicationDeadline()).isNotEqualTo(date);
		assertThat(prev.getJobOfferDefault().getContractName()).isNotEqualTo("patched name JO");
		assertThat(prev.getJobOfferDefault().getContractBaanCode()).isNotEqualTo("patched baan JO");
		assertThat(prev.getJobOfferDefault().getContractStartDate()).isNotEqualTo(date);
		assertThat(prev.getJobOfferDefault().getContractEndDate()).isNotEqualTo(date);
		assertThat(prev.getJobOfferDefault().getDescription()).isNotEqualTo("patched Description");
		assertThat(prev.getJobOfferDefault().getHiringMessage()).isNotEqualTo("patched Hiring Message");
		assertThat(prev.getJobOfferDefault().getRejectMessage()).isNotEqualTo("patched Reject Message");
		assertThat(prev.getExtraWorkDefault().getContractName()).isNotEqualTo("patched name EW");
		assertThat(prev.getExtraWorkDefault().getContractBaanCode()).isNotEqualTo("patched baan EW");
		assertThat(prev.getExtraWorkDefault().getDeclarationDeadline()).isNotEqualTo(date);
		assertThat(prev.getCoordinatorContractName()).isNotEqualTo("patched coordinator name");
		assertThat(prev.getCoordinatorContractEmail()).isNotEqualTo("patched coordinator email");

		service.patchCoordinatorDefault(CSE_DEFAULT_ID, patchDto);

		CoordinatorDefault post = service.findByIdOrThrow(CSE_DEFAULT_ID);

		assertThat(post.getJobOfferDefault().getApplicationDeadline()).isEqualTo(date);
		assertThat(post.getJobOfferDefault().getContractName()).isEqualTo("patched name JO");
		assertThat(post.getJobOfferDefault().getContractBaanCode()).isEqualTo("patched baan JO");
		assertThat(post.getJobOfferDefault().getContractStartDate()).isEqualTo(date);
		assertThat(post.getJobOfferDefault().getContractEndDate()).isEqualTo(date);
		assertThat(post.getJobOfferDefault().getDescription()).isEqualTo("patched Description");
		assertThat(post.getJobOfferDefault().getHiringMessage()).isEqualTo("patched Hiring Message");
		assertThat(post.getJobOfferDefault().getRejectMessage()).isEqualTo("patched Reject Message");
		assertThat(post.getExtraWorkDefault().getContractName()).isEqualTo("patched name EW");
		assertThat(post.getExtraWorkDefault().getContractBaanCode()).isEqualTo("patched baan EW");
		assertThat(post.getExtraWorkDefault().getDeclarationDeadline()).isEqualTo(date);
		assertThat(post.getCoordinatorContractName()).isEqualTo("patched coordinator name");
		assertThat(post.getCoordinatorContractEmail()).isEqualTo("patched coordinator email");
	}

}
