/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.mockito.Mockito.*;

import java.time.LocalDate;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.CsvSource;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.EditionControllerApi;
import nl.tudelft.labracore.api.dto.CohortSummaryDTO;
import nl.tudelft.labracore.api.dto.CourseSummaryDTO;
import nl.tudelft.labracore.api.dto.EditionCreateDTO;
import nl.tudelft.labracore.api.dto.EditionDetailsDTO;
import nl.tudelft.tam.cache.EditionCacheManager;
import nl.tudelft.tam.dto.view.details.EditionDetailsOffersWorkDTO;
import nl.tudelft.tam.dto.view.summary.ExtraWorkSummaryDTO;
import nl.tudelft.tam.dto.view.summary.JobOfferSummaryDTO;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class EditionServiceTest {

	@Autowired
	EditionService service;

	@Autowired
	EditionControllerApi editionControllerApi;

	@Autowired
	ModelMapper mapper;

	@MockBean
	JobOfferService jobOfferService;

	@MockBean
	ExtraWorkService extraWorkService;

	@MockBean
	EditionCacheManager editionCache;

	List<JobOfferSummaryDTO> offers;
	List<ExtraWorkSummaryDTO> work;
	EditionDetailsDTO ed1, ed2;
	EditionDetailsOffersWorkDTO edd1, edd2;

	@BeforeEach
	void setUp() {
		offers = List.of(
				JobOfferSummaryDTO.builder()
						.id(1011L).name("jo-11").editionId(10L).deadline(LocalDate.now())
						.build(),
				JobOfferSummaryDTO.builder()
						.id(1021L).name("jo-21").editionId(20L).deadline(LocalDate.now())
						.build());
		work = List.of(
				ExtraWorkSummaryDTO.builder()
						.id(2011L).name("ex-11").editionId(10L).deadline(LocalDate.now())
						.build(),
				ExtraWorkSummaryDTO.builder()
						.id(2021L).name("ex-21").editionId(20L).deadline(LocalDate.now())
						.build());

		CourseSummaryDTO course = new CourseSummaryDTO().id(1L).name("course-1").code("cc1");
		CohortSummaryDTO cohort = new CohortSummaryDTO().name("coho-1");

		ed1 = new EditionDetailsDTO().id(10L).name("edi-1").course(course).cohort(cohort);
		ed2 = new EditionDetailsDTO().id(20L).name("edi-2").course(course).cohort(cohort);

		edd1 = mapper.map(ed1, EditionDetailsOffersWorkDTO.class);
		edd1.setOffers(List.of(offers.get(0)));
		edd1.setWork(List.of(work.get(0)));
		edd2 = mapper.map(ed2, EditionDetailsOffersWorkDTO.class);
		edd2.setOffers(List.of(offers.get(1)));
		edd2.setWork(List.of(work.get(1)));
	}

	@Test
	void combineOffersAndWorkWithEditionsTest() {
		when(editionControllerApi.getEditionsById(anyList())).thenReturn(Flux.just(ed1, ed2));

		List<EditionDetailsOffersWorkDTO> res = service.combineOffersAndWorkWithEditions(offers, work);
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(edd1, edd2);

		verify(editionControllerApi).getEditionsById(anyList());
	}

	@Test
	void combineOffersAndWorkWithEditionsTestEmpty() {
		assertThat(service.combineOffersAndWorkWithEditions(List.of(), List.of())).isEmpty();

		verify(editionControllerApi, never()).getEditionsById(anyList());
	}

	@ParameterizedTest
	@CsvSource({ ", 2, 2",
			"cc1, 2, 2",
			"course-1, 2, 2",
			"coho, 2, 2",
			"edi-1, 2, 0",
			"edi-2, 0, 2",
			"jo, 1, 1",
			"ex, 1, 1",
			"asdfuhasdfha, 0, 0" })
	void filterTest(String filter, int count1, int count2) {
		List<EditionDetailsOffersWorkDTO> res = service.filter(List.of(edd1, edd2), filter);
		if (count1 == 0 && count2 == 0) {
			assertThat(res).isEmpty();
		} else if (count1 == 0) {
			assertThat(res.size()).isEqualTo(1);
			assertThat(res.get(0).getId()).isEqualTo(20L);
			assertThat(res.get(0).getWork().size() + res.get(0).getOffers().size()).isEqualTo(count2);
		} else if (count2 == 0) {
			assertThat(res.size()).isEqualTo(1);
			assertThat(res.get(0).getId()).isEqualTo(10L);
			assertThat(res.get(0).getWork().size() + res.get(0).getOffers().size()).isEqualTo(count1);
		} else {
			assertThat(res.size()).isEqualTo(2);
			assertThat(res.get(0).getId()).isEqualTo(10L);
			assertThat(res.get(0).getWork().size() + res.get(0).getOffers().size()).isEqualTo(count1);
			assertThat(res.get(1).getId()).isEqualTo(20L);
			assertThat(res.get(1).getWork().size() + res.get(1).getOffers().size()).isEqualTo(count2);
		}
	}

	@Test
	void getEditionsWithJobsAndWorkTest() {
		when(jobOfferService.getAllInEditions(anyList())).thenReturn(offers);
		when(extraWorkService.getAllInEditions(anyList())).thenReturn(work);
		when(editionControllerApi.getEditionsById(anyList())).thenReturn(Flux.just(ed1, ed2));

		List<EditionDetailsOffersWorkDTO> res = service.getEditionsWithJobsAndWork(List.of(10L, 20L));
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(edd1, edd2);

		verify(jobOfferService).getAllInEditions(anyList());
		verify(extraWorkService).getAllInEditions(anyList());
		verify(editionControllerApi).getEditionsById(anyList());
	}

	@Test
	void getAllEditionIdsTest() {
		when(editionControllerApi.getAllEditions()).thenReturn(Flux.just(ed1));

		assertThat(service.getAllEditionIds()).isEqualTo(List.of(ed1.getId()));

		verify(editionControllerApi).getAllEditions();
	}

	@Test
	void removePersonFromEditionTest() {
		when(editionControllerApi.removePersonFromEdition(anyLong(), anyLong())).thenReturn(Mono.empty());

		service.removePersonFromEdition(1L, 1L);

		verify(editionControllerApi).removePersonFromEdition(eq(1L), eq(1L));
	}

	@Test
	void getOrThrowTest() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(1L);

		when(editionCache.getOrThrow(anyLong())).thenReturn(edition);

		assertThat(service.getOrThrow(1L)).isEqualTo(edition);

		verify(editionCache).getOrThrow(eq(1L));
	}

	@Test
	void getEditionsByIdTest() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(1L);

		when(editionCache.get(anyList())).thenReturn(List.of(edition));

		assertThat(service.getEditionsById(List.of(1L))).isEqualTo(List.of(edition));

		verify(editionCache).get(eq(List.of(1L)));
	}

	@Test
	void getEditionByIdTest() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(1L);

		when(editionCache.getOrThrow(anyLong())).thenReturn(edition);

		assertThat(service.getEditionById(1L)).isEqualTo(edition);

		verify(editionCache).getOrThrow(eq(1L));
	}

	@Test
	void getListTest() {
		EditionDetailsDTO edition = new EditionDetailsDTO().id(1L);

		when(editionCache.get(anyList())).thenReturn(List.of(edition));

		assertThat(service.get(List.of(1L))).isEqualTo(List.of(edition));

		verify(editionCache).get(eq(List.of(1L)));
	}

	@Test
	void addCohortToEditionTest() {
		when(editionControllerApi.addCohortToEdition(anyLong(), anyLong())).thenReturn(Mono.empty());

		service.addCohortToEdition(1L, 1L);

		verify(editionControllerApi).addCohortToEdition(eq(1L), eq(1L));
	}

	@Test
	void addEditionTest() {
		EditionCreateDTO editionCreateDto = new EditionCreateDTO().name("Test");

		when(editionControllerApi.addEdition(any())).thenReturn(Mono.just(1L));

		assertThat(service.addEdition(editionCreateDto)).isEqualTo(1L);

		verify(editionControllerApi).addEdition(editionCreateDto);
	}
}
