/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.time.LocalDate;
import java.util.ArrayList;
import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.CsvSource;
import org.mockito.ArgumentCaptor;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.CohortSummaryDTO;
import nl.tudelft.labracore.api.dto.CourseSummaryDTO;
import nl.tudelft.labracore.api.dto.EditionDetailsDTO;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.tam.dto.create.ExtraWorkCreateDTO;
import nl.tudelft.tam.dto.patch.ExtraWorkPatchDTO;
import nl.tudelft.tam.dto.view.details.ExtraWorkDetailsDTO;
import nl.tudelft.tam.dto.view.summary.ExtraWorkSummaryDTO;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.exception.WrongDataCallException;
import nl.tudelft.tam.model.Declaration;
import nl.tudelft.tam.model.ExtraWork;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.repository.ExtraWorkRepository;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class ExtraWorkServiceTest {

	@SpyBean
	DeclarationService declarationService;

	@Autowired
	ExtraWorkService service;

	@Autowired
	ExtraWorkRepository repository;

	@MockBean
	CSVService csvService;

	@MockBean
	EditionService editionService;

	@MockBean
	PersonService personService;

	@Autowired
	ModelMapper mapper;

	final Long EDITION_ID_1 = 1001L;
	final Long EDITION_ID_2 = 1002L;

	long WORK_ID_1A, WORK_ID_1B, WORK_ID_2;

	ExtraWork work1a, work1b, work2;

	EditionDetailsDTO edition1, edition2;

	@BeforeEach
	void setup() {
		repository.deleteAll();

		work1a = repository.save(ExtraWork.builder().name("work 1a").editionId(EDITION_ID_1)
				.contractName("Extra-Work-1a-Contract")
				.baanCode("Extra Work Baan Code").maxHours(100).build());
		work1b = repository.save(ExtraWork.builder().name("work 1b").editionId(EDITION_ID_1)
				.deadline(LocalDate.now().minusDays(7)).contractName("Extra-Work-1b-Contract")
				.baanCode("Extra Work Baan Code").maxHours(100).build());
		work2 = repository.save(ExtraWork.builder().name("work 2").editionId(EDITION_ID_2)
				.deadline(LocalDate.now().plusDays(7)).contractName("Extra-Work-2-Contract")
				.baanCode("Extra Work Baan Code").maxHours(100).build());

		WORK_ID_1A = work1a.getId();
		WORK_ID_1B = work1b.getId();
		WORK_ID_2 = work2.getId();

		edition1 = new EditionDetailsDTO().name("Edition 1").id(EDITION_ID_1)
				.course(new CourseSummaryDTO().name("course-1").code("code-1"))
				.cohort(new CohortSummaryDTO().name("COHOrt-1"));
		edition2 = new EditionDetailsDTO().name("Edition 2").id(EDITION_ID_2)
				.course(new CourseSummaryDTO().name("course-2").code("code-2"))
				.cohort(new CohortSummaryDTO().name("Cohort-2"));
	}

	@Test
	void getByIdTest() {
		when(editionService.getEditionById(anyLong())).thenReturn(edition1);

		ExtraWorkDetailsDTO dto = mapper.map(work1a, ExtraWorkDetailsDTO.class);
		dto.setEdition(edition1);

		ExtraWorkDetailsDTO res = service.getById(WORK_ID_1A);
		assertThat(res).isEqualTo(dto);

		verify(editionService).getEditionById(EDITION_ID_1);
	}

	@Test
	void getEditionFromExtraWorkTest() {
		assertThat(service.getEditionFromExtraWork(WORK_ID_1A)).isEqualTo(EDITION_ID_1);
		assertThat(service.getEditionFromExtraWork(WORK_ID_2)).isEqualTo(EDITION_ID_2);
	}

	@Test
	void getAllOpenTest() {
		ExtraWorkSummaryDTO w1 = mapper.map(work1a, ExtraWorkSummaryDTO.class);
		ExtraWorkSummaryDTO w2 = mapper.map(work2, ExtraWorkSummaryDTO.class);

		List<ExtraWorkSummaryDTO> res = service.getAllOpen();
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(w1, w2);
	}

	@Test
	void getAllClosedTest() {
		ExtraWorkSummaryDTO w1 = mapper.map(work1b, ExtraWorkSummaryDTO.class);

		List<ExtraWorkSummaryDTO> res = service.getAllClosed();
		assertThat(res.size()).isEqualTo(1);
		assertThat(res).contains(w1);
	}

	@Test
	void newExtraWorkTest() {
		Long edId = 5008L;
		assertThat(repository.findAllByEditions(List.of(edId))).isEmpty();

		ExtraWorkCreateDTO dto = ExtraWorkCreateDTO.builder().name("new-work").editionId(edId)
				.contractName("new-work-contract").baanCode("new-baan").maxHours(100).build();
		service.newExtraWork(dto);

		List<ExtraWork> res = repository.findAllByEditions(List.of(edId));
		assertThat(res.size()).isEqualTo(1);
		assertThat(res.get(0).getName()).isEqualTo(dto.getName());
		assertThat(res.get(0).getEditionId()).isEqualTo(edId);
		assertThat(res.get(0).getContractName()).isEqualTo(dto.getContractName());
		assertThat(res.get(0).getBaanCode()).isEqualTo(dto.getBaanCode());
		assertThat(res.get(0).getMaxHours()).isEqualTo(dto.getMaxHours());
	}

	@Test
	void patchExtraWorkTest() {
		ExtraWorkPatchDTO dto = ExtraWorkPatchDTO.builder().name("patched-name")
				.contractName("patched-contract-name")
				.baanCode("patched-baan-code")
				.maxHours(23829)
				.deadline(LocalDate.now().plusWeeks(42)).build();

		ExtraWork res = repository.findByIdOrThrow(WORK_ID_1A);
		assertThat(res.getName()).isNotEqualTo(dto.getName());
		assertThat(res.getContractName()).isNotEqualTo(dto.getContractName());
		assertThat(res.getBaanCode()).isNotEqualTo(dto.getBaanCode());
		assertThat(res.getDeadline()).isNotEqualTo(dto.getDeadline());
		assertThat(res.getMaxHours()).isNotEqualTo(dto.getMaxHours());

		service.patchExtraWork(WORK_ID_1A, dto);

		res = repository.findByIdOrThrow(WORK_ID_1A);
		assertThat(res.getName()).isEqualTo(dto.getName());
		assertThat(res.getContractName()).isEqualTo(dto.getContractName());
		assertThat(res.getBaanCode()).isEqualTo(dto.getBaanCode());
		assertThat(res.getDeadline()).isEqualTo(dto.getDeadline());
		assertThat(res.getMaxHours()).isEqualTo(dto.getMaxHours());
	}

	@Test
	void closeExtraWorkTest() {
		assertThat(repository.findByIdOrThrow(WORK_ID_2).getDeadline()).isAfter(LocalDate.now().minusDays(1));

		service.closeExtraWork(WORK_ID_2);

		assertThat(repository.findByIdOrThrow(WORK_ID_2).getDeadline()).isBefore(LocalDate.now());
	}

	private List<ExtraWorkSummaryDTO> mapToSummary(List<ExtraWorkDetailsDTO> dtos) {
		List<ExtraWorkSummaryDTO> res = new ArrayList<>();
		for (ExtraWorkDetailsDTO dto : dtos) {
			res.add(ExtraWorkSummaryDTO.builder().id(dto.getId()).name(dto.getName())
					.contractName(dto.getContractName()).baanCode(dto.getBaanCode())
					.maxHours(dto.getMaxHours())
					.deadline(dto.getDeadline()).editionId(dto.getEditionId()).build());
		}
		return res;
	}

	@Test
	void getFilteredOpenTest() {
		when(editionService.getEditionsById(anyList())).thenReturn(List.of(edition1, edition2));
		assertThat(mapToSummary(service.getFilteredOpen(""))).isEqualTo(service.getAllOpen());
		assertThat(service.getFilteredOpen("auhsgljfsd").size()).isEqualTo(0);
	}

	@Test
	void getFilteredClosedTest() {
		when(editionService.getEditionsById(anyList())).thenReturn(List.of(edition1, edition2));
		assertThat(mapToSummary(service.getFilteredClosed(""))).isEqualTo(service.getAllClosed());
		assertThat(service.getFilteredClosed("auhsgljfsd").size()).isEqualTo(0);
	}

	@ParameterizedTest
	@CsvSource({ ", 3",
			"aghaufdns, 0",
			"work 1, 2",
			"WoRK 2, 1",
			"edition 1, 2",
			"course-1, 2",
			"code-2, 1",
			"cohort, 3",
			"cohort-2, 1" })
	void getFilteredForTest(String filter, int count) {
		when(editionService.getEditionsById(anyList())).thenReturn(List.of(edition1, edition2));

		List<ExtraWorkDetailsDTO> res = service.getFilteredFor(List.of(work1a, work1b, work2), filter);
		assertThat(res.size()).isEqualTo(count);

		verify(editionService).getEditionsById(List.of(EDITION_ID_1, EDITION_ID_1, EDITION_ID_2));
	}

	@Test
	void getFilteredForTestException() {
		when(editionService.getEditionsById(anyList())).thenReturn(List.of());

		assertThrows(WrongDataCallException.class,
				() -> service.getFilteredFor(List.of(work1a, work1b, work2), "filter"));
	}

	@Test
	void getAllInEditionsTest() {
		ExtraWorkSummaryDTO w1 = mapper.map(work1a, ExtraWorkSummaryDTO.class);
		ExtraWorkSummaryDTO w2 = mapper.map(work1b, ExtraWorkSummaryDTO.class);

		List<ExtraWorkSummaryDTO> res = service.getAllInEditions(List.of(EDITION_ID_1));
		assertThat(res.size()).isEqualTo(2);
		assertThat(res).contains(w1, w2);
	}

	@Test
	void getFilteredAllInEditionsTest() {
		when(editionService.getEditionsById(anyList())).thenReturn(List.of(edition1));

		assertThat(mapToSummary(service.getFilteredAllInEditions(List.of(EDITION_ID_1), "")))
				.isEqualTo(service.getAllInEditions(List.of(EDITION_ID_1)));
		assertThat(service.getFilteredAllInEditions(List.of(EDITION_ID_1), "auhsgljfsd").size()).isEqualTo(0);

		verify(editionService, times(2)).getEditionsById(List.of(EDITION_ID_1, EDITION_ID_1));
	}

	@Test
	void exportExtraWork() throws IOException {
		doReturn(List.of(
				Declaration.builder().amount(4)
						.person(Profile.builder().id(1L).payScale(PayScale.SA1).build())
						.extraWork(work1a)
						.status(Status.ACCEPTED).build(),
				Declaration.builder().amount(15)
						.person(Profile.builder().id(2L).payScale(PayScale.SA3).build())
						.extraWork(work1a)
						.status(Status.SUBMITTED).build(),
				Declaration.builder().amount(8)
						.person(Profile.builder().id(1L).payScale(PayScale.SA1).build())
						.extraWork(work1a)
						.status(Status.ACCEPTED).build()))
				.when(declarationService)
				.findAllByWorkId(anyLong());

		when(personService.getOrThrow(eq(1L)))
				.thenReturn(new PersonSummaryDTO().id(1L).number(100).username("user1")
						.email("user1@tudelft.nl").displayName("User 1"));
		when(personService.getOrThrow(eq(2L)))
				.thenReturn(new PersonSummaryDTO().id(2L).number(101).username("user2")
						.email("user2@tudelft.nl").displayName("User 2"));

		when(editionService.getOrThrow(anyLong()))
				.thenReturn(new EditionDetailsDTO().id(1L).name("Edition")
						.course(new CourseSummaryDTO().name("Course")));

		doCallRealMethod().when(declarationService).exportDeclarations(anyString(), anyList());

		@SuppressWarnings("unchecked") ArgumentCaptor<List<String[]>> captor = ArgumentCaptor
				.forClass(List.class);
		service.exportExtraWork(work1a.getId());
		verify(csvService).writeToCSV(captor.capture(), any(), eq("work_1a.csv"));

		assertThat(captor.getValue())
				.contains(
						new String[] { "Student Number", "Email", "Name", "PayScale", "Hours", "Status" },
						new String[] { "100", "user1@tudelft.nl", "User 1", "SA1", "3.0", "ACCEPTED" },
						new String[] { "101", "user2@tudelft.nl", "User 2", "SA3", "3.75", "SUBMITTED" });
	}

	@Test
	void findByIdOrThrowTest() {
		assertThat(service.findByIdOrThrow(WORK_ID_1A)).isEqualTo(work1a);
	}

}
