/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.AssertionsForClassTypes.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.MockitoAnnotations;
import org.modelmapper.ModelMapper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.dto.PersonSummaryDTO;
import nl.tudelft.labracore.api.dto.ProgramSummaryDTO;
import nl.tudelft.labracore.lib.security.user.Person;
import nl.tudelft.tam.dto.create.RaiseRequestCreateDTO;
import nl.tudelft.tam.dto.view.details.RaiseRequestDetailsDTO;
import nl.tudelft.tam.enums.PayScale;
import nl.tudelft.tam.enums.Status;
import nl.tudelft.tam.exception.InvalidDatabaseEntryException;
import nl.tudelft.tam.model.Profile;
import nl.tudelft.tam.model.RaiseRequest;
import nl.tudelft.tam.repository.RaiseRequestRepository;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
class RaiseRequestServiceMockTest {

	@Autowired
	RaiseRequestService raiseRequestService;

	@MockBean
	FileService fileService;

	@MockBean
	RaiseRequestRepository raiseRequestRepository;

	@MockBean
	PersonService personService;

	@MockBean
	ProgramService programService;

	@MockBean
	ProfileService profileService;

	@Autowired
	ModelMapper mapper;

	private static final Long REQUEST_ID = 7943816855L;
	private static final Long STUDENT_ID = 1656987665L;
	private static final Long TEACHER_ID = 8463319541L;
	private static final PayScale PS_FROM = PayScale.SA1;
	private static final PayScale PS_TO = PayScale.SA4;

	private static RaiseRequest request;
	private static Profile student;
	private static Profile teacher;

	@BeforeEach
	void setup() {
		MockitoAnnotations.openMocks(this);

		student = Profile.builder()
				.id(STUDENT_ID)
				.payScale(PS_FROM)
				.build();

		request = RaiseRequest.builder()
				.id(REQUEST_ID)
				.requestedBy(student)
				.payScale(PS_TO)
				.status(Status.SUBMITTED)
				.build();

		teacher = Profile.builder()
				.id(TEACHER_ID)
				.build();
	}

	@Test
	void approve() {
		when(raiseRequestRepository.findByIdOrThrow(anyLong())).thenReturn(request);
		when(profileService.findByIdOrThrow(anyLong())).thenReturn(teacher);

		assertThat(student.getPayScale()).isEqualTo(PS_FROM);
		assertThat(request.getStatus()).isEqualTo(Status.SUBMITTED);

		raiseRequestService.approve(request.getId(), Person.builder().id(TEACHER_ID).build());

		assertThat(student.getPayScale()).isEqualTo(PS_TO);
		assertThat(request.getStatus()).isEqualTo(Status.ACCEPTED);

		verify(raiseRequestRepository).findByIdOrThrow(REQUEST_ID);
		verify(profileService).findByIdOrThrow(TEACHER_ID);
	}

	@Test
	void reject() {
		when(raiseRequestRepository.findByIdOrThrow(anyLong())).thenReturn(request);
		when(profileService.findByIdOrThrow(anyLong())).thenReturn(teacher);

		assertThat(student.getPayScale()).isEqualTo(PS_FROM);
		assertThat(request.getStatus()).isEqualTo(Status.SUBMITTED);

		raiseRequestService.reject(request.getId(), Person.builder().id(TEACHER_ID).build());

		assertThat(student.getPayScale()).isEqualTo(PS_FROM);
		assertThat(request.getStatus()).isEqualTo(Status.REJECTED_BY_TEACHER);

		verify(raiseRequestRepository).findByIdOrThrow(REQUEST_ID);
		verify(profileService).findByIdOrThrow(TEACHER_ID);
	}

	@Test
	void getAllSubmittedThrows() {
		when(raiseRequestRepository.findAllByStatus(any(Status.class))).thenReturn(List.of(request));

		when(personService.getPeopleById(anyList())).thenReturn(List.of());

		assertThrows(InvalidDatabaseEntryException.class,
				() -> raiseRequestService.getAllSubmitted(List.of(1L)));

		verify(personService).getPeopleById(List.of(STUDENT_ID));
	}

	@Test
	void getAllSubmittedPass() throws InvalidDatabaseEntryException {
		when(raiseRequestRepository.findAllByStatus(any(Status.class))).thenReturn(List.of(request));
		String studentName = "abc-student-4984654";
		when(personService.getPeopleById(anyList()))
				.thenReturn(List.of(new PersonSummaryDTO().id(STUDENT_ID).externalId(studentName)));
		when(programService.getAllRelevantPrograms(anyLong()))
				.thenReturn(List.of(new ProgramSummaryDTO().id(1L).name("Test")));

		List<RaiseRequestDetailsDTO> result = raiseRequestService.getAllSubmitted(List.of(1L));

		assertThat(result.size()).isEqualTo(1);
		assertThat(result.get(0).getId()).isEqualTo(REQUEST_ID);
		assertThat(result.get(0).getCurrentPayScale()).isEqualTo(PS_FROM);
		assertThat(result.get(0).getPayScale()).isEqualTo(PS_TO);
		assertThat(result.get(0).getPerson().getId()).isEqualTo(STUDENT_ID);
		assertThat(result.get(0).getPerson().getExternalId()).isEqualTo(studentName);

		verify(personService).getPeopleById(List.of(STUDENT_ID));
	}

	@Test
	void newRaiseRequestFailUploadFile() throws IOException {
		RaiseRequestCreateDTO dto = mapper.map(request, RaiseRequestCreateDTO.class);

		MultipartFile file = new MockMultipartFile("abc", "abc".getBytes());

		when(raiseRequestRepository.findByRequesterAndSubmitted(anyLong())).thenReturn(Optional.empty());
		when(raiseRequestRepository.save(any(RaiseRequest.class))).thenReturn(request);

		doThrow(new IOException("test exception")).when(fileService).uploadFile(any(MultipartFile.class),
				anyLong(), anyString());

		assertFalse(raiseRequestService.newRaiseRequest(dto, file));

		verify(raiseRequestRepository).save(dto.apply());
		verify(raiseRequestRepository).delete(request);
		verify(fileService).uploadFile(eq(file), eq(request.getId()), anyString());
	}

}
