/*
 * TAM
 * Copyright (C) 2021 - Delft University of Technology
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package nl.tudelft.tam.service;

import static org.assertj.core.api.AssertionsForClassTypes.assertThat;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Set;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.transaction.annotation.Transactional;

import application.test.TestTAMApplication;
import nl.tudelft.labracore.api.EditionControllerApi;
import nl.tudelft.labracore.api.RoleControllerApi;
import nl.tudelft.labracore.api.dto.*;
import reactor.core.publisher.Flux;
import reactor.core.publisher.Mono;

@SpringBootTest(classes = TestTAMApplication.class)
@Transactional
public class RoleServiceTest {

	@Autowired
	RoleService service;

	@MockBean
	JobOfferService jobOfferService;

	@MockBean
	PersonService personService;

	@MockBean
	EditionService editionService;

	@Autowired
	RoleControllerApi roleControllerApi;

	@Autowired
	EditionControllerApi editionControllerApi;

	final long JOFFER_ID = 1001;
	final long PERSON_ID = 5001;
	final long EDITION_ID = 7000;

	RoleEditionDetailsDTO roleCorrectEdition, roleCorrectEditionHeadTA;
	RoleEditionDetailsDTO roleIncorrectEdition;

	@BeforeEach
	void setup() {
		roleCorrectEdition = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.STUDENT)
				.edition(new EditionSummaryDTO().id(EDITION_ID));

		roleCorrectEditionHeadTA = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.HEAD_TA)
				.edition(new EditionSummaryDTO().id(EDITION_ID));

		roleIncorrectEdition = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.STUDENT)
				.edition(new EditionSummaryDTO().id(0L));
	}

	@Test
	void hasRoleNonMatching() {
		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleIncorrectEdition));

		Boolean resultNotMatching = service.hasRole(PERSON_ID, EDITION_ID);

		assertThat(resultNotMatching).isFalse();
	}

	@Test
	void hasRoleMatching() {
		RoleEditionDetailsDTO matching = new RoleEditionDetailsDTO()
				.type(RoleEditionDetailsDTO.TypeEnum.STUDENT)
				.edition(new EditionSummaryDTO().id(EDITION_ID));

		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleCorrectEdition));

		Boolean resultMatching = service.hasRole(PERSON_ID, EDITION_ID);

		assertThat(resultMatching).isTrue();
	}

	@Test
	void removeRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleCorrectEdition));

		doNothing().when(editionService).removePersonFromEdition(anyLong(), anyLong());

		service.removeRole(PERSON_ID, JOFFER_ID);

		verify(editionService).removePersonFromEdition(EDITION_ID, PERSON_ID);
	}

	@Test
	void changeToTaPatchRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleCorrectEdition));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new Id().personId(PERSON_ID).editionId(EDITION_ID)));

		when(editionControllerApi.getEditionParticipants(anyLong())).thenReturn(
				Flux.just(new RolePersonDetailsDTO().person(new PersonSummaryDTO().id(PERSON_ID)))
		);

		service.changeToTa(PERSON_ID, JOFFER_ID, false);

		verify(roleControllerApi).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TA), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToTaPatchRole_HeadTaAlready() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleCorrectEdition));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new Id().personId(PERSON_ID).editionId(EDITION_ID)));

		when(editionControllerApi.getEditionParticipants(anyLong())).thenReturn(
				Flux.just(new RolePersonDetailsDTO()
						.person(new PersonSummaryDTO().id(PERSON_ID))
						.type(RolePersonDetailsDTO.TypeEnum.HEAD_TA)
				)
		);

		service.changeToTa(PERSON_ID, JOFFER_ID, false);

		verify(roleControllerApi, times(0)).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TA), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToTaPatchRole_HeadTaAlready_Demote() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleCorrectEdition));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new Id().personId(PERSON_ID).editionId(EDITION_ID)));

		when(editionControllerApi.getEditionParticipants(anyLong())).thenReturn(
				Flux.just(new RolePersonDetailsDTO()
						.person(new PersonSummaryDTO().id(PERSON_ID))
						.type(RolePersonDetailsDTO.TypeEnum.HEAD_TA)
				)
		);

		service.changeToTa(PERSON_ID, JOFFER_ID, true);

		verify(roleControllerApi).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TA), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToTaAddRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleIncorrectEdition));

		when(roleControllerApi.addRole(any()))
				.thenReturn(Mono.just(new Id().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToTa(PERSON_ID, JOFFER_ID, false);

		verify(roleControllerApi).addRole(new RoleCreateDTO()
				.edition(new EditionIdDTO().id(EDITION_ID))
				.person(new PersonIdDTO().id(PERSON_ID))
				.type(RoleCreateDTO.TypeEnum.TA));
	}

	@Test
	void changeToHeadTaPatchRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleCorrectEdition));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new Id().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToHeadTa(PERSON_ID, JOFFER_ID);

		verify(roleControllerApi).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.HEAD_TA), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToHeadTaAddRole() {
		when(jobOfferService.getEditionFromJobOffer(JOFFER_ID)).thenReturn(EDITION_ID);

		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleIncorrectEdition));

		when(roleControllerApi.addRole(any()))
				.thenReturn(Mono.just(new Id().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToHeadTa(PERSON_ID, JOFFER_ID);

		verify(roleControllerApi).addRole(new RoleCreateDTO()
				.edition(new EditionIdDTO().id(EDITION_ID))
				.person(new PersonIdDTO().id(PERSON_ID))
				.type(RoleCreateDTO.TypeEnum.HEAD_TA));
	}

	@Test
	void changeToTeacherPatchRole() {
		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleCorrectEdition));

		when(roleControllerApi.patchRole(any(), anyLong(), anyLong()))
				.thenReturn(Mono.just(new Id().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToTeacher(PERSON_ID, EDITION_ID);

		verify(roleControllerApi).patchRole(new RolePatchDTO().type(RolePatchDTO.TypeEnum.TEACHER), PERSON_ID,
				EDITION_ID);
	}

	@Test
	void changeToTeacherAddRole() {
		when(personService.getRolesForPerson(anyLong())).thenReturn(List.of(roleIncorrectEdition));

		when(roleControllerApi.addRole(any()))
				.thenReturn(Mono.just(new Id().personId(PERSON_ID).editionId(EDITION_ID)));

		service.changeToTeacher(PERSON_ID, EDITION_ID);

		verify(roleControllerApi).addRole(new RoleCreateDTO()
				.edition(new EditionIdDTO().id(EDITION_ID))
				.person(new PersonIdDTO().id(PERSON_ID))
				.type(RoleCreateDTO.TypeEnum.TEACHER));
	}

	@Test
	void getRolesByIdTest() {
		RoleDetailsDTO role = new RoleDetailsDTO().id(new Id().personId(PERSON_ID).editionId(EDITION_ID));

		when(roleControllerApi.getRolesById(anySet(), anySet())).thenReturn(Flux.just(role));

		assertThat(service.getRolesById(List.of(EDITION_ID), List.of(PERSON_ID))).isEqualTo(List.of(role));

		verify(roleControllerApi).getRolesById(Set.of(EDITION_ID), Set.of(PERSON_ID));
	}

}
